#!/usr/bin/env bash
#
# Copyright 2020 bonree.
# All rights reserved.
#
# Install script for Bonree c/c++ SDK agent
#

usage() {
cat << EOF
  Usage: `basename $0`
  Options:
    -h,--help                     Show Help
    --prefix                      install architecture-independent files in prefix,default /opt
    --ignore-permissions          Ignores lack of write permissions on log and read permissions on the sdk folder
EOF
}

prefix=/opt
canonicalNameOfThisFile=`readlink -f "$0"`
containingDir=`dirname "$canonicalNameOfThisFile"`
agentVersionId=$(cat ${containingDir}/version.txt)

ok=1

while test $# -gt 0 
do
    # get optarg
    case "$1" in
        -*=*) optarg=`echo "$1" | sed 's/[-_a-zA-Z0-9]*=//'` ;;
        *) optarg= ;; 
    esac

    case "$1" in
        -h|--help)
            usage
            exit 1
            ;;
        --ignore-permissions)
            ignoreFilePermissions="true"
            ;;
        --prefix=*)
            prefix="$optarg"
            ;; 
        *)
            ok=0
            echo "Invalid option: '--${OPTARG}'" >&2
            ;;
    esac
    shift
done

if [ ! -d "${prefix}" ] ; then
    mkdir ${prefix}
    if [ ! $? -eq 0 ] ; then
        echo -e "\033[;31;1mprefix installation directory, \"${prefix}\", cant't be created. \033[m" >&2
        ok=0
    fi
fi

installDir=${prefix}/bonree/apm/agent/c

if [ ! -d "${installDir}/${agentVersionId}" ] ; then
    mkdir -p ${installDir}/${agentVersionId}
    if [ ! $? -eq 0 ] ; then
        echo -e "\033[;31;1mprefix installation directory, \"${installDir}\", cant't be created. \033[m" >&2
        ok=0
    fi
fi

sdkLogDir=${prefix}/bonree/apm/logs/c
if [ ! -d "${sdkLogDir}" ] ; then
    mkdir -p "${sdkLogDir}"
    if [ ! $? -eq 0 ] ; then
        echo -e "\033[;31;1m c agent log output directory, \"${sdkLogDir}\", cant't be created. \033[m" >&2
        ok=0
    fi
fi

if [ "$ok" != 1 ]; then
    usage
    exit 1
fi

################################################################################
# Define functions
################################################################################

log() {
    echo "$@" >> ${install_log}
}

log_error() {
    echo "[Error] $@" >> ${install_log}
    echo "[Error] $@" >&2
}

fatal_error() {
    log_error $1
    cat >&2 <<EOF

Bonree c/c++ SDK Agent installation has failed. Please check $install_log for
possible causes. Please also attach it when filing a bug report.
EOF

    exit 1
}

checkPathIsAccessible() {
    local __resultVarName currentPath findResult __result
    __resultVarName="$1"
    currentPath="$2"
    __result=""
    while [ "${currentPath}" != "/" ] ; do
        findResult=$(find "${currentPath}" -maxdepth 0 -perm -0005 2>/dev/null)
        if [ -z "${findResult}" ] ; then
            __result="${currentPath}"
            break
        fi
        currentPath=$(dirname "${currentPath}")
    done
    eval $__resultVarName="$__result"
}


datestamp=`date +%Y%m%d_%H%M%S 2>/dev/null`
install_log=/tmp/bonree_sdk_install_${datestamp}.log
rm -f $install_log > /dev/null 2>&1
cat > $install_log <<EOF
Bonree c/c++ SDK agent installation log
Version: ${agentVersionId}
Date: `date 2>/dev/null`

Hostname: `hostname`
Location: ${containingDir}
Install: ${installDir}
System: `uname -a 2>/dev/null`
User: `id`
Environment:
################################################################################
`env | sort`
################################################################################
EOF

echo "Install script for Bonree c/c++ SDK Agent ${agentVersionId}. sdk will install at ${installDir}."


## Installed files to installDir
ldconfDir="/etc/ld.so.conf.d"
ldconfFile="${ldconfDir}/bonree_sdk_native.conf"

rootDirCheck=
checkPathIsAccessible rootDirCheck "${installDir}"
if [ -n "${rootDirCheck}" ] ; then
    if [ "${ignoreFilePermissions}" = "true" ] ; then
        log "c/c++ SDK directory '${installDir}' is not readable by all users"
    else
        log_error "c/c++ SDK directory '${installDir}' is not readable by all users"
        fatal_error "Change the permissions of '${rootDirCheck}' or specify --ignore-permissions"
    fi
fi

# copy files to installDir
if [ ${containingDir} != ${installDir}/${agentVersionId} ]; then
        cp -rf ${containingDir}/* ${installDir}/${agentVersionId}/.
else
        echo "installDir is same to containingDir, ignore copy files"
fi

agentLogDir="${installDir}/${agentVersionId}"

chmod 0777 "$agentLogDir"
chmod 0770 -R "${installDir}/${agentVersionId}/lib" "${installDir}/${agentVersionId}/examples"

# Write shared library path to /etc/ld.so.conf.d/
if [ -w "${ldconfDir}" ]; then
    echo "${installDir}/${agentVersionId}/lib" > "${ldconfFile}"
    [ -x `which ldconfig` ] && ldconfig
else
    # No permissions to dir or to an existing file. Ignore.
    log "Cound not write shared library path to '${ldconfFile}'"
fi

echo "Install success"
