'use strict';

var os = require('os');
var fs = require('fs');
var util = require('util');
var path = require('path');
var EventEmitter = require('events').EventEmitter;
var Logger = require('./logger').Logger;
var Timers = require('./timers').Timers;
var System = require('./system').System;
var Wrap = require('./wrap').Wrap;
var Thread = require('./thread').Thread;
var Profiler = require('../services/profiler').Profiler;
var Configuration = require('../services/config').Configuration;
var Tracer = require('../models/tracer').Tracer;
var BaseHook = require('../hook/core/base_hook').BaseHook;
var Reporter = require('../services/reporter').Reporter;
var AgentVer = require('../../version.json').version;
function Agent() {
    this.initialized = false;
    this.nextId = Math.round(Math.random() * Math.pow(10, 6));
    // predefine options
    this.debug = false;

    EventEmitter.call(this);

    // create modules
    this.logger = new Logger(this);
    this.timers = new Timers(this);
    this.system = new System(this);
    this.wrap = new Wrap(this);
    this.thread = new Thread(this);
    //create services
    this.profiler = new Profiler(this);
    this.config = new Configuration(this);

    this.baseHook = new BaseHook(this);
    this.tracer = new Tracer(this);
    var versionArr=process.version.split('.');
    var majorVer = +versionArr[0].substr(1);
    if (majorVer < 8) {
        var CoTracer=require('../models/co_trace');
        this.asyncTracer=new CoTracer(this);
        this.asyncTracer.enableAsyncHooks();
    }
    else if (majorVer != 0) {
        var AsyncTracer = require('../models/async_trace');
        this.asyncTracer = new AsyncTracer(this);
        this.asyncTracer.enableAsyncHooks();
    }
}

util.inherits(Agent, EventEmitter);


Agent.prototype.init = function(opts) {
    var self = this;
    var ret;
    if (self.initialized) return;
    opts.agentVersion = AgentVer;
    opts.agentInstallDir = __dirname;
    opts.agentInstallTimestamp = fs.statSync(__filename).ctime.getTime();

    self.initialized = true;
    
    self.setMaxListeners(15);

    self.config.init(opts);

    self.timers.init();

    self.system.init();

    self.wrap.init();

    self.thread.init();

    self.logger.init();

    //判断sdk库是否存在
    try{
        var path;
        path = process.env.BONREE_SMARTAGENT_SDK_PATH;
        fs.accessSync(path,fs.F_OK);
    }catch (e) {
        self.logger.error("failed to open sdk library :"+process.env.BONREE_SMARTAGENT_SDK_PATH);
        return;
    }
    var nodeModuleVersion = process.versions.modules;
    //判断当前node版本是否被探针支持
    if (!(nodeModuleVersion >= 64 || nodeModuleVersion == 48 || nodeModuleVersion == 57)) {
        self.logger.error("Bonree Node.js agent doesn't support this version. NODE_MODULE_VERSION : " + nodeModuleVersion);
        return;
    }

    self.profiler.init();

    // Prepare probes.
    self.loadWraps();
    try{
        if(self.config.config)
        {
            self.reporter = new Reporter(this);
        }
        ret = self.reporter.init();
    }catch (e) {
		self.logger.error("catch exception:" + e);
        return false;
    }
    return ret;

};

Agent.prototype.loadWraps = function() {
    var self = this;

    // Dynamic probes.
    var hookCons = [];
    var keyhookCons = [];
    var constructHooks = ['next','nuxt'];
    // var constructHooks = ['promise'];
    // if(process.version<'v8.0.0'){
    //     constructHooks.push('request');
    // };
    hookCons.push(require('../hook/core/fs_hook').FsHook); //add new
    // hookCons.push(require('../hook/core/net_hook').NetHook); //add new
    hookCons.push(require('../hook/core/event_hook').EventHook); //add new
    hookCons.push(require('../hook/http_hook').HttpHook);
    hookCons.push(require('../hook/mysql_hook').MysqlHook);
    hookCons.push(require('../hook/mongodb_hook').MongodbHook);
    hookCons.push(require('../hook/redis_hook').RedisHook);
    hookCons.push(require('../hook/hiredis_hook').HiRedisHook);
    hookCons.push(require('../hook/rabbitjs_hook').RabbitJsHook);
    hookCons.push(require('../hook/express_hook').ExpressHook);
    hookCons.push(require('../hook/ioredis_hook').IoredisHook);
    hookCons.push(require('../hook/koa2_hook').Koa2Hook);
    hookCons.push(require('../hook/egg_hook').EggHook);
    hookCons.push(require('../hook/pg_hook').PgHook);
    hookCons.push(require('../hook/mysql2_hook').Mysql2Hook);
    // hookCons.push(require('../hook/promise_hook').PromiseHook);
    hookCons.push(require('../hook/bluebird_hook').BluebirdHook);
    hookCons.push(require('../hook/when_hook').WhenHook);
    hookCons.push(require('../hook/q_hook').QHook);
    hookCons.push(require('../hook/async_hook').AsyncHook);
    //hookCons.push(require('../hook/core/any_promise_hook').AnyPromiseHook);
    hookCons.push(require('../hook/fullgoal_redis_hook').FullgoalRedisHook);
    hookCons.push(require('../hook/loopback_hook').LoopbackHook);
    hookCons.push(require('../hook/strong_remoting_hook').StrongRemotineHook);
    hookCons.push(require('../hook/loopback_datasource_hook').LoopbackDatasourceHook);
    hookCons.push(require('../hook/loopback_connector_mongodb_hook').LoopbackMongodbHook);
    hookCons.push(require('../hook/loopback_connector_mysql_hook').LoopbackMysqlHook);
    hookCons.push(require('../hook/loopback_connector_pg_hook').LoopbackPgHook);
    hookCons.push(require('../hook/connect_hook').ConnectHook);
    hookCons.push(require('../hook/nuxt_hook').NuxtHook);
    hookCons.push(require('../hook/next_hook').NextHook);
    hookCons.push(require('../hook/request_hook').RequestHook);
    hookCons.push(require('../hook/superagent_hook').SuperagentHook);
    hookCons.push(require('../hook/hapi_hook').HapiHook);
    hookCons.push(require('../hook/cron_hook').CronHook); //add new
    keyhookCons.push(require('../hook/amqplib_hook').AmqplibHook);

    var packageHooks = {};
    var keyHooks = {};
    hookCons.forEach(function(hookCon) {
        var hook = new hookCon(self);
        hook.packages.forEach(function(pkg) {
            packageHooks[pkg] = hook;
        });
    });
    keyhookCons.forEach(function (hookCon) {
        var hook = new hookCon(self);
        hook.keys.forEach(function (key) {
          keyHooks[key] = hook;
        });
    });

    var cnt=0;
    // on demand probe attaching
    var orig = Object.getPrototypeOf(module).require;
    Object.getPrototypeOf(module).require = function() {
        var ret = orig.apply(this, arguments);
        var args = arguments;
        if (args[0].indexOf('meap_redis_connection') != -1) args[0] = path.basename(args[0],'.js');
        var hook = packageHooks[args[0]];
        if (constructHooks.indexOf(args[0]) !== -1 && !!hook) {
            ret = hook.hookConstructor(ret);
        }

        var obj = this;

        try{
            if(hook) {
                hook.attach(ret, args[0]);
            }
            else {
                hook = keyHooks[args[0]];
                if(hook && hook.match(obj.id))
                {
                    hook.attach(ret, args[0]);
                }
            }
        }catch(err)
        {
            self.logger.error('Error on attaching'+args[0]);
        }
        return ret;
    };

    // Explicit probes.
    var ProcessHook = require('../hook/process_hook').ProcessHook;
    new ProcessHook(self).attach(process);
    // var GlobalHook = require('../hook/globals_hook').GlobalHook;
    // new GlobalHook(self).attach(global);
};


Agent.prototype.getNextId = function() {
    return this.nextId++;
};


Agent.prototype.destroy = function() {
    try {
        this.emit('destroy');
    }
    catch(err) {
        this.logger.error(err);
    }

    this.removeAllListeners();
};

Agent.prototype.profile = Agent.prototype.init;

Agent.prototype.isEnabled = function () {
    var self = this;
    return self.config.isEnabled();
};

Agent.prototype.mainModule = function (appModule, callBack) {
    let ModuleWrapper = require('../hook/module_wrapper').ModuleWrapper;
    new ModuleWrapper(this).attach(appModule, callBack);
}
var Bonree = function() {
    var self = this;
    var agent = new Agent();
    [ 'profile', 'mainModule'
    ].forEach(function(meth) {
        self[meth] = function() {
            return agent[meth].apply(agent, arguments);
       };
    });
};

Bonree.Agent = Agent;

//exports = module.exports = new Bonree();
module.exports = new Bonree();
