'use strict';

var util = require('util');
var BackendType =  require('../common/constant').BackendType;
function MongodbHook(agent) {
    this.agent = agent;
    this.baseHook = agent.baseHook;
    this.packages = ['mongodb', 'mongodb-core', 'mongoose'];
    this.errContexs = [];
}
exports.MongodbHook = MongodbHook;

function shouldIgnore(segment) {
    if (!segment) return true;
    var filePath = segment.filename;
    if (filePath.indexOf('/node_modules/mongodb/') != -1 || filePath.indexOf('/node_modules/mongodb-core/') != -1)
    {
        return true;
    }
    return false;
}

MongodbHook.prototype.makeExitCallInfo = function(serverPool, commandName, commandArgs) {
    if (!commandArgs.collectionName){
        commandArgs.collectionName = 'unknown';
    }
    var supportedProperties = {
        'serverpool': serverPool.map(function (serverinfo) {
            return serverinfo.host+':'+serverinfo.port;
        }).join(';'),
        DATABASE : commandArgs.databaseName,
        COLLECTION: commandArgs.collectionName,
        VENDOR : "Mongo"
    };


    return {
        'backendType': BackendType.BACKEND_TYPE_MONGODB,
        'name':util.format('%s://%s/%s',supportedProperties.VENDOR,supportedProperties.serverpool,supportedProperties.COLLECTION),
        'connType': 'Mongo',
        'dbVendor': supportedProperties.VENDOR,
        'dbVersion': supportedProperties.VERSION,
        'dbSchema': supportedProperties.DATABASE,
        'is_async': false,
        'params': JSON.stringify(commandArgs),
        'operation': commandName,
        'host': serverPool.map(function (serverinfo) {
            return serverinfo.host;
        }).join(';'),
        'port': serverPool.map(function (serverinfo) {
            return serverinfo.port;
        }).join(';'),
        'url':util.format('%s://%s/%s',supportedProperties.VENDOR,supportedProperties.serverpool,supportedProperties.COLLECTION),
    };
};

MongodbHook.prototype.makeConnectExitCallInfo = function(conStr) {
    var serverPool = [];
    serverPool.push({'host':'unknown','port':'unknown'});

    var commandArgs = {
        databaseName: 'unknown',
        collectionName: 'unknown'
    };

    return this.makeExitCallInfo(serverPool, 'connect', commandArgs);
};

MongodbHook.prototype.makeModelExitCallInfo = function(obj, args, commandName, txn) {
    if (!this.baseHook.needExitCall(txn)) return null;
    var host = obj.host;
    var port = obj.port;
    var database = obj.name;
    var serverPool = [];
    if (obj.hosts && typeof Array.isArray(obj.hosts) && obj.hosts.length > 0) {
        serverPool = obj.hosts;
    } else {
        serverPool.push({'host':host,'port':port});
    }

    var readMethod = ['find', 'findOne', 'findById', 'count'];
    
    var colName;
    if (obj.collections) {
        for (var k in obj.collections) {
            if (k !== '_proto_') {
                colName = k;
                break;
            }
        }
    }
    var commandArgs = {
        databaseName: database,
        collectionName: !colName ? 'unknown' : colName
    };
    if (readMethod.indexOf(commandName) !== -1) {
        commandArgs.query = args[0] || undefined;
    }

    return this.makeExitCallInfo(serverPool, commandName, commandArgs);
};

MongodbHook.prototype.getServerPool = function(db) {
    var serverPool = [];

    var serverConfig = db && db.serverConfig || db.serverDetails || db.options;
    if(serverConfig) {
        if(serverConfig.s && serverConfig.s.host && serverConfig.s.port) {
            serverPool.push({'host': serverConfig.s.host, 'port': serverConfig.s.port});
        }
        else if(serverConfig.host && serverConfig.port) {
            serverPool.push({'host': serverConfig.host, 'port' : serverConfig.port});
        }
        else if(Array.isArray(serverConfig.servers)) {
            serverConfig.servers.forEach(function(server) {
                serverPool.push({'host': server.host,'port': server.port});
            });
        }
    }

    if (serverPool.length) {
        serverPool = serverPool.sort();
    }

    return serverPool;
};

MongodbHook.prototype.makeDbExitCallInfo = function(obj, args, commandName, txn, conExitCall) {
    var self = this;
    var exitCallInfo;
    var query;
    if (!this.baseHook.needExitCall(txn)) return null;

    var s =  (obj._isDb ? obj : obj.s.db);
    var serverPool = self.getServerPool(s);

    if (serverPool.length) {
        var opts = {};
        if (args[1] && args[1].query) {
            query = args[1].query;//JSON.stringify(args[1].query);
            Object.keys(args[1]).forEach(function(key) {
                if (key !== 'query') {
                    opts[key] = args[1][key];
                }
            });
        }

        var commandArgs = {
            databaseName: obj.s.dbName || obj.s.databaseName,// : args[0].split('.')[0],
            collectionName: obj.s.name || obj.collectionName || 'unknown',// : args[0].split('.')[1],
            query: query,
            queryOptions: opts,
            numberToSkip: opts.skip,
            numberToReturn: opts.limit
        };

        if(obj.s.auths && obj.s.auths.length > 0) {
            commandArgs.auth = obj.s.auths[0];
        }

        exitCallInfo = this.makeExitCallInfo(serverPool, commandName, commandArgs);
        if (!!conExitCall && !!exitCallInfo) {
            conExitCall.name = exitCallInfo.name;
            conExitCall.host = exitCallInfo.host;
            conExitCall.port = exitCallInfo.port;
            conExitCall.url = exitCallInfo.url;
            conExitCall.dbSchema = exitCallInfo.dbSchema;
            var conParamArgs = {
                databaseName:commandArgs.databaseName,
                collectionName:commandArgs.collectionName
            };
            conExitCall.params = JSON.stringify(conParamArgs);
        }
    }

    return exitCallInfo;
};




MongodbHook.prototype.checkAndaddErrorInfo = function(args, txn) {
    if (!args || !txn) return;
    var error = args[0];
    var reply = args[1];
    if(!error && reply && reply.result && reply.result.writeErrors)
    {
        error = reply.result.writeErrors[0];
    }
    // var error = this.bonree.proxy.getErrorObject(args);
    if ( error && error.errmsg)
    {
        error.stack = error.errmsg;
        error.name='mongodb';
    }
    if(error) txn.addException(error);
};


MongodbHook.prototype.attach = function(obj, name) {
    var self = this;
    var wrap = self.agent.wrap;
    var profiler = self.agent.profiler;
    var localBaseHook = self.baseHook;
    var tracer = self.agent.tracer;

    if(obj.__bonrAttached__) return;
    obj.__bonrAttached__ = true;

    function wrapDb(db, conExitCall) {
        var dbMethodsWithExitCall = ['addUser','authenticate','collections','command','createCollection',
            'createIndex','dropCollection','dropDatabase','ensureIndex','eval','executeDbAdminCommand',
            'indexInformation','removeUser','renameCollection','stats'];
        var dbMethodsWithoutExitCall = ['logout','open','close'];
        //Simultaneous compatible synchronous and asynchronous invocation
        dbMethodsWithExitCall.forEach(function(method) {
            //var txn = db._txn;
            localBaseHook.wrapAsyncMethod(db, 'db.' + method, null, function (obj, args, locals) {
                locals.transaction = db._txn;
                if (locals.transaction) locals.parentSegment = locals.transaction.getActiveSegment();
                locals.wrapBefore = function(obj, args) {
                    self.checkAndaddErrorInfo(args, locals.transaction);
                };
                obj._isDb = true;
                if (method == 'createCollection' || method == 'dropCollection') {
                    obj.collectionName = args[0];}
                else if (method == 'renameCollection') obj.collectionName = args[1];
                locals.exitCallInfo = self.makeDbExitCallInfo(obj, args, method, locals.transaction, conExitCall);
                if (!!locals.exitCallInfo) {
                    if (method == 'command' || method == 'createIndex') {
                        locals.needTrace = true;
                        locals.shouldIgnore = shouldIgnore;
                    }
                }
            });
        });
        dbMethodsWithoutExitCall.forEach(function(method) {
            localBaseHook.wrapAsyncMethod(db, 'db.' + method);
        });

        localBaseHook.wrapSyncMethod(db, 'db.collection', null, function(obj, args, ret) {
            var collection = ret;
            var collectionMethods = [
                'find','count','findOne','insert', 'insertOne','bulkWrite', 'insertMany', 'update', 'updateOne', 'replaceOne',
                'findOneAndReplace','findOneAndUpdate','findAndModify','updateMany','remove','removeOne','removeMany',
                'deleteOne','deleteMany','findOneAndDelete','findAndRemove','drop','aggregate','mapReduce', 'group',
                'parallelCollectionScan','createIndex','createIndexes','dropIndex','dropIndexes','dropAllIndexes','reIndex',
                'indexExists','indexes','distinct','geoHaystackSearch','geoNear','isCapped','options','rename','save','stats',
                'indexInformation','ensureIndex'
            ];
            collectionMethods.forEach(function(command) {
                var txn = db._txn;
                if (collection.hasOwnProperty('_txn')) txn = collection._txn;
                var time;
                if (command == 'find') {
                    localBaseHook.wrapSyncMethod(collection, 'collection.' + command, function (obj, args, locals) {
                        locals.transaction = txn || tracer.getTransaction();
                        if (locals.transaction) locals.parentSegment = locals.transaction.getActiveSegment();
                        locals.exitCallInfo = self.makeDbExitCallInfo(obj, args, command, locals.transaction, conExitCall);
                        },
                        function(obj, args, ret) {
                            var curAsyncMethods = ['each','forEach','toArray','count','close'];
                            curAsyncMethods.forEach(function(method) {
                                localBaseHook.wrapAsyncMethod(ret, 'Cursor.' + method);
                            });
                        });
                } else {
                    localBaseHook.wrapAsyncMethod(collection, 'collection.' + command, null, function (obj, args, locals) {
                        locals.transaction = txn || tracer.getTransaction();
                        if (locals.transaction) locals.parentSegment = locals.transaction.getActiveSegment();
                        locals.wrapBefore = function(obj, args) {
                            self.checkAndaddErrorInfo(args, locals.transaction);
                        };
                        locals.exitCallInfo = self.makeDbExitCallInfo(obj, args, command, locals.transaction, conExitCall);
                    });
                }
            });
        });
    }

    if (name == 'mongoose') {
        var propMethods = ['save', 'increment', 'remove'];
        var staticMethods = ['connect','init','ensureIndexes','createIndexes','remove',
            'deleteOne','deleteMany','where','find','findById','findOne','count',
            'distinct','findOneAndUpdate','findByIdAndUpdate','update',
            'findOneAndRemove','findByIdAndRemove','insertMany','bulkWrite',
            'updateMany','updateOne','replaceOne','mapReduce','aggregate',
            'geoSearch'];
        propMethods.forEach(function(method) {
            localBaseHook.wrapAsyncMethod(obj.Model.prototype, 'model.' + method, null, function(obj, args, locals) {
                locals.wrapBefore = function(obj, args) {
                    self.checkAndaddErrorInfo(args, locals.transaction);
                };
                locals.transaction = tracer.getTransaction();
                if (locals.transaction) locals.parentSegment = locals.transaction.getActiveSegment();
                var command = (method == 'save' ? 'insert' : method);
                locals.exitCallInfo = self.makeModelExitCallInfo(obj.db, args, command, locals.transaction);
            });
        });
        staticMethods.forEach(function(method) {
            localBaseHook.wrapAsyncMethod(obj.Model, 'model.' + method, null, function(obj, args, locals) {
                locals.wrapBefore = function(obj, args) {
                    self.checkAndaddErrorInfo(args, locals.transaction);
                };
                locals.transaction = tracer.getTransaction();
                if (locals.transaction) locals.parentSegment = locals.transaction.getActiveSegment();
                var command;
                if (method == 'init') command = 'ensureIndexes';
                else if (method == 'where') command = 'find';
                else command = method;
                locals.exitCallInfo = self.makeModelExitCallInfo(obj.db, args, command, locals.transaction);
            });
        });
    }
    if (name == 'mongodb') {
        localBaseHook.wrapAsyncMethod(obj.MongoClient, 'MongoClient.connect', null, function(obj, args, locals) {
            if (args.length < 1) return;
            locals.transaction = obj._txn || tracer.getTransaction();
            if (locals.transaction) locals.parentSegment = locals.transaction.getActiveSegment();
            locals.exitCallInfo = self.makeConnectExitCallInfo(args[0]);
            self.errContexs.push(locals.transaction);

            if (args.length < 2) return;
            locals.wrapBefore = function(obj, args) {
                if (args.length < 2) return;
                self.checkAndaddErrorInfo(args, locals.transaction);
                var db = args[1];
                if (db == null) return;
                db._txn = locals.transaction;
                var conExitCall = locals.curExitCall;
                localBaseHook.wrapSyncMethod(db, 'db.db', null, function(obj, args, ret) {
                    ret._txn = locals.transaction;
                    wrapDb(ret, conExitCall);
                });
                wrapDb(db, conExitCall);

            };
        });
    }
    if (name == 'mongodb-core') {
        // driver 2.x
        var commands = [
            'connect','cursor', 'insert', 'update', 'remove'
        ];
        // queries via a cursor find command must be reported after they complete:
        commands.forEach(function(command) {
            var commandName = command == 'cursor' ? 'find' : command;
            //The class name requires manual expenditure, or the class name of the called function is presumed to be inaccurate according to the name of the caller's class.
            localBaseHook.wrapAsyncMethod(obj.Server.prototype, 'mongodb.' + command, null, function(obj, args) {
                var errContex = self.errContexs.shift();
                // obj.on('error', function(err) {
                //     if (errContex) errContex.addException(err);
                //     return;
                // });
            });
        });
    }


    if (obj.Db && obj.Db.prototype && obj.Db.prototype._executeQueryCommand) {
        // driver 1.x
        var internalCommands = [
            '_executeQueryCommand',
            '_executeInsertCommand',
            '_executeUpdateCommand',
            '_executeRemoveCommand'
        ];

        var commandMap = {
            '_executeQueryCommand': 'find',
            '_executeInsertCommand': 'insert',
            '_executeUpdateCommand': 'update',
            '_executeRemoveCommand': 'remove'
        };

        internalCommands.forEach(function(internalCommand) {
            localBaseHook.wrapAsyncMethod(obj.Db.prototype, internalCommand, null, function(obj, args, locals) {
                var commandName = commandMap[internalCommand] || internalCommand;
                var command = (args && args.length > 0) ? args[0] : undefined;
                var commandArgs;
                //var exitCall;
                //var time;
                locals.transaction = tracer.getTransaction();
                if (locals.transaction) locals.parentSegment = locals.transaction.getActiveSegment();
                if(command && command.db) {
                    var serverPool = self.getServerPool(command.db);
                    var category;

                    if (serverPool.length) {
                        commandArgs = {
                            databaseName: command.db.databaseName,
                            collectionName: command.collectionName,
                            query: command.query ? JSON.stringify(command.query) : '',
                            queryOptions: command.queryOptions,
                            numberToSkip: command.numberToSkip,
                            numberToReturn: command.numberToReturn
                        };

                        if(command.db.auths && command.db.auths.length > 0) {
                            commandArgs.auth = command.db.auths[0];
                        }

                        if(internalCommand === '_executeQueryCommand') {
                            category = "read";
                        }
                        else {
                            category = "write";
                        }
                        if(command.query)
                        {
                            if(command.query.hasOwnProperty('delete'))
                            {
                                category = "write";
                                commandName = "remove";
                            }
                            if(command.query.hasOwnProperty('insert'))
                            {
                                category = "write";
                                commandName = "insert";
                            }
                            if(command.query.hasOwnProperty('update'))
                            {
                                category = "write";
                                commandName = "update";
                            }

                        }
                        locals.exitCallInfo = self.makeExitCallInfo(serverPool, commandName, commandArgs);
                    }
                }
            });
        });
    }
};

