'use strict';
var BackendType =  require('../common/constant').BackendType;
var util = require('util');
function MysqlHook(agent) {
    this.agent = agent;
    this.baseHook = agent.baseHook;
    this.packages = ['mysql'];
}
exports.MysqlHook = MysqlHook;

function hookStream(hookHd, query) {
    hookHd.wrapSyncMethod(query, 'stream', null, function (obj, args, stream) {
        hookHd.wrapSyncMethod(stream, 'pipe');
    });
}

function makeExitCallInfo(config, command) {
    return {
        'backendType': BackendType.BACKEND_TYPE_MYSQL,
        'name':!config ? 'MySQL://unknown:unknown/unknown' : util.format('MySQL://%s:%s/%s', config.host, config.port, config.database),
        'connType': 'MySQL',
        'dbVendor': 'MySQL',
        'dbVersion': (!!config && !!config.version) ? config.version : 'unknown',
        'dbSchema': !config ? 'unknown' : config.database,
        'is_async': false,
        'operation': command,
        'host': !config ? 'unknown' : config.host,
        'port': !config ? 'unknown' : config.port,
        'url':!config ? 'MySQL://unknown@unknown:unknown/unknown' : util.format('MySQL://%s@%s:%s/%s',config.user,config.host,config.port,config.database)
    };
}

MysqlHook.prototype.addConnectExitCall = function(config, error, profiler, txn) {
    var exitCallInfo = makeExitCallInfo(config, 'connect');
    var exitCall = profiler.createExitCall(exitCallInfo, txn);
    if (!exitCall) {
        this.agent.logger.debug('In mysql addConnectExitCall, create null exit call!');
        return;
    }
    profiler.addExitCall(error, exitCall, txn);
};

function prepareExitCall(config, client, wrap, profiler, baseHook) {
    if(client === undefined)
    {
        return;
    }
    if( client.__bonrAttach__ ) return;
    client.__bonrAttach__ = true;
    var orig = client['query'];
    baseHook.wrapAsyncMethod(client, 'connection.query',null, function(obj, args, locals) {
        var command;
        var txn = obj.hasOwnProperty('_txn') ? obj._txn : profiler.agent.tracer.getTransaction();
        if (!txn) return;
        locals.transaction = txn;

        var version;
        if (obj !== undefined && obj._protocol &&
            obj._protocol._handshakeInitializationPacket &&
            obj._protocol._handshakeInitializationPacket.serverVersion
        )
        {
            version = obj._protocol._handshakeInitializationPacket.serverVersion;
        } else {
            version = '';
        }
        if (!config.version) config.version = version;

        command = args.length > 0 ? (args[0].hasOwnProperty('sql') ? args[0].sql : args[0]) : undefined;

        var exitCallInfo = makeExitCallInfo(config, command);
        locals.exitCall = profiler.createExitCall(exitCallInfo, txn);
        locals.parentSegment = txn.getActiveSegment();
        var explain_args = [];
        for(var i = 0;i < args.length; i++) {
            explain_args[i] = args[i];
        }
        explain_args[0] = 'explain ' + args[0];
        var Client = obj;

        function processQueryCb(obj, args) {
            if (!locals.exitCall) {
                return;
            }
            var error = wrap.getErrorObject(args);
            if(error) error.name='mysql';

            if(!error &&!!locals.exitCall && locals.exitCall.needExplainPlan(profiler.agent)) {
                var callback = function (err, results, fields){
                    if(!err && results && fields) {
                        locals.exitCall.collectMySqlExplainPlan(fields, results);
                    }
                    profiler.addExitCall(error, locals.exitCall, locals.transaction);
                };
                explain_args[explain_args.length - 1] = callback;
                orig.apply(Client, explain_args);
            } else {
                profiler.addExitCall(error, locals.exitCall, locals.transaction);
            }
        }

        if (args[0] && args[0].__bonreeCbFn) {
            wrap.before(args[0], '_callback', processQueryCb);
        } else if(typeof args[args.length - 1] === 'function') {
            locals.wrapBefore = processQueryCb;
        } else {
            processQueryCb(obj, args);
        }
    }, function(obj, args, query) {
        hookStream(baseHook, query);
    });
}


MysqlHook.prototype.attach = function(obj) {
    var self = this;
    var wrap = self.agent.wrap;
    var profiler = self.agent.profiler;
    var cmds = ['createClient', 'createConnection', 'createPool'];
    var localBaseHook = self.baseHook;
    var tracer = self.agent.tracer;
    if(obj.__bonrAttached__) return;
    obj.__bonrAttached__ = true;

    localBaseHook.wrapSyncMethod(obj, "mysql.createPoolCluster");
    cmds.forEach(function(createCmd) {
        localBaseHook.wrapSyncMethod(obj, 'mysql.' + createCmd,null, function(obj, args, ret, locals) {
            if (createCmd === 'createPool') {
                localBaseHook.wrapAsyncMethod(ret, 'pool.getConnection', null,function(obj, args, locals) {
                    var txn = obj.hasOwnProperty('_txn') ? obj._txn : tracer.getTransaction();
                    //if (!txn) txn = tracer.startTransactionWithoutBe();
                    if (!txn) return;
                    locals.parentSegment = txn.getActiveSegment();
                    locals.wrapBefore = function(obj, args) {
                        //调用失败,args[0]返回错误
                        if (args[0]) {
                            self.addConnectExitCall(null, args[0], profiler, txn);
                            self.agent.logger.debug('getConnection error, url:', txn.url);
                        }
                        //调用成功,args[1]返回connection
                        else if (args[1]) {
                            var connection = args[1],
                                config = connection.config;
                            self.addConnectExitCall(config, null, profiler, txn);
                            connection._txn = txn;
                            //对connection.query进行hook
                            prepareExitCall(config, connection, wrap, profiler, localBaseHook);
                        }
                    };
                },null);
                localBaseHook.wrapAsyncMethod(ret, 'pool.query', null, null, function (obj, args, query) {
                    query.__bonreeCbFn = true;
                    hookStream(localBaseHook, query);
                });
            } else {
                var client = ret;
                var config = (createCmd === 'createClient' ? client : client.config);
                if (!config) {
                    return;
                }
                localBaseHook.wrapAsyncMethod(client, 'connection.connect',null, function(obj, args, locals) {
                    locals.transaction = obj.hasOwnProperty('_txn') ? obj._txn : tracer.getTransaction();
                    locals.exitCallInfo = makeExitCallInfo(config, 'connect');
                });
                prepareExitCall(config, client, wrap, profiler, localBaseHook);
            }
        }, function(obj, args, locals) {
            var txn = obj.hasOwnProperty('_txn') ? obj._txn : tracer.getTransaction();
            if(!txn) {
                return;
            }
            self.addConnectExitCall(null, locals.error, profiler, txn);
        });
    });
};
