'use strict';

var util = require('util');
var EventEmitter = require('events').EventEmitter;
var urlparser = require('url');
var HTTPCODETYPE =  require('../common/constant').HttpCodeType;

var CreateException = require('../models/error').CreateException;
var CreateHttpError = require('../models/error').CreateHttpError;

function capitalizeFirstLetter (string) {
    return string.charAt(0).toUpperCase() + string.slice(1);
};

function Transaction(agent, id) {
    //Transaction basic info
    this.bt = undefined;
    this.name = undefined;
    this.uri = undefined;
    this.btType = 150;

    this.agent = agent;
    this.id = id;
    //http info
    this.url = undefined;
    this.requestHeaders = undefined;
    this.responseHeaders = undefined;

    this.statusCode = undefined;
    this.httpMethod = undefined;
    this.httpparameters = undefined;

    this.startedExitCalls = undefined;
    this.exitCallCounter = 0;
    this.exitCalls = undefined;

    this.errors = [];
    this.bException = false;
    this.bError = false;
    this.isResponseSent = false;
    this.isFinished = false;
    this.needTrace = false;

    this.baseSegment = undefined;
    this.lastSegment = undefined;
    this.segmentNum = 0;
    this.exitCallCnt = 0;
    if (this.agent.config) {
        this.maxSegmentNum = this.agent.config.config.maxSegmentNum;
    } else {
        this.maxSegmentNum = 3000;
    }
    this.exitCallMap = {};
    this.component = undefined;
}

Transaction.prototype.getActiveSegment = function () {
    var segment = this.lastSegment;
    if (!segment) return null;
    do {
        if (segment.isAlived && segment.className) return segment;
        segment = segment.parent;
    } while (!!segment);

    return null;
};

Transaction.prototype.clear = function () {
    if (!!this.baseSegment) {
        var segments = [this.baseSegment];
        while (segments.length > 0) {
            var segment = segments.shift();
            for (var i = 0; i < segment.childs.length; ++i) {
                segments.push(segment.childs[i]);
            }
            if (segment.childs.length > 0) {
                segment.childs = [];
            }
            if (segment.childIds.length > 0) {
                segment.childIds = [];
            }
            segment.traceNode = null;
            if (!!segment.parent) {
                segment.parent = null;
            }
            if (!!segment.transaction) {
                segment.transaction = null;
            }
        }
    }
    this.baseSegment = null;
    this.lastSegment = null;
    if (!!this.startedExitCalls && Array.isArray(this.startedExitCalls)) {
        if (this.startedExitCalls.length > 0) {
            this.startedExitCalls = [];
        }
    }

    if (!!this.exitCalls && Array.isArray(this.exitCalls)) {
        if (this.exitCalls.length > 0) {
            this.exitCalls = [];
        }
    }

    if (!!this.errors && Array.isArray(this.errors)) {
        if (this.errors.length > 0) {
            this.errors = [];
        }
    }
    this.segmentNum = 0;
    this.exitCallMap = {};
};

Transaction.prototype.__ignoreException = function (excpname) {
    var self = this;
    var err_name = self.agent.reporter.Str({value:'', length:0});
    err_name.value = excpname;
    var isIgnore = self.agent.reporter.so.ignore_exception(self.agent.reporter.error_cfg.ref(),err_name.ref());
    if(isIgnore){
        return true;
    }else{
        return false;
    }
};

Transaction.prototype.__ignoreError = function (errname) {
    var self = this;
    var err_name = self.agent.reporter.Str({value:'', length:0});
    err_name.value = errname;
    var isIgnore = self.agent.reporter.so.ignore_err_msg(self.agent.reporter.error_cfg.ref(),err_name.ref());
    if(isIgnore){
        return true;
    }else{
        return false;
    }
};

Transaction.prototype.addException = function (err) {
    if (!err) return;
    var self = this;
    var reporter = self.agent.reporter;
    var e = CreateException(err, self);
    if(!self.__ignoreException(e.errSummary))
    {
        self.bException = true;
        self.sdkexcpname = reporter.Str({value:'', length:0});
        self.sdkexcpsummary = reporter.Str({value:'', length:0});
        self.sdkexcpdetail = reporter.Str({value:'', length:0});
        self.sdkexcperror_type_str = reporter.Str({value:'', length:0});

        self.sdkexcpname.value = capitalizeFirstLetter(e.errName) + 'Exception';
        self.sdkexcpsummary.value = e.errSummary;
        self.sdkexcpdetail.value =  e.errorDetails;
        self.sdkexcperror_type_str.value = e.errorTypeStr;

        var btErr = reporter.BtError();
        btErr.time_stamp = e.timeStamp;
        btErr.error_type = e.errorType;
        btErr.error_name = self.sdkexcpname;
        btErr.error_type_str = self.sdkexcperror_type_str;
        btErr.error_sumary = self.sdkexcpsummary;
        btErr.error_details = self.sdkexcpdetail;
        var ret = reporter.so.add_bt_error(self.bt.ref(), btErr.ref());
        if(ret != 0) {
            self.agent.logger.logger.error('failed to so.add_bt_error.');
        }

        var seg = self.getActiveSegment();
        if(seg) {
            seg.setError(e.errSummary, e.errorDetails);
        }
    }

};

Transaction.prototype.addHttpError = function (statusCode) {
    var self = this;
    var e = CreateHttpError(statusCode, self);

    var reporter = self.agent.reporter;
    var erro_code = statusCode;
    var erro_desc = self.agent.reporter.Str({value:'', length:0});
    var erro_type = self.agent.reporter.Int;

    var ret = reporter.so.get_custom_http_code_desc(self.agent.config.config.errorConfig.hErrCfg.ref(),erro_code,erro_desc.ref(),erro_type);
    self.sdkerrname = reporter.Str({value:'', length:0});
    self.sdkerrsummary = reporter.Str({value:'', length:0});
    self.sdkerrdetail = reporter.Str({value:'', length:0});
    self.sdkerrerror_type_str = reporter.Str({value:'', length:0});

    if (ret == 0) {
        var errTypeCode= erro_type.readUInt8();
        if(errTypeCode == HTTPCODETYPE.HTTP_CODE_TYPE_ERROR) {
            self.sdkerrname.value = 'HTTP ERROR (' + erro_desc.value +':'+statusCode +')';
            self.sdkerrsummary.value = self.sdkerrname.value;
            self.sdkerrdetail.value = self.sdkerrname.value;
        } else {
            return;
        }
    } else if (statusCode >= 400) {
        self.sdkerrname.value = 'HTTP ERROR (' + statusCode + ')';
        self.sdkerrsummary.value = self.sdkerrname.value;
        self.sdkerrdetail.value = self.sdkerrname.value;
    } else {
        return;
    }

    self.bError = true;
    self.sdkerrerror_type_str.value = e.errorTypeStr;

    var errInfo = reporter.BtError();
    errInfo.time_stamp = e.timeStamp;
    errInfo.error_type = e.errorType;
    errInfo.error_name = self.sdkerrname;
    errInfo.error_type_str = self.sdkerrerror_type_str;
    errInfo.error_sumary = self.sdkerrsummary;
    errInfo.error_details = self.sdkerrdetail;
    var ret = reporter.so.add_bt_error(self.bt.ref(), errInfo.ref());
    if(ret != 0) {
        self.agent.logger.logger.error('failed to so.set_trace_tag.');
    }


    var seg = this.getActiveSegment();
    if(!!seg) {
        seg.setError(e.errSummary, e.errorDetails);
    }
};

function transactionName(agent, request) {
    var config = agent.config.config;
    var nameconfig = config.appConfig.namingConfigs;
    var urlobj = urlparser.parse(request.url);

    if (!nameconfig || !nameconfig.enable) {
        var tmp = '/' + urlobj.pathname.replace(/^\//, '').split('/').slice(0, 2).join('/');
        return '/' + urlobj.pathname.replace(/^\//, '').split('/').slice(0, 2).join('/');
    }

    var accumulator,segments;
    var filters = nameconfig.properties.split(';').reduce(function (accumulater, filter) {
        var kv = filter.split('=');
        accumulater[kv[0]] = kv[1];
        return accumulater;
    }, {});

    if (filters.hasOwnProperty('uri-length')) {
        switch (filters['uri-length']) {
            case 'full':
                accumulator = urlobj.pathname;
                break;
            case 'first-n-segment':
                segments = parseInt(filters['segment-length']);
                accumulator = '/' + urlobj.pathname.replace(/^\//, '').split('/').slice(0, segments).join('/');
                break;
            case 'last-n-segment':
                segments = parseInt(filters['segment-length']);
                accumulator = '/' + urlobj.pathname.replace(/^\//, '').split('/').slice(0 - segments).join('/');
                break;
        }
    }

    if (filters.hasOwnProperty('ext') && filters.ext === "false") {
        return accumulator;
    }

    if (filters.hasOwnProperty('any-segment')) {
        var cutpath = accumulator.replace(/^\//, '').split('/');
        accumulator = '/' + filters['any-segment'].split(',').filter(function (seg) {
            return parseInt(seg) <= cutpath.length;
        }).map(function (str) {
            var index = parseInt(str);
            return cutpath[index - 1];
        }).join('/');
        if (accumulator === '/' && request.url !== '/') {
            accumulator = 'unknown';
        }

    }
    var kvlist = [];
    if (filters.hasOwnProperty('parameter-key') && urlobj.query) {
        var paramSegments = filters['parameter-key'].split(',');
        var queryobj = urlobj.query.split('&').reduce(function (finalobj, subquery) {
            var kv = subquery.split('=');
            finalobj[kv[0]] = kv[1];
            return finalobj;
        }, {});
        kvlist = kvlist.concat(paramSegments.filter(function (k) {
            return queryobj.hasOwnProperty(k);
        }).map(function (key) {
            return key + '=' + queryobj[key].toString();
        }));
    }
    if (filters.hasOwnProperty('header-key') && request.headers) {
        var headerSegments = filters['header-key'].split(',');
        kvlist = kvlist.concat(headerSegments.filter(function (k) {
            return request.headers && (request.headers.hasOwnProperty(k) || request.headers.hasOwnProperty(k.toLowerCase()));
        }).map(function (key) {
            if (!!request.headers.hasOwnProperty(key))
                return key + '=' + request.headers[key].toString();
            else {
                return key + '=' + request.headers[key.toLowerCase()].toString();
            }
        }));
    }
    if (kvlist.length > 0) {
        accumulator = accumulator + '?' + kvlist.join('&');
    }
    return accumulator;
}

//添加traceTag
Transaction.prototype.setTraceTags = function () {
    var self = this;
    var tags = [];
    if(!self.tracer) {
        return;
    }
    tags.push({
        key: 'http.method', value: self.httpMethod || ''
    }, {
        key: 'http.status_code', value: self.statusCode ? self.statusCode.toString() : ''
    }, {
        key: 'http.url', value: self.url
    }, {
        key: 'component', value: self.component
    });
    if (self.bException || self.bError) {
        tags.push({key: 'error', value: true});
    }

    tags.forEach(function (tag) {
        if(tag.key == undefined || tag.value == undefined) {
            return;
        }
        var key = self.agent.reporter.Str({value:'', length:0});
        var val = self.agent.reporter.Str({value:'', length:0});
        key.value = tag.key;
        key.length = tag.key.length;
        val.value = tag.value.toString();
        val.length = tag.value.toString().length;

        var ret = self.agent.reporter.so.set_trace_tag(self.tracer.ref(), key.ref(), val.ref());
        if(ret != 0) {
            self.agent.logger.logger.error('failed to so.set_trace_tag.');
        }
    });
    return tags;
};


exports.Transaction = Transaction;
exports.transactionName = transactionName;
