'use strict';
var fs = require('fs');
var yamlParser = require('js-yaml');
var loggerfuncs = require('../core/logger').Logger_funcs;
var os=require('os');
var path=require('path');
var TRACEMODE = require('../common/constant').TraceMode;

const {
    AgentType,
    UserExperienceType
} = require('../common/constant');

var yamlConfig = {
    configfile: '',
    log: {
        level: "info",
        dir: undefined,
        logFileCount: 2,
        logLimitInKbytes: 20480
    },
    js_inject_pos: 0, // JS自动嵌码位置 0: 在第一个<script>标签之前；1: </head>标签之前
    enable_datafusion: false,
    enable_sql_explain: false,
    maxSegmentNum: 3000,
    blacklist: []
};
function Configuration(agent) {
    this.agent= agent;
    this.config = {};
    this.configRecved = false;
}

exports.Configuration=Configuration;

function RecursiveCopy(src,dest) {
    for(var key in src)
    {
        if(dest.hasOwnProperty(key))
        {
            if(src[key] instanceof Array) {
                dest[key] = src[key];
            }
            else if(src[key] instanceof Object)
            {
                RecursiveCopy(src[key],dest[key]);
            }
            else
            {
                dest[key] = src[key];
            }
        }
    }
}

Configuration.prototype.readOpts = function (opts) {

    if(opts.hasOwnProperty('configFile'))
    {
        if(opts.configFile.match(/.yml$/))
        {
            var configOpts = yamlParser.safeLoad(fs.readFileSync(opts.configFile,'utf8'));

            RecursiveCopy(configOpts,yamlConfig);

            yamlConfig.configfile = opts.configFile;
        }
    }

    if(opts.hasOwnProperty('enablelogging'))
    {
        yamlConfig.log.enablelogging = opts.enablelogging;
    }
    else
    {
        yamlConfig.log.enablelogging = false;
    }

    if (yamlConfig.enablelogging && opts.hasOwnProperty('log_path'))
    {
        yamlConfig.log.dir = opts.log_path;
    }
    if (yamlConfig.enablelogging && opts.hasOwnProperty('level'))
    {
        yamlConfig.log.level = opts.level;
    }
    if(opts.hasOwnProperty('js_inject_pos')){
        yamlConfig.js_inject_pos = opts.js_inject_pos;
    }
    if(opts.hasOwnProperty('enable_datafusion')){
        yamlConfig.enable_datafusion=opts.enable_datafusion;
    }
    yamlConfig.agentInstallTimestamp = opts.agentInstallTimestamp;
    yamlConfig.agentInstallDir = opts.agentInstallDir;
    yamlConfig.agentVersion = opts.agentVersion;
    return opts;
};

Configuration.prototype.init= function(opts) {
    var self = this;
    try{
        //加载配置文件配置
        self.readOpts(opts);
    }catch (err)
    {
        self.agent.logger.error('Failed to load config file' + err);
        process.exit();
    }

    self.config = {
        configTime: 0,
        enable: false,
        dir: yamlConfig.log.dir,
        logFileCount:yamlConfig.log.logFileCount,
        logLimitInKbytes:yamlConfig.log.logLimitInKbytes,
        enablelogging: yamlConfig.log.enablelogging,
        loglevel: yamlConfig.log.level,
	maxSegmentNum:yamlConfig.maxSegmentNum,
        js_inject_pos:yamlConfig.js_inject_pos,
        enable_datafusion:yamlConfig.enable_datafusion,
        enable_sql_explain:yamlConfig.enable_sql_explain,
        blacklist:yamlConfig.blacklist,
        agentID: undefined,
        agentIDS : {
            accountGUID: undefined,
            accountID: 0,
            agentID: 0,
            agentType:AgentType.AGENT_NODEJS,
        },
        containerInfo: {
            agentName: undefined,
            agentVersion:yamlConfig.agentVersion,
            agentInstallTimestamp:yamlConfig.agentInstallTimestamp,
            agentInstallDir:yamlConfig.agentInstallDir,
            processId: process.pid ,
            listens: self.getIp(),
            containerVersion: process.version,
            containerInstallDir: process.execPath,
            VMInfo: 'node '+process.version,
            VMStartupOptions: process.argv.join(' '),
            VMSystemOptions: Object.keys(process.env).map(function (k) {
                return k+=+process.env[k];
            }).join('\r\n'),
        },
        nodeInfo: {
            uniqueHost: undefined,
            OSInfo: undefined,
            environmentVariables: undefined,
        },
        appConfig:{
            dataAggregateEnable: false,
            namingConfigs: undefined,
        },
        backendConfig:{
            enableCaptureNosqlParams : false,
            nosqlMaxParamLength : 0,
            sqlMaxRecords:0,
            sqlExeTimeExplainThreshold: 0,
            remoteCallGetURISegmentType : undefined,
            remoteCallURISegments : 0,
            enableSqlExplain: false
        },
        errorConfig: {
            hErrCfg : undefined
        },
        btConfig:{
            btFilterCfg: undefined,
        },
        injectJsConfig:{
            injectJsEnable:false,
            jsMd5:undefined,
            injectJs:undefined
        },
        traceConfig:{
            enable:false,
            defaultGathererPercent:10,
            enableCollectHttpParams : false,
            traceMode:TRACEMODE.TRACE_MODE_SAMPLE,
            enableCollectHttpReqHeaders : false,
            smartModeLevel : 2
        }
    };

    self.agent.on('updateConfig',function (configResp) {
        if(!self.configRecved)
        {
            self.configRecved = true;
        }
        for( var section in self.config)
        {
            if(configResp.hasOwnProperty(section)&& !!configResp[section] && !(configResp[section] instanceof Object))
            {
                self.config[section] = configResp[section];
                continue;
            }
            for( var key in self.config[section])
            {
                if(configResp.hasOwnProperty(section)&& !!configResp[section] && configResp[section].hasOwnProperty(key))
                {
                    self.update(section,key,configResp[section][key]);
                }
            }
        }
        loggerfuncs.metric_debug(self.agent.logger,"New configration recved "+  JSON.stringify(self.config,null,4));
    });
};

Configuration.prototype.update = function(section,key,value) {
    var self = this;
    try {
        self.config[section][key] = value;
    }catch (err)
    {
        loggerfuncs.metric_error(self.agent.logger,'section: '+ section +'; key: '+ key +'; value: '+value);
        throw err;
    }
};

Configuration.prototype.isEnabled = function () {
    var self = this;
    return self.config.enable;
};
Configuration.prototype.jsInjectEnabled = function () {
    var self = this;
    return  self.config.injectJsConfig.injectJsEnable
        && !!self.config.injectJsConfig.injectJs
        && self.config.injectJsConfig.injectJs!="";
}

Configuration.prototype.dataAggregateEnabled = function () {
    var self = this;
    return self.config.enable_datafusion
        && self.config.appConfig.dataAggregateEnable;
}

Configuration.prototype.getIp=function(){
    var IPv4,hostName;
    hostName=os.hostname();
    var arr=os.networkInterfaces();
    for (let devName in arr) {
        let iface = arr[devName];
        for (let i = 0; i < iface.length; i++) {
            let alias = iface[i];
            if (alias.family === 'IPv4' && alias.address !== '127.0.0.1' && !alias.internal) {
                var localIPAddress = alias.address;
                return [{
                    ip:localIPAddress
                }];
            }
        }
    }
};

exports.ExitCallUeType = function (agent,exitcall) {
    if (!agent || !exitcall) {
        return undefined;
    }
    var config = agent.config.config;

    if (exitcall.have_error) {
        return UserExperienceType.USER_EXPERIENCE_TYPE_ERROR;
    }
    else if (exitcall.duration_time > config.appConfig.remoteCallExeTimeThreshold) {
        return UserExperienceType.USER_EXPERIENCE_TYPE_SLOW;
    }
    else {
        return UserExperienceType.USER_EXPERIENCE_TYPE_GOOD;
    }
};


Configuration.prototype.updateConfig = function () {
    var self = this;
    var configOpts = self.agent.reporter.getConfigResp();

    if(configOpts){
        self.agent.emit('updateConfig',configOpts);
    }
    self.agent.logger.info('configuration received :'+JSON.stringify(self.config,null,2));
};
