'use strict';

var Time = require('../core/time').Time;
var Transaction = require('../models/transaction').Transaction;
var transactionName = require('../models/transaction').transactionName;
var ExitCall = require('../models/exitcall').ExitCall;
var urlparser = require('url');
var util = require('util');
var BTTYPE =  require('../common/constant').BTType;
var AGENTTYPE =  require('../common/constant').AgentType;

var LEN =  require('../common/constant').LEN;
var ERRORCODE = require('../common/constant').ErrCode;
var TRACEMODE =  require('../common/constant').TraceMode;
var HEADERKEY =  require('../common/constant').HeaderKey;
var ERRORTYPE =  require('../common/constant').ErrorType;



function Profiler(agent) {
    this.agent = agent;
    this.transactions = {};
    this.sampleBtMap = new Map();
}

exports.Profiler = Profiler;

Profiler.prototype.grepExitInfo = function(transaction) {
    var infos = [];
    if (transaction && transaction.startedExitCalls) {
        for (var i = 0, l = transaction.startedExitCalls.length; i < l; i++) {
            infos.push({agentGUID:transaction.startedExitCalls[i].agentGUID,host:transaction.startedExitCalls[i].host,
                port:transaction.startedExitCalls[i].port,name:transaction.startedExitCalls[i].name,operation:transaction.startedExitCalls[i].operation});
        }
    }
    return infos;
};

Profiler.prototype.init = function () {
    var self = this;
    // cleanup timeout transactions
    self.agent.timers.setInterval(function() {
        // expire transactions older than 5 minutes, which have not ended
        var now = Date.now();
        for(var txnId in self.transactions) {
            // if(self.transactions[txnId].ts + 300000 < now) {
             if(self.transactions[txnId].ts + 60000 < now) {

            self.agent.logger.debug('timeout transaction  exitCalls:'+
                    JSON.stringify(self.grepExitInfo(self.transactions[txnId])) + (self.transactions[txnId].isResponseSent ? ' response had sent' :
                        'response had not send!'));
                var txn = self.transactions[txnId];
                if (!!txn && !!txn.startedExitCalls) {
                    for (var i = 0,l = txn.startedExitCalls.length; i < l; i++) {
                        if (!txn.exitCallMap[txn.startedExitCalls[i].id]) {
                            self.addExitCall(null, txn.startedExitCalls[i], txn);
                        }
                    }
                }
                self._endTransaction(self.transactions[txnId]);
                self.transactions[txnId].clear();
                delete self.transactions[txnId];
            }
        }
        self.sampleBtMap.clear();
    }, 5000);
};

Profiler.prototype.time = function(isTransaction) {
    var t =  new Time(this.agent, isTransaction);
    t.start();

    return t;
};

Profiler.prototype.needTrace = function (name) {
    var self = this;
    var config = self.agent.config.config;
    if(!config.traceConfig.enable) {
        return false;
    }
    if(config.traceConfig.traceMode === TRACEMODE.TRACE_MODE_SAMPLE) {
        var defaultGatherPersent = config.traceConfig.defaultGathererPercent;
        var seed = Math.floor(Math.random()*1000)+1;
        if (seed <= defaultGatherPersent) {
            return true;
        }
    } else if(config.traceConfig.traceMode === TRACEMODE.TRACE_MODE_SMART) {
        if(!self.sampleBtMap.has(name)) {
            self.sampleBtMap.set(name, 1);
            return true;
        }
        var cnt = self.sampleBtMap.get(name);
        if(!!cnt && cnt < config.traceConfig.smartModeLevel) {
            self.sampleBtMap.set(name, ++cnt);
            return true;
        }
    }
    return false;
};

Profiler.prototype.createBGTaskTranaction = function (trans_name,btType) {
    //设置业务名称
    var self = this;
    var reporter = self.agent.reporter;
    var trans = reporter.Instance();
    var transaction;
    var config = self.agent.config.config;
    if(!config.enable) {
        return null;
    }
    var ret = reporter.so.open_transaction(trans.ref());
    if (ret !== 0) {
        self.agent.logger.error( 'failed to so.open_transaction.');
        return null;
    } else {
        transaction = new Transaction(self.agent, self.agent.tracer.nextTransactionId)
        transaction.instance = trans
        transaction.entryType = btType
        transaction.name = trans_name

        if(self.needTrace(transaction.name)) {
            transaction.needTrace = true;
        }

        //判断是否在业务黑名单中
        if(self.isInBlacklist(transaction.name)){
            return null;
        }

        self.transactions[transaction.id] = transaction
        var bt = reporter.Instance()
        var bt_options = reporter.BtOptions()
        bt_options.name.value = reporter.ref.NULL
        bt_options.name.length = 0
        bt_options.uri.value = reporter.ref.NULL
        bt_options.uri.length = 0
        bt_options.internal_name.value = reporter.ref.NULL
        bt_options.internal_name.length = 0

        self.sdkname = reporter.ref.allocCString(transaction.name)
        self.sdkuri = reporter.ref.allocCString(transaction.name)
        self.sdkinternal_name = reporter.ref.allocCString(transaction.name)

        bt_options.type = BTTYPE.BT_BACKGROUND_TASK ;
        bt_options.name.value = self.sdkname;
        bt_options.uri.value = self.sdkuri;
        bt_options.internal_name.value = self.sdkinternal_name;
        bt_options.start_time_stamp = 0;

        ret = reporter.so.get_or_create_bt(trans.ref(), bt_options.ref(), reporter.ref.NULL, bt.ref());
        if (ret !== 0) {
            self.agent.logger.error('failed to so.get_or_create_bt.');
            return null;
        }
        transaction.bt = bt;
        var cross_req = reporter.Instance();

        var cross_req_options = reporter.CrossReqOptions();
        cross_req_options.cross_req.data[0]=0;

        cross_req_options.is_sampling = transaction.needTrace? 1: 0;
        ret = reporter.so.get_or_create_cross_req(trans.ref(), cross_req_options.ref(), cross_req.ref());
        if (ret !== 0) {
            self.agent.logger.error('failed to so.get_or_create_cross_req.');
            return null;
        }
    }
    return transaction;
}
Profiler.prototype.createTransaction = function(req, btType, moduleName){
    var self = this;
    var reporter = self.agent.reporter;
    var trans = reporter.Instance();
    var transaction;
    var config = self.agent.config.config;
    if(!self.agent.config.config.enable) {
        return ERRORCODE.ERROR_AGENT_STATUS_DISABLED;
    }
    var ret = reporter.so.open_transaction(trans.ref());
    if (ret !== 0) {
        self.agent.logger.error( 'failed to so.open_transaction.');
        return ERRORCODE.ERROR_CALL_SDK_FAILURE;
    } else {
        transaction = new Transaction(self.agent, self.agent.tracer.nextTransactionId);
        transaction.instance = trans;
        transaction.entryType = btType;
        if(config.traceConfig.enableCollectHttpReqHeaders) {
            transaction.requestHeaders = req.headers;
        }
        transaction.ts = Date.now();
        transaction.uri = req.url;
        transaction.url = util.format('%s://%s%s', moduleName,req.headers.host,req.url);
        transaction.httpMethod = req.method;
        if(config.traceConfig.enableCollectHttpParams){
            transaction.httpparameters = urlparser.parse(req.url).query;
        }
        transaction.name = transactionName(self.agent,req);

        //判断是否在业务黑名单中
        if(self.isInBlacklist(transaction.name)){
            return ERRORCODE.ERROR_BT_IN_BLACKLIST;
        }
        if(self.isSampling(req) || self.needTrace(transaction.name)) {
            transaction.needTrace = true;
        }



        self.transactions[transaction.id] = transaction;

        var bt = reporter.Instance();
        var bt_options = reporter.BtOptions();
        self.sdkname = reporter.Str({value:'', length:0});
        self.sdkuri = reporter.Str({value:'', length:0});
        self.sdkinternal_name = reporter.Str({value:'', length:0});

        self.sdkname.value = transaction.name;
        self.sdkuri.value = transaction.uri;
        self.sdkinternal_name.value = transaction.name;

        bt_options.type = BTTYPE.BT_NODEJS_WEB ;
        bt_options.name = self.sdkname;
        bt_options.uri = self.sdkuri;
        bt_options.internal_name = self.sdkinternal_name;
        bt_options.start_time_stamp = 0;


        var http_req_data = reporter.HttpReqData();

        self.sdkurl = reporter.Str({value:'', length:0});
        self.sdkhttp_method = reporter.Str({value:'', length:0});
        self.sdkparameters = reporter.Str({value:'', length:0});
        self.sdkreq_headers = reporter.Str({value:'', length:0});
        self.sdkreq_content =  reporter.Str({value:'', length:0});

        if(!!transaction.url) {
            self.sdkurl.value = transaction.url;
        }
        if(!!transaction.httpMethod) {
            self.sdkhttp_method.value = transaction.httpMethod;
        }
        if(!!transaction.requestHeaders) {
            self.sdkreq_headers.value = (Object.keys(transaction.requestHeaders).map(
                function (key) {
                    return key + ': ' + transaction.requestHeaders[key];
                })).join('\r\n');
        }
        if(!!transaction.httpparameters) {
            self.sdkparameters.value = transaction.httpparameters;
        }
        http_req_data.url = self.sdkurl;
        http_req_data.http_method = self.sdkhttp_method;
        http_req_data.req_headers = self.sdkreq_headers;
        http_req_data.parameters = self.sdkparameters;
        http_req_data.req_content = self.sdkreq_content;
        http_req_data.start_time_stamp = 0;


        ret = reporter.so.get_or_create_bt(trans.ref(), bt_options.ref(),http_req_data.ref(), bt.ref());
        if (ret !== 0) {
            self.agent.logger.error('failed to so.get_or_create_bt.');
            return ERRORCODE.ERROR_CALL_SDK_FAILURE;
        }
        transaction.bt = bt;
        var cross_req = reporter.Instance();

        var cross_req_options = reporter.CrossReqOptions();
        cross_req_options.cross_req.data[0]=0;

        cross_req_options.is_sampling = transaction.needTrace? 1: 0;
        if (req.headers.hasOwnProperty(HEADERKEY.CXREQ))
        {
            var reqValue = req.headers[HEADERKEY.CXREQ];
            for(var i=0; i<reqValue.length && i<LEN.CROSS_HEAD; i++) {
                cross_req_options.cross_req.data[i] = reqValue[i];
            }
        }
        ret = reporter.so.get_or_create_cross_req(trans.ref(), cross_req_options.ref(), cross_req.ref());
        if (ret !== 0) {
            self.agent.logger.error('failed to so.get_or_create_cross_req.');
            return ERRORCODE.ERROR_CALL_SDK_FAILURE;
        }
    }
    return transaction;
};

Profiler.prototype.isSampling = function (req) {
    if(req.headers.hasOwnProperty(HEADERKEY.CXREQ)) {
        var cxOptions = req.headers[HEADERKEY.CXREQ].split(';;');
        for(var i = 0; i < cxOptions.length; i++) {
            var keyval = cxOptions[i].split('=');
            if(!keyval || keyval.length != 2) return;
            if(keyval[0]=='a' && keyval[1]==1) {
                return true;
            }            
        }
    }
    return false;
};

Profiler.prototype.asyncEndTransaction = function(exitCallTime, transaction) {
    var self = this;

    if(transaction.isFinished) {
        self.agent.logger.error('transaction with request id:'+transaction.request_guid+' had finished!');
        // ignore finished transactions
        return;
    }

    if(transaction.isResponseSent) {
        self.agent.logger.error('transaction with request id:'+transaction.request_guid+' had send!');
        // ignore "synchronous" transactions, which are waiting for their exit calls
        //Sometimes, when an error occurs, the upper service fails to call to HTTP end, which requires an asynchronous end of a transaction
        return;
    }
    transaction.isResponseSent = true;
    if(transaction.startedExitCalls) {

        if(!transaction.exitCalls || transaction.startedExitCalls.length !== transaction.exitCalls.length) {
            self.agent.logger.error('transaction with request id:'+transaction.request_guid+
                ' have mismatched backend, startedExitCalls:'+JSON.stringify(self.grepExitInfo(transaction)));
            // waiting for exit calls to end
            // exit calls will call endTransaction
            return;
        }
    }
    exitCallTime.is_async = true;
    self._endTransaction(exitCallTime, transaction);
};

Profiler.prototype.endBGTaskTransaction= function (transaction) {
    var self = this;
    var ret = -1;
    if(transaction.isFinished) {
        return;
    }
    var reporter = this.agent.reporter;
    var bt_finish_options = reporter.BtFinishOptions();
    if(transaction.bException) {
        bt_finish_options.flags = ERRORTYPE.ERROR_TYPE_EXCEPTION;
    }
    bt_finish_options.finish_time_stamp = 0;
    ret = reporter.so.finish_bt(transaction.bt.ref(), bt_finish_options.ref(), reporter.ref.NULL);
    if (ret != 0) {
        self.agent.logger.error('failed to so.finish_bt.');
    }
    if(transaction.tracer) {
        transaction.setTraceTags();
        ret = reporter.so.finish_tracer(transaction.tracer.ref());
        if (ret != 0) {
            self.agent.logger.error('failed to so.finish_tracer.');
        }
    }
    var ptran = transaction.instance.ref();
    transaction.instance = null;
    transaction.isFinished = true;
    ret = reporter.so.close_transaction(ptran);
    if (ret != 0) {
        self.agent.logger.error('failed to so.close_transaction.');
    }
}

Profiler.prototype._endTransaction = function(transaction) {
    var self = this;
    var ret = -1;
    if(transaction.isFinished) {
        return;
    }
    var reporter = this.agent.reporter;
    var bt_finish_options = reporter.BtFinishOptions();
    if (transaction.bError) {
        bt_finish_options.flags = ERRORTYPE.ERROR_TYPE_HTTP;
    }
    if(transaction.bException) {
        bt_finish_options.flags = ERRORTYPE.ERROR_TYPE_EXCEPTION;
    }
    bt_finish_options.finish_time_stamp = 0;

    var response_data  = reporter.HttpRespData();
    self.sdkstatus_code = reporter.Str({value:'', length:0});
    self.sdkres_header = reporter.Str({value:'', length:0});
    self.sdkresp_content = reporter.Str({value:'', length:0});
    response_data.status_code = self.sdkstatus_code;
    response_data.res_header = self.sdkres_header;
    response_data.resp_content = self.sdkresp_content;

    if (!!transaction.statusCode) {
        response_data.status_code.value = transaction.statusCode.toString();
    }

    if (!!transaction.responseHeaders) {
        response_data.res_header.value = (Object.keys(transaction.responseHeaders).map(
            function (key) {
                return key + ': ' + transaction.responseHeaders[key];
            })).join('\r\n');
    }

    ret = reporter.so.finish_bt(transaction.bt.ref(), bt_finish_options.ref(), response_data.ref());
    if (ret != 0) {
        self.agent.logger.error('failed to so.finish_bt.');
    }
    if(transaction.tracer) {
        transaction.setTraceTags();
        ret = reporter.so.finish_tracer(transaction.tracer.ref());
        if (ret != 0) {
            self.agent.logger.error('failed to so.finish_tracer.');
        }
    }

    var ptran = transaction.instance.ref();
    transaction.instance = null;
    transaction.isFinished = true;
    ret = reporter.so.close_transaction(ptran);
    if (ret != 0) {
        self.agent.logger.error('failed to so.close_transaction.');
    }
};

Profiler.prototype.endTransaction = function (transaction) {
    var self = this;

    //self.agent.logger.debug('endTransaction, transaction id:' + transaction.id + ' segment num:' + transaction.segmentNum);
    transaction.segmentNum = 0;

    if(transaction.isResponseSent) {
        self.agent.logger.error('transaction id:'+transaction.id+' had send!');
        return;
    }
    transaction.isResponseSent = true;

    if (!!transaction.baseSegment && !transaction.baseSegment.updateAllTime()) {
            self.agent.logger.error('updateAllTime failed!');
    }

    if(transaction.startedExitCalls) {
        if(!transaction.exitCalls || transaction.startedExitCalls.length !== transaction.exitCalls.length) {
            // waiting for exit calls to end
            // exit calls will call endTransaction
            self.agent.logger.debug('transaction have mismatched backend, startedExitCalls:'+JSON.stringify(self.grepExitInfo(transaction)));
            return;
        }
    }

    self._endTransaction(transaction);
    var id = transaction.id;
    transaction.clear();
    delete self.transactions[id];

};

Profiler.prototype.__getNextSequenceInfo = function(transaction) {
    var self = this;
    transaction.exitCallCounter++;
    return transaction.exitCallCounter.toString();
};

Profiler.prototype.createExitCall = function (exitCallInfo, txn) {
    var self = this;
    var exitCall;

    var transaction = !!txn ? txn : this.agent.tracer.getTransaction();
    if (!transaction) {
        return;
    }

    if (transaction && exitCallInfo) {
        if (exitCallInfo.operation && exitCallInfo.operation.length === 0) return undefined;
        exitCall = new ExitCall(exitCallInfo,transaction);

        if (!transaction.startedExitCalls) {
            transaction.startedExitCalls = [];
        }
        transaction.startedExitCalls.push(exitCall);
    }
    else {
        return undefined;
    }

    return exitCall;
};

Profiler.prototype.addExitCall = function (error, exitCall, txn,toEnd) {
    var self = this;
    if(!exitCall || exitCall.isFinished) {
        return;
    }
    var transaction = !!txn ? txn : this.agent.tracer.getTransaction();
    if (!transaction ||transaction.isFinished) {
        return;
    }
    var seg =transaction.getActiveSegment();
    if(!!seg) {
        seg.linkBeCall(exitCall);
    }


    exitCall.stop({
        flags : error? 1:0,
        errorName : error? error.name || error.code : undefined,
        errorDetail : error ? error.stack.toString() : undefined,
    });
    exitCall.have_error = !!error;
    exitCall.isFinished = true;


    if (!transaction.exitCalls) {
        transaction.exitCalls = [];
    }
    if(error)//后端的错误已经设置了，业务错误还需要设置吗，看下如何统计的
    {
        exitCall.have_error = true;
        transaction.addException(error);
    }
    transaction.exitCalls.push(exitCall);
    transaction.exitCallMap[exitCall.id] = true;
};

Profiler.prototype.isInBlacklist = function (btName) {
    var self = this;
    var reporter = self.agent.reporter;
    if(!self.agent.config || !self.agent.config.config) {
        return false;
    }
    var config = self.agent.config.config;

    var ymlBlacklist = config.blacklist;
    if(ymlBlacklist instanceof Array) {
        for (var i=0; i < ymlBlacklist.length; i++) {
            if (btName.startsWith(ymlBlacklist[i]) || btName.startsWith('/' + ymlBlacklist[i])) {
                return true;
            }
        }
    }

    if(!!config && !!config.btConfig && !!config.btConfig.btFilterCfg) {
        var btFilterCfg = config.btConfig.btFilterCfg;
        var name = reporter.Str('',0);
        name.value = btName;
        var ret = reporter.so.in_bt_blacklist(btFilterCfg.ref(), name.ref());
        if(ret == 1) {
            return true;
        }
    }
    return false;
};

