# Copyright (c) Bonree, Inc., and its affiliates
# 2015
# All Rights Reserved

"""The agent itself.

"""
import os
import time

from Bonree import get_version, config
from Bonree.agent.services.config import ConfigService
from Bonree.agent.core import timer, registries, naming
from Bonree.agent.services.cross import make_cross_server
from Bonree.agent.services.trace import make_traceshot_server
from Bonree.logger import BT_LOGGER, SYS_LOGGER
from Bonree.lib import set_current_ident_btr, unset_current_ident_btr, set_time_unit
from Bonree.agent.models.transactions import make_transaction_req_factory
from Bonree.agent.services.transport import TransportSDK


class Agent(object):
    """The entry point for the Python agent.

    """
    def __init__(self):

        self._txr_factory = None

        # Threads
        self.config_svc = None


        # Registries
        self.backend_reg = None
        self.naming_reg= None

        self.enable = False

        self.agent_id = 0
        self.agent_guid = None

        # register
        self.reg_trans = {}
        self.reg_backend = {}

        # Program startup arguments
        self.startup_arguments = os.environ.get('BONR_STARTUP_ARGUMENTS', '')

        # transport
        self.transport_sdk = None
        self.transport_be = None

    def start(self):
        if self.config_svc is None:
            self.config_svc = ConfigService(agent=self)
            # service set here
            self._txr_factory = make_transaction_req_factory(self, self.timer)
            self.enable = True
            SYS_LOGGER.info('Starting Bonree Python agent (build %s) ...' % get_version())
            


    def timer(self, *args, **kwargs):
        t = timer.Timer(self.time, *args, **kwargs)
        return t

    @staticmethod
    def time():
        return set_time_unit(time.time())
    
    @property
    def enabled(self):
        """Return true if the agent has started and is enabled.

        """
        return self.enable

    @property
    def backend_registry(self):
        if not self.backend_reg:
            self.backend_reg = registries.BackendRegistry()
        return self.backend_reg

    @property
    def naming_registry(self):
        if not self.naming_reg:
            self.naming_reg = naming.NameRegistry()
        return self.naming_reg

    def update_config_response(self):
        self.config_svc.update_config_response()


    def check_and_open_sdk(self):
        return self.config_svc.check_and_open_sdk()

    def start_transaction_request(self,  entry_type, request=None, name=None):
        """
        Start a transaction request processing on request arriving or background task on name.

        Parameters
        ----------
        entry_type: Bonree.agent.models.transactions define some entry_types the agent supports
        request: Bonree.lib.LazyWsgiRequest, the coming transaction request, only http for now
        name: the transaction request name, This is for backend task.

        Returns
        -------
        a Bonree.agent.models.transactions.TransactionReq object
        """
        if not self.check_and_open_sdk():
            return None
        self.update_config_response()
        if not config.get_enabled():
            return None

        # AppConfig
        # config.updateNamingConfigs(self)
        if request is not None:
            name = self.naming_registry.get_name(entry_type, request)
        elif not name:
            return None

        if config.getFilterURIPatterns(name):
            BT_LOGGER.debug('{0} in blacklist.'.format(name))
            return None

        btr = self._start_btr(name, entry_type)

        transport_sdk = TransportSDK()
        transport_sdk.start_bt_transaction(name, entry_type, request)
        btr.transport_sdk = transport_sdk

        btr.request = request
        btr.timer.start()
        btr.cross_svc = make_cross_server(btr, request)
        btr.traceshot_svc = make_traceshot_server(btr.cross_svc.cross_header, btr.transport_sdk, name)

        set_current_ident_btr(btr)
        BT_LOGGER.debug('start btr:{0}'.format(btr))

        return btr

    def _start_btr(self, name, entry_type):
        """Create the TransactionReq according to TransactionInfo and incoming request

        Parameters:
        ---------------
        name: The name of business transaction request.
        entry_type: The type of business transaction request.
        trace_header: The trace request head.

        Returns:
        ---------------
        a TransactionReq object
        """
        return self._txr_factory(name=name, entry_type=entry_type)


    def end_transaction_request(self, btr):
        """End the active transaction on this thread.

        Parameters
        ----------
        btr : Bonree.agent.models.transactions.TransactionReq

        A specific BT request to end.
        """
        if btr and hasattr(btr, 'transport_sdk'):

            btr.timer.stop()
            btr.transport_sdk.end_sdk_transaction(btr)
            unset_current_ident_btr()
            BT_LOGGER.debug('finish btr:{0}'.format(btr))


    def handle_exception(self, btr, exc_type, exc_value, exc_tb):
        btr.add_exception(exc_type, exc_value, exc_tb)
