from Bonree import config
from Bonree.logger import SYS_LOGGER
from Bonree.lang import urlparse, str
import re


class NameRegistry(object):
    URL_HEAD = 0
    URL_TAIL = 1
    URL_FULL = 2

    def __init__(self):
        self.registered_names = {}
        self.name_config = {}

    def get_confusion_url(self, url):
        if len(url) == 0:
            return url
        confusion_url = re.split('\\d+', url)
        return '*'.join(confusion_url)

    def get_name(self, btr_type, request):
        request_url = request.path
        confusion_url = self.get_confusion_url(request_url)
        key = (btr_type, confusion_url)
        try:
            return self.registered_names[key]
        except KeyError:
            name = self.make_btr_name(btr_type, request, request_url)
            name = self.get_confusion_url(name)
            self.registered_names[key] = name
            return name

    def update(self, rules):
        new = self.make_name_config(rules)
        if new != self.name_config:
            self.name_config = new
            self.registered_names.clear()

    def clear(self, rules):
        free = self.make_name_config(rules)
        if free == self.name_config:
            self.name_config.clear()
            self.registered_names.clear()

    def make_name_config(self, rules):
        """Since currently we just support BT_PYTHON_WSGI, So just a signal dict is enough.
        And we just ignore 'enable'. Otherwise there will no rule for bt naming.
        """
        new_config = {}
        for rule in rules:
            properties = rule['properties']
            couples = properties.split(';')

            def trans_url_length_rule(string):
                mapper = {
                    'full': self.URL_FULL,
                    'first-n-segment': self.URL_HEAD,
                }
                return mapper.setdefault(string, self.URL_TAIL)

            handlers = {
                "uri-length": trans_url_length_rule,
                "segment-length": int,
                "any-segment": lambda l: [int(x) for x in l.split(',')],
                "parameter-key": lambda l: l.split(','),
                "header-key": lambda l: l.split(','),
                "ext": config.on_off,
            }

            temp_name = {}

            for couple in couples:
                if not couple:
                    continue

                key, value = couple.split("=")
                handler = handlers.get(key)
                if handler:
                    value = handler(value)
                temp_name[key] = value

            if not temp_name.get('ext', False):
                temp_name.pop("any-segment", None)
                temp_name.pop("parameter-key", None)
                temp_name.pop("header-key", None)

            new_config[int(rule['type'])] = temp_name

        return new_config

    def make_btr_name(self, btr_type, request, request_url):
        """Get name from http request.

        """
        try:
            ruler = self.name_config[btr_type]
        except KeyError:
            SYS_LOGGER.warning("BTR[{0}] don't have name config.".format(btr_type))
            return request_url

        try:
            direction = ruler['uri-length']
            if direction == self.URL_FULL:
                url = request_url.split('?')[0] # full url without parameter
                return url

            parse = urlparse(request_url)
            cut_path = parse.path.strip('/').split('/')
            segments = ruler['segment-length']
            if direction == self.URL_HEAD:
                url = '/' + '/'.join(cut_path[:segments])
            else:
                url = '/' + '/'.join(cut_path[-segments:])
        except KeyError:
            SYS_LOGGER.warning("BTR[{0}] have damaged name config:{1}".format(btr_type, ruler))
            return request_url

        if not ruler.get('ext', False):
            return url

        if "any-segment" in ruler:
            new_cut = url.strip('/').split('/')

            tmp = []
            for i in ruler['any-segment']:
                try:
                    tmp.append(new_cut[i-1])
                except IndexError:
                    pass

            if tmp:
                url = '/' + '/'.join(tmp)
            else:
                url = "unknown"

        tail = []
        if "parameter-key" in ruler and parse.query:
            parameters = parse.query.split('&')
            for key in ruler['parameter-key']:
                for param in parameters:
                    if key+'=' in param:
                        tail.append(param)

        if "header-key" in ruler:
            for key in ruler['header-key']:
                key = key.lower()
                value = request.headers.get(key, None)
                if value:
                    tail.append(key+'='+str(value))

        if tail:
            url += "?" + "&".join(tail)

        return url
