# Copyright (c) Bonree, Inc., and its affiliates
# 2015
# All Rights Reserved

"""Registries store information about the agent's state and configuration.

"""
import threading

from Bonree import rule
from Bonree.agent.proto import pb_type
from Bonree.agent.services.wrapper import Be
from Bonree.lang import items
from Bonree.agent.models import methods


def backend_hash(tipe, properties):
    """Return a unique identifier for this backend.

    Parameters
    ----------
    tipe : enum
        The type of the exit call being made.
    properties : dict
        A dict representing the backend properties.

    """
    return (tipe, tuple(sorted(items(properties))))


def backend_key(beinfo):
    return backend_hash(beinfo.tipe, beinfo.identifying_properties)


class BackendRegistry(object):
    """Track backends that have been registered.

    """
    __slots__ = ('registered_backends', 'lock')

    URL_PROPERTY_MAX_LEN = 100
    HOST_PROPERTY_MAX_LEN = 50
    DB_NAME_PROPERTY_MAX_LEN = 50
    VENDOR_PROPERTY_MAX_LEN = 50
    db_backend_naming_format_string = rule.common_name(issql=True)

    property_to_config_field_name = {
        'HOST': 'host',
        'PORT': 'port',
        'URL': 'url',
        'DATABASE': 'database',
        'VERSION': 'version',
        'VENDOR': 'vendor',
        'SERVER POOL': 'serverPool',
        'QUERY STRING': 'queryString',
        'ROUTING KEY': 'routingKey',
        'EXCHANGE': 'exchange',
    }

    backend_type_from_vendor = {
        'mysql': pb_type.BACKEND_TYPE_MYSQL,
        'memcached': pb_type.BACKEND_TYPE_MEMCACHE,
        'postgresql': pb_type.BACKEND_TYPE_POSTGRESQL,
        'oracle': pb_type.BACKEND_TYPE_ORACLE,
        'hive': pb_type.BACKEND_TYPE_HIVE,
    }

    def __init__(self):
        self.lock = threading.Lock()
        self.registered_backends = {}

    def get_backend(self, tipe, properties, format_string):
        """Get a backend from its type and properties.

        After a config update, every unique backend has identifying properties
        re-generated based on the naming config.  If there is a registered
        backend which matches the generated identifying properties, it is
        returned.  Otherwise a new backend is created.

        Parameters
        ----------
        tipe : exitcalls type.
            The type of the exit call being made. More detail see pb_backend.proto
        properties : dict
            A dict of strings representing all the supported properties for a
            backend of the given exit call type.
        format_string : str
            A string defining the format of the backend name.  This is a
            simplified version of the 'Format String Syntax' as described
            in PEP-3101.  See exitcalls.BackendNamer for more details.

        Returns
        -------
        backend : exitcalls.Backend
            The backend. If no backend has been added to the registry of the
            specified type with the given identifying properties, a new
            backend is created and added to the registry.

        See Also
        --------
        get_db_backend

        """
        back_key = backend_hash(tipe, properties)
        try:
            backend = self.registered_backends[back_key]
        except KeyError:
            identifying_properties = properties
            display_name = methods.BackendNamer().format(format_string, identifying_properties)
            backend = methods.Backend(tipe, properties, display_name)

            with self.lock:
                self.registered_backends[back_key] = backend

        return backend

    def get_db_backend(self, vendor, host, port, dbname, dbversion):
        """Get a (SQL) database backend.

        This method is a wrapper that just invokes `get_backend` with the
        correct type, identifying properties, and name.

        Parameters
        ----------
        vendor : str
            The name of the DB server vendor, such as 'MySQL'.
        host : str
            The host the outgoing connection is being made to.
        port : int
            The port the outgoing connection is being made to.
        dbname : str
            The name of the database we are connecting to. Do not confuse this
            with the `vendor` parameter. If you are using MySQL and connecting
            to the 'ecommerce' database, the `vendor` would be 'MySQL' and
            the `dbname` would be 'ecommerce'.
        dbversion: The version of database

        Returns
        -------
        backend : exitcalls.Backend
            The HTTP backend for the given host, port, and URL.

        """
        key = (vendor, host, port, dbname, dbversion)
        try:
            backend = self.registered_backends[key]
        except KeyError:
            vendor = rule.vendor_conversion(vendor[:self.VENDOR_PROPERTY_MAX_LEN])
            host = host[:self.HOST_PROPERTY_MAX_LEN]
            dbname = dbname[:self.DB_NAME_PROPERTY_MAX_LEN]

            backend_properties = {
                'VENDOR': vendor,
                'HOST': host,
                'PORT': str(port),
                'DATABASE': dbname,
                'VERSION': dbversion or 'unknown',
                'URL': rule.url_sql_conversion(vendor, host, port, dbname),
                'dbVendor': vendor,
                'dbVersion': dbversion or 'unknown',
                'dbSchema': dbname,
            }
            # it seems that 'is' keyword may have different result with '=='
            # when comparing unicode strings (from code imported unicode_literal)

            tipe = self.backend_type_from_vendor[vendor.lower()]
            display_name = methods.BackendNamer().format(self.db_backend_naming_format_string, backend_properties)
            backend = methods.Backend(tipe, backend_properties, display_name)
            backend.be = self.create_be(backend, rule.connType_conversion(vendor))
            # backend = self.get_backend(tipe, backend_properties, self.db_backend_naming_format_string)
            self.registered_backends[key] = backend
        return backend

    def create_be(self, backend, vendor):
        be = Be()
        be.type = int(backend.tipe)
        be.conn_type.with_value(vendor)
        be.name.with_value(backend.name)
        be.host.with_value(backend.identifying_properties.get('HOST', 'unknown'))
        be.port = int(backend.identifying_properties.get('PORT', '0'))
        be.url.with_value(backend.identifying_properties.get('URL', backend.name))

        if 'dbVendor' in backend.identifying_properties:
            be.db_vendor.with_value(backend.identifying_properties['dbVendor'])
        if 'dbVersion' in backend.identifying_properties:
            be.db_version.with_value(backend.identifying_properties['dbVersion'])
        if 'dbSchema' in backend.identifying_properties:
            be.db_schema.with_value(backend.identifying_properties['dbSchema'])
        return be


