# Copyright (c) Bonree, Inc., and its affiliates
# 2015
# All Rights Reserved

"""Interceptors hook into imports to modify third-party module behavior.

"""

import sys

from Bonree.agent.interceptor.frameworks import intercept_apistar_request
from Bonree.lang import reload
from Bonree.logger import SYS_LOGGER

try:
    from importlib.machinery import ModuleSpec, BuiltinImporter
except ImportError:
    BuiltinImporter = None

# These definitions must come before the subsequent imports to avoid a
# circular reference.
HOST_PROPERTY_MAX_LEN = 50
DB_NAME_PROPERTY_MAX_LEN = 50
VENDOR_PROPERTY_MAX_LEN = 50

from . import cache, frameworks, http, mongodb, sql, concurrent, background, log, custommethod


BT_INTERCEPTORS = (
    # concurrency
    ('concurrent.futures', concurrent.intercept_futures),
    ('threading', concurrent.intercept_threading),
    ('thread', concurrent.intercept_thread),
    # ('greenlet', concurrent.intercept_greenlet),

    # Entry points
    # 1. Flask
    ('flask', frameworks.intercept_flask),
    ('flask.wrappers', frameworks.intercept_flask_response),
    ('flask.ctx', frameworks.intercept_flask_request),
    # 2. Django
    ('django.http.response', frameworks.intercept_django_response),
    ('django.template.response', frameworks.intercept_django_render),
    ('django.core.handlers.wsgi', frameworks.intercept_django_wsgi_handler),
    ('django.core.handlers.asgi', frameworks.intercept_django_asgi_handler),
    ('django.core.handlers.base', frameworks.intercept_django_base_handler),
    ('django.core.handlers.exception', frameworks.intercept_uncaught_exception),
    ('django.core.management.commands.runserver', frameworks.intercept_django_runserver),
    # 3. Tornado, pip install tornado
    ('tornado.web', frameworks.intercept_tornado_web),
    # 4. bottle, pip install bottle
    ('bottle', frameworks.intercept_bottle),
    # 5. CherryPy, pip install CherryPy
    ('cherrypy', frameworks.intercept_cherrypy),
    ('cherrypy.process.wspbus', frameworks.intercept_cherrypy_wspbus),
    ('cherrypy._cprequest', frameworks.intercept_cherrypy_request),
    # 6.1 Pyramid, pip install pyramid
    ('pyramid.router', frameworks.intercept_pyramid),
    ('pyramid.response', frameworks.intercept_pyramid_response),
    # 6.2 wsgiref, pip install wsgiref
    ('wsgiref.simple_server', frameworks.intercept_wsgiref),
    # 7. web.py, pip install web.py
    ('web.httpserver', frameworks.interceptor_web_static),
    ('web.application', frameworks.interceptor_web),
    ('web.wsgiserver', frameworks.interceptor_start_response),
    # 8. sanic, pip3.6 install sanic
    ('sanic', frameworks.intercept_sanic),
    ('sanic.server', frameworks.intercept_sanic_response),
    ('sanic.response', frameworks.intercept_sanic_response_send),
    # 9.1 Odoo(v10.0 v11.0)
    ('odoo.service.wsgi_server', frameworks.intercept_odoo_wsgi_handler),
    ('odoo.http', frameworks.intercept_odoo_request_handler),
    # 9.2 Odoo(v8.0 v9.0 v7.0)
    ('openerp.service.wsgi_server', frameworks.intercept_odoo_wsgi_handler),
    # 9.3 odoo(v6.1)
    ('openerp.wsgi.core', frameworks.intercept_odoo6_wsgi_handler),
    # 10. APIStar, pip3.6 install apistar
    ('apistar', frameworks.intercept_APIStar),
    ('apistar.server.injector', intercept_apistar_request),
    ('apistar.server.adapters', frameworks.intercept_APIStar_ASGItoWSGIAdapter),
    #11. Flask/Odoo depends on werkzeug
    ('werkzeug.serving', frameworks.intercept_werkzeug_serving),
    #12. gunicorn
    ('gunicorn.app.base', frameworks.intercept_gunicorn),
    #13 quart require python3.6+
    ('quart', frameworks.intercept_quart),
    ('quart.wrappers', frameworks.intercept_quart_response),
    # ('quart.ctx', frameworks.intercept_quart_request),
    #14 uvicorn
    ('uvicorn.main', frameworks.intercept_uvicorn),
    #15starlette
    ('starlette.applications', frameworks.intercept_starlette),
    ('starlette.responses', frameworks.intercept_starlette_response),
    # 16 daphne
    ('daphne.server', frameworks.intercept_daphne),

    # HTTP exit calls
    ('httplib', http.intercept_httplib),
    ('http.client', http.intercept_http_client),
    ('urllib3', http.intercept_urllib3),
    ('tornado.httpclient', http.intercept_tornado_httpclient),
    ('requests', http.intercept_requests),
    ('kombu.messaging', http.intercept_rabbitmq),
    ('thrift.transport.TSocket', background.intercept_thrift_server_listen),
    # ('thrift.protocol.TProtocol', http.interceptor_thrift),
    ('thrift.protocol.TBinaryProtocol', http.interceptor_thrift_binary_field),
    ('thrift.protocol.TJSONProtocol', http.interceptor_thrift_json_field),
    ('thrift.protocol.TCompactProtocol', http.interceptor_thrift_compact_field),
    ('thrift.protocol.TProtocol', http.interceptor_thrift_skip),

    ('elasticsearch', http.intercept_elasticsearch),

    # RPC
    ('grpc', http.intercept_grpc_channel),
    ('grpc._channel', http.intercept_grpc),
    ('grpc._server', http.intercept_grpc_server),
    ('xmlrpc.server', http.intercept_xmlrpc_server),
    ('SimpleXMLRPCServer', http.intercept_xmlrpc_server), # python2
    ('stomp.protocol', http.intercept_activemq),
    ('stomp.connect', http.intercept_listen_activemq),
    # stompy - Python STOMP client library
    ('stompy.stomp', http.intercept_stompy),
    ('stompy.frame', http.intercept_stompy_frame),
    # kafka-python
    ('kafka', http.intercept_kafka_python),
    ('kafka.consumer.group', http.intercept_kafka_consumer),

    # SQL exit calls
    # pip install psycopg2
    ('psycopg2', sql.psycopg2.intercept_psycopg2_connection),
    # pip install PyMySQL
    ('pymysql.connections', sql.pymysql.intercept_pymysql_connections),
    # pip install mysql-connector-python-rf
    ('mysql.connector.connection', sql.mysql_connector.intercept_mysql_connector_connection),
    # pip install MySQL-python
    ('MySQLdb.connections', sql.mysqldb.intercept_MySQLdb_connection),
    # pip install cx-Oracle
    ('cx_Oracle', sql.cx_oracle.intercept_cx_oracle_connection),

    # pip install pyhive sasl thrift thrift-sasl
    ('pyhive.hive', sql.pyhive.intercept_pyhive_connections),
    # thrift
    # ('TCLIService.TCLIService', background.intercept_thrift_client),

    # Caches
    ('memcache', cache.intercept_memcache),
    ('redis.client', cache.intercept_redis_client),
    ('pyssdb', cache.intercept_ssdb),

    # MongoDB
    ('pymongo', mongodb.intercept_pymongo),

    # Background
    ('celery.app.task', background.intercept_celery),
    ('celery.result', background.intercept_result),
    ('invoke.runners', background.intercept_invoke),
    ('invoke.executor', background.intercept_invoke_execute),

    # Logging
    ('logging', log.intercept_logging),
    ('logging', log.intercept_logging_append_trace_id),
    ('logbook.handlers', log.intercept_logbook),
    ('logbook.base', log.intercept_logbook_append_trace_id),
    ('eliot', log.intercept_eliot),

    # aiohttp
    ('aiohttp.client', http.intercept_aiohttp_request),

    # openstack
    ('nova.conductor.rpcapi', frameworks.intercept_nova_compute_api),
    ('nova.conductor.rpcapi', frameworks.intercept_nova_conductor_api),
    ('nova.scheduler.rpcapi', frameworks.intercept_nova_scheduler_api),
    ('nova.conductor.manager', frameworks.intercept_nova_conductor_manage),
    ('nova.compute.manager', frameworks.intercept_nova_compute_manage),
    ('nova.scheduler.manager.', frameworks.intercept_nova_schedule_manage),
    ('webob.response', frameworks.intercept_webob_response),
    ('webob.client', frameworks.intercept_webob_client),
    ('webob.dec', frameworks.intercept_webob_dec_wsgify),
    ('oslo_messaging.rpc.client', http.intercept_oslo_message_rpc_client),
    ('oslo_messaging._drivers.impl_rabbit', http.intercept_oslo_message_rabbit),
)


def add_hook(agent):
    """Add the module interceptor hook for Bonree, if it's not already registered.

    """
    interceptor = ModuleInterceptor(agent)

    for mod, patch in BT_INTERCEPTORS:
        interceptor.call_on_import(mod, patch)

    sys.meta_path.insert(0, interceptor)
    return interceptor


class ModuleInterceptor(object):
    """Intercepts finding and loading modules in order to monkey patch them on load.

    """

    def __init__(self, agent):
        super(ModuleInterceptor, self).__init__()
        self.agent = agent
        self.imported_modules = set()
        self.module_hooks = {}
        self.intercepted_modules = set()


    def call_on_import(self, module_name, cb):

        if module_name in sys.modules:
            self._intercept_module(sys.modules[module_name], [cb])
        else:
            self.module_hooks.setdefault(module_name, [])
            self.module_hooks[module_name].append(cb)


    def my_find_module(self, module_name, path, target=None):
        is_hook_module = (module_name in self.module_hooks)
        path_is_list = isinstance(path, list)
        # print('my_find_module module name[%s] path[%s] is_hook_module[%s] path_is_list[%s]' % (module_name, path, is_hook_module, path_is_list))
        if module_name in self.imported_modules:
            return False
        if (not is_hook_module):
            if not path_is_list:
                return False
            if len(path) == 0:
                return False
            if not isinstance(path[0], str):
                return False
            # not checkout unhooked libs because ...
            # TODO: ImportError: No module named django.contrib.messages.models
            if path[0].find('site-packages') >= 0:
                return False
            if path[0].find('/lib64/python') >= 0:
                return False
            if path[0].find('/lib/python') >= 0:
                return False
            # TODO: django 3.1.1 raise Exception
            # KeyError: 'asyncio.selectors'
            if module_name.find('asyncio') >= 0:
                return False
            # TODO: /usr/local/python3/lib/python3.7/importlib
            if module_name.find('importlib') >= 0:
                return False
            if module_name.find('odoo.addons.base_import') >= 0:
                return False
            # require builtin `Importer` can import it.
            if BuiltinImporter:
                if not self.other_loader_can_import(module_name, path, target):
                    return False
        self.imported_modules.add(module_name)
        is_python_2 = sys.version_info[0] <= 2
        if not module_name in sys.modules and is_python_2:
            try:
                # load module here and not in function `load_module`
                # when load module fail, not using our ModuleInterceptor
                # __import__('a.b.c') returns <module a>, not <module a.b.c>
                # print('1 __import__(%s) start' % (module_name,))
                __import__(module_name)
                # print('1 __import__(%s) start' % (module_name,))
            except Exception:
                # print('import[%s] path[%s] fail, return False' % (module_name, path))
                return False

        return True

    def other_loader_can_import(self, module_name, path, target):
        for loader in sys.meta_path:
            if loader is self:
                continue
            if hasattr(loader, 'find_spec') and loader.find_spec(module_name, path, target) is not None:
                return True
        return False

    def find_spec(self, module_name, path, target=None):
        if self.my_find_module(module_name, path, target):
            return ModuleSpec(module_name, self)
        return None

    def find_module(self, module_name, path=None):
        if self.my_find_module(module_name, path, None):
            return self
        return None

    def load_module(self, module_name):
        module = None
        if not module_name in sys.modules:
            # __import__('a.b.c') returns <module a>, not <module a.b.c>
            # in Python 3.8.5 __import__ module 'therad' will throw Exception
            # do not catch this Exception, just let it throw
            #print('__import__(%s) start' % (module_name,))
            __import__(module_name)
            #print('__import__(%s) end' % (module_name,))
        if not module_name in sys.modules:
            return None
        module = sys.modules[module_name]  # ...so get <module a.b.c> from sys.modules
        if module is None:
            # print('module_name[%s] is None' % (module_name))
            return None
        hasIface = hasattr(module, "Iface")
        SYS_LOGGER.debug('module [%-40s] hasIface[%s]' % (module_name, hasIface))
        if hasIface:
            self.module_hooks.setdefault(module_name, [])
            self.module_hooks[module_name].append(background.intercept_thrift_server)
            self.module_hooks[module_name].append(background.intercept_thrift_client)

        # Remove the module from the list of hooks so that we never see it again.
        hooks = self.module_hooks.pop(module_name, [])
        self._intercept_module(module, hooks)
        return module

    def _intercept_module(self, module, hooks):
        try:
            if hooks is None:
                return
            for hook in hooks:
                SYS_LOGGER.debug(('hook_module[%-40s] using [%s]' % (module.__name__, hook.__name__, )))
                hook(self.agent, module)
            self.intercepted_modules.add(module)
        except BaseException:
            # SYS_LOGGER.exception('Exception in %s hook.' % module.__name__)

            # Re-import to ensure the module hasn't been partially patched.
            # SYS_LOGGER.exception('Re-importing %s after error in module hook' % module.__name__)
            reload(module)
