# Copyright (c) Bonree, Inc., and its affiliates
# 2020
# All Rights Reserved

"""Definition of base entry and exit point interceptors.
This module require python3.6+
"""
import sys
import asyncio
from functools import wraps
from Bonree.agent.models.methods import MethodData
from Bonree.agent.interceptor.base import BaseInterceptor, OnlyStartInterceptor, OnlyEndInterceptor, CustomBtInterceptor
from Bonree.agent.models.transactions import ENTRY_PYTHON_BACKGROUND


class AsyncBaseInterceptor(BaseInterceptor):
    def _create_method_attach(self, original_method):
        if not asyncio.iscoroutinefunction(original_method):
            return super(AsyncBaseInterceptor, self)._create_method_attach(original_method)

        @wraps(original_method)
        async def method_attach(*args, **kwargs):
            btr = self.btr
            if not btr:
               return await original_method(*args, **kwargs)
            md = MethodData(args, kwargs, self.cls, original_method, btr)
            with self.log_exceptions():
                self._start_attach(md)
            try:
                response = await original_method(*md.args, **md.kwargs)
                md.response = response
            except Exception:
                with self.log_exceptions():
                    md.error = self.add_error(*sys.exc_info())
                raise
            finally:
                with self.log_exceptions():
                    self._end_attach(md)
            return response

        return method_attach

    def transaction_attach(self, method_attach, patched_method_name, original_method, entry):
        if entry:
            patched_method = self.get_patched_method(patched_method_name, original_method.__name__)
            if not asyncio.iscoroutinefunction(patched_method):
                @wraps(original_method)
                def entrance_attach(*args, **kwargs):
                    return patched_method(method_attach, *args, **kwargs)
                return entrance_attach
            else:
                @wraps(original_method)
                async def entrance_attach(*args, **kwargs):
                    return await patched_method(method_attach, *args, **kwargs)

                return entrance_attach
        else:
            return method_attach


class AsyncOnlyStartInterceptor(OnlyStartInterceptor):
    def _create_method_attach(self, original_method):
        if not asyncio.iscoroutinefunction(original_method):
            return super(AsyncOnlyStartInterceptor, self)._create_method_attach(original_method)
        @wraps(original_method)
        async def method_attach(*args, **kwargs):
            md = MethodData(args, kwargs, self.cls, original_method)
            with self.log_exceptions():
                self.start_data(md, *md.args, **md.kwargs)
            return await original_method(*md.args, **md.kwargs)

        return method_attach


class AsyncOnlyEndInterceptor(OnlyEndInterceptor):
    def _create_method_attach(self, original_method):
        if not asyncio.iscoroutinefunction(original_method):
            return super(AsyncOnlyEndInterceptor, self)._create_method_attach(original_method)
        @wraps(original_method)
        async def method_attach(*args, **kwargs):
            md = MethodData(args, kwargs, self.cls, original_method)
            response = await original_method(*md.args, **md.kwargs)
            md.response = response
            with self.log_exceptions():
                self.end_data(md, *md.args, **md.kwargs)
            return response

        return method_attach


class AsyncCustomBtInterceptor(AsyncBaseInterceptor, CustomBtInterceptor):

    def create_wrapper(self, origin_method):
        origin_method_wrapper = self._create_method_attach(origin_method)
        @wraps(origin_method)
        async def wrapper(*args, **kwargs):
            return await self.custom_bt_entry(origin_method_wrapper, *args, **kwargs)
        return wrapper

    async def custom_bt_entry(self, origin_method, *args, **kwargs):
        if callable(self.bt_name):
            bt_name = self.bt_name(*args, **kwargs)
        else:
            bt_name = self.bt_name
        btr = self.start_transaction_request(ENTRY_PYTHON_BACKGROUND, name=bt_name)
        try:
            response = await origin_method(*args, **kwargs)
        except Exception:
            with self.log_exceptions():
                if btr:
                    btr.add_exception(*sys.exc_info())
            raise
        finally:
            self.end_transaction_request(btr)

        return response


class AsyncCustomTraceInterceptor(AsyncBaseInterceptor):

    def __init__(self, agent, cls):
        super(AsyncCustomTraceInterceptor, self).__init__(agent, cls)
        self.btrace_data = True

    def create_wrapper(self, origin_method):
        return self._create_method_attach(origin_method)
