import sys

from Bonree.agent.interceptor.base import ExitCallInterceptor
from Bonree.agent.services.cross import parse_cross_response_header
from Bonree.logger import SYS_LOGGER, BT_LOGGER
from types import MethodType,FunctionType
from Bonree import rule
from Bonree.agent.proto.pb_type import BACKEND_TYPE_THRIFT
from Bonree.agent.models.cross import BR_CS_REQ, BR_CS_RSP


class ThriftClientFunctionInterceptor(ExitCallInterceptor):
    def __init__(self, agent, cls, clientFunctionName):
        super(ThriftClientFunctionInterceptor, self).__init__(agent, cls)
        self.clientFunctionName = clientFunctionName
    def get_host_and_port(self, client):
        host = 'localhost'
        port = '0'
        _oprot = getattr(client, '_iprot')
        if _oprot is None:
            return host, port
        transport = None
        for prop in dir(_oprot):
            pp = getattr(_oprot, prop)
            if pp is None:
                continue
            for base_class in pp.__class__.__bases__:
                if base_class.__name__ == 'TTransportBase':
                    transport = pp
                    break
        if transport is None:
            return host, port
        tsocket = None
        if transport.__class__.__name__ == 'TSocket':
            tsocket = transport
        else:
            for prop in dir(transport):
                pp = getattr(transport, prop)
                if pp is None:
                    continue
                if pp.__class__.__name__ == 'TSocket':
                    tsocket = pp
                    break
        if tsocket is None:
            return host, port
        host = getattr(tsocket, 'host')
        port = getattr(tsocket, 'port')
        if host is None:
            host = 'localhost'
        if port is None:
            port = '0'
        return host, port

    def get_backend(self, host, port, version):
        """

        Parameters
        ----------
        host : rabbitmq server host
        port : rabbirmq server port

        """

        backend_properties = {
            'HOST': host,
            'PORT': str(port),
            'VENDOR': 'Thrift',
            'URL': rule.url_conversion('thrift', host, port),
            'VERSION': version,
        }
        backend_name_format_string = rule.common_name()
        backend = self.agent.backend_registry.get_backend(BACKEND_TYPE_THRIFT, backend_properties, backend_name_format_string)
        backend.connType = 'thrift'
        return backend
    def make_tags(self, md, client, *args, **kwargs):
        with self.log_exceptions():
            tags  = {}
            tags['component'] = 'Thrift'
            host, port = self.get_host_and_port(client)
            # print('host[%s] port[%s]' % (host, port,))
            tags['peer.address'] = rule.peer_address_conversion(address='thrift', ip=host, port=port)
            tags['peer.service'] = self.clientFunctionName
            md.tags = tags

    def start_data(self, md, client, *args, **kwargs):
        btr = self.btr
        if not btr:
            return
        host, port = self.get_host_and_port(client)
        # print('host[%s] port[%s]' % (host, port,))
        md.backend = self.get_backend(host, port, 'version')
        # print('backend[%s]' % (md.backend,))
        if md.backend:
            md.summary = self.clientFunctionName
            md.operation = self.clientFunctionName

        md.is_rpc = True
        md.custom = True
        self.btrace_data = True
        md.response_method = self.parse_thrift_response
        self.start_exit_call(md)
        self.start_trace_call(md)
        btr.rpc_headers = {BR_CS_REQ: md.cross_head}
        self.btrace_data = False

    def end_data(self, md, *args, **kwargs):
        self.btrace_data = True
        self.end_trace_call(md)
        self.end_exit_call(md)

    def parse_thrift_response(self, response):
        btr = self.btr
        if not btr:
            return None
        rpc_headers = btr.rpc_headers
        if BR_CS_RSP in rpc_headers:
            return parse_cross_response_header(rpc_headers)
        return None


def intercept_thrift_client(agent, mod):
    iface = getattr(mod, 'Iface', None)
    client = getattr(mod, 'Client', None)
    # They are required be `class`
    if not (isinstance(iface, type) and isinstance(client, type)):
        return

    clientIsIface = issubclass(client, iface)
    SYS_LOGGER.debug('clientIsIface %s' % (clientIsIface,))
    if not (clientIsIface):
        return

    module_name = mod.__name__
    client_name = client.__name__
    # hook client methods
    for clientFunctionName in dir(iface):
        if clientFunctionName.startswith('__'):
            continue
        clientFunction = getattr(client, clientFunctionName, None)
        isFunctionType = isinstance(clientFunction, FunctionType)
        isMethodType = isinstance(clientFunction, MethodType)
        SYS_LOGGER.debug('clientFunction[%s] isFunctionType[%s] isMethodType[%s]' % (clientFunction, isFunctionType, isMethodType))
        if not (isFunctionType or isMethodType):
            continue
        fullClientFunctionName = '%s.%s.%s' % (module_name, client_name, clientFunction.__name__)
        SYS_LOGGER.debug('hook [%s]' % (fullClientFunctionName))
        ThriftClientFunctionInterceptor(agent, client, fullClientFunctionName).attach(clientFunctionName, btrace_data=True)
