# Copyright (c) Bonree, Inc., and its affiliates
# 2015
# All Rights Reserved

"""Definition of base entry and exit point interceptors.

"""
import re
import sys
import traceback
from contextlib import contextmanager
from functools import wraps

from Bonree import config, rule
from Bonree.agent.models.errors import get_exception_name, get_exception_message
from Bonree.agent.models.methods import MethodData, HOOK_FLAG
from Bonree.agent.models.transactions import ENTRY_PYTHON_BACKGROUND
from Bonree.agent.proto.pb_config import HTTP_CODE_TYPE_SUCESS
from Bonree.agent.services.trace import module_name_to_component_name
from Bonree.agent.services.transport import TransportTrace, TransportBe
from Bonree.logger import SYS_LOGGER, BT_LOGGER, only_log_exception
from Bonree.lib import set_current_ident_btr, unset_current_ident_btr, get_btr_from_attr, now

pattern = r"""^[ \n]*<!DOCTYPE html"""
html_re = re.compile(pattern, re.IGNORECASE)


def simple_wrapper(cls_or_module, method_name, patch_function):
    original_method = getattr(cls_or_module, method_name, None)
    if not original_method:
        return

    if hasattr(original_method, HOOK_FLAG):
        return

    @wraps(original_method)
    def method_attach(*args, **kwargs):
        return patch_function(original_method, *args, **kwargs)

    patched_method = method_attach
    setattr(patched_method, HOOK_FLAG, True)
    setattr(cls_or_module, method_name, patched_method)


class BaseInterceptor(object):
    def __init__(self, agent, cls):
        self.agent = agent
        self.cls = cls
        self.btrace_data = False
        self.component_name = None

    def _fix_dunder_method_name(self, method, class_name):
        # If `method` starts with '__', then it will have been renamed by the lexer to '_SomeClass__some_method'
        # (unless the method name ends with '__').
        if method.startswith('__') and not method.endswith('__'):
            method = '_' + class_name + method
        return method

    def get_patched_method(self, patched_method_name, method_name):
        patched_method_name = patched_method_name or '_' + method_name
        patched_method_name = self._fix_dunder_method_name(patched_method_name, self.__class__.__name__)
        return getattr(self, patched_method_name)

    def get_original_method(self, original_method_name):
        method = self._fix_dunder_method_name(original_method_name, self.cls.__name__)
        original_method = getattr(self.cls, method, None)
        return original_method

    def add_error(self, exc_type, exc_value, exc_tb):
        if exc_type or exc_value:
            error = {}
            error['errorName'] = get_exception_name(exc_value)
            error['errorSumary'] = get_exception_message(exc_value)
            error['errorDetails'] = ''.join(traceback.format_exception(exc_type, exc_value, exc_tb))
            return error

    def transaction_attach(self, method_attach, patched_method_name, original_method, entry):
        if entry:
            patched_method = self.get_patched_method(patched_method_name, original_method.__name__)

            @wraps(original_method)
            def entrance_attach(*args, **kwargs):
                return patched_method(method_attach, *args, **kwargs)
            return entrance_attach
        else:
            return method_attach

    def _attach(self, method, patched_method_name, entry):
        if method is None:
            return
        original_method = self.get_original_method(method)

        if original_method is None:
            SYS_LOGGER.warning("[%s] has not [%s] property" % (self.cls.__name__, method))
            return

        if hasattr(original_method, '_bonr_intercepted'):
            return

        method_attach = self._create_method_attach(original_method)

        hook_method =  self.transaction_attach(method_attach, patched_method_name, original_method, entry)
        if hook_method is None:
            return
        hook_method._bonr_intercepted = True
        setattr(self.cls, method, hook_method)

    def _create_method_attach(self, original_method):
        @wraps(original_method)
        def method_attach(*args, **kwargs):
            btr = self.btr
            if not btr:
                return original_method(*args, **kwargs)
            md = MethodData(args, kwargs, self.cls, original_method, btr)
            with self.log_exceptions():
                self._start_attach(md)
            try:
                response = original_method(*md.args, **md.kwargs)
                md.response = response
            except Exception:
                with self.log_exceptions():
                    md.error = self.add_error(*sys.exc_info())
                raise
            finally:
                with self.log_exceptions():
                    self._end_attach(md)
            return response
        return method_attach

    def _start_attach(self, md):
        # BT_LOGGER.debug("_start_attach: [{0}.{1}]".format(md.cls_name, md.method_name))
        self.start_data(md, *md.args, **md.kwargs)
        self.make_tags(md, *md.args, **md.kwargs)
        if md.custom:
            return
        self.start_exit_call(md)
        self.start_trace_call(md)

    def _end_attach(self, md):
        md.end_time = now()
        # BT_LOGGER.debug("_end_attach:   [{0}.{1}]".format(md.cls_name, md.method_name))
        self.end_data(md, *md.args, **md.kwargs)
        if md.custom:
            return
        self.end_trace_call(md)
        self.end_exit_call(md)

    def start_data(self, md, *args, **kwargs):
        pass

    def end_data(self, md, *args, **kwargs):
        pass

    def start_exit_call(self, md):
        pass

    def end_exit_call(self, md):
        pass

    def make_tags(self, md, *args, **kwargs):
        pass

    def start_trace_call(self, md, btrace_data=False):
        if not (btrace_data or self.btrace_data):
            return
        btr = md.btr
        if not (btr and btr.traceshot_svc):
            return
        transport_trace = TransportTrace()
        transport_trace.start_trace_call(md)
        md.transport_trace = transport_trace

    def end_trace_call(self, md, btrace_data=False):
        if md.end_time is None:
           md.end_time = now()
        transport_trace = md.transport_trace
        if transport_trace is None:
            return

        if self.component_name:
            md.tags['component'] = self.component_name

        transport_trace.end_trace_call(md)

    @property
    def btr(self):
        return get_btr_from_attr()

    def log_exceptions(self):
        return only_log_exception

    def attach(self, method_or_methods, patched_method_name=None, entry=False, btrace_data=False):
        self.btrace_data = btrace_data

        if getattr(self.cls, '__module__', None):
            self.component_name = module_name_to_component_name(self.cls.__module__)
        elif self.cls:
            self.component_name = module_name_to_component_name(self.cls.__name__)

        if not isinstance(method_or_methods, list):
            method_or_methods = [method_or_methods]
        for method in method_or_methods:
            self._attach(method, patched_method_name, entry)


class OnlyStartInterceptor(BaseInterceptor):
    def _create_method_attach(self, original_method):
        @wraps(original_method)
        def method_attach(*args, **kwargs):
            md = MethodData(args, kwargs, self.cls, original_method)
            with self.log_exceptions():
                self.start_data(md, *md.args, **md.kwargs)
            return original_method(*md.args, **md.kwargs)

        return method_attach


class OnlyEndInterceptor(BaseInterceptor):
    def _create_method_attach(self, original_method):
        @wraps(original_method)
        def method_attach(*args, **kwargs):
            md = MethodData(args, kwargs, self.cls, original_method)
            response = original_method(*md.args, **md.kwargs)
            md.response = response
            with self.log_exceptions():
                self.end_data(md, *md.args, **md.kwargs)
            return response

        return method_attach


class ExitCallInterceptor(BaseInterceptor):
    def attach(self, method_or_methods, patched_method_name=None, entry=False, btrace_data=True):
        super(ExitCallInterceptor, self).attach(method_or_methods, btrace_data=btrace_data)

    def start_exit_call(self, md, *args, **kwargs):
        """Start an exit call.

        """
        # if not self.agent.check_and_open_sdk():
        #     return None
        # self.agent.update_config_response()
        # if not config.get_enabled():
        #     return None
        if not md.backend or md.exit_flag:
            return None
        btr = self.btr

        transport_be = TransportBe()
        md.transport_be = transport_be
        if md.is_rpc and btr:
            transport_be.start_be_rpc(md)
        else:
            transport_be.start_be_call(md)
        # with_btr = (btr is None)
        # BT_LOGGER.debug("start exitcall:[{0}.{1}] with btr[{2}] backend[{3}]".format(md.cls_name, md.method_name, with_btr, md.backend))

    def end_exit_call(self, md, *args, **kwargs):
        """End the exit call.

        """
        btr = self.btr
        transport_be = md.transport_be
        if transport_be is None:
            return

        if md.exc_info and btr:
            btr.add_exception(*md.exc_info)
        if md.is_rpc and btr:
            transport_be.end_be_rpc(md)
        else:
            transport_be.end_be_call(md)
        # with_btr = (btr is not None)
        # BT_LOGGER.debug("end   exitcall:[{0}.{1}] with btr[{2}] backend[{3}]".format(md.cls_name, md.method_name, with_btr, md.backend))


class EntryPointInterceptor(BaseInterceptor):
    def attach(self, method_or_methods, patched_method_name=None, entry=True, btrace_data=False):
        super(EntryPointInterceptor, self).attach(method_or_methods, patched_method_name, entry, btrace_data)

    def start_transaction_request(self, entry_type, request=None, name=None):
        if self.btr:
            return None

        with self.log_exceptions():
            return self.agent.start_transaction_request(entry_type, request, name)

    def end_transaction_request(self, btr):
        if btr is None:
            return

        with self.log_exceptions():
            self.agent.end_transaction_request(btr)

    def handle_http_status_code(self, btr, status_code, msg):
        status_code = int(status_code)
        btr.status_code = status_code
        if status_code == 200:
            # for speed faster
            return

        error_msg, error_type = config.getErrorMsg(status_code)
        if error_type and int(error_type) == HTTP_CODE_TYPE_SUCESS:
            return

        if error_msg and error_msg != 'None':
            btr.add_http_error(status_code, error_msg)
            return

        if status_code < 400:
            return

        btr.add_http_error(status_code, msg)

    def make_tags(self, md, *args, **kwargs):
        tags  = {}
        btr = self.btr
        if btr:
            tags['http.method'] = btr.request.method if btr.request is not None else ''
            tags['http.status_code'] = str(btr.status_code) if btr.status_code else '0'
            tags['http.url'] = btr.request.url if btr.request is not None else ''
            btr.tags = tags
        # md.tags = tags


class CustomTraceInterceptor(BaseInterceptor):

    def __init__(self, agent, cls):
        super(CustomTraceInterceptor, self).__init__(agent, cls)
        self.btrace_data = True

    def create_wrapper(self, origin_method):
        return self._create_method_attach(origin_method)


class CustomBtInterceptor(EntryPointInterceptor):

    def __init__(self, agent, cls):
        super(CustomBtInterceptor, self).__init__(agent, cls)
        self.bt_name = ''
        self.btrace_data = True

    def set_bt_name(self, bt_name):
        self.bt_name = bt_name
        return self

    def create_wrapper(self, origin_method):
        origin_method_wrapper = self._create_method_attach(origin_method)
        @wraps(origin_method)
        def wrapper(*args, **kwargs):
            return self.custom_bt_entry(origin_method_wrapper, *args, **kwargs)
        return wrapper

    def custom_bt_entry(self, origin_method, *args, **kwargs):
        if callable(self.bt_name):
            bt_name = self.bt_name(*args, **kwargs)
        else:
            bt_name = self.bt_name
        btr = self.start_transaction_request(ENTRY_PYTHON_BACKGROUND, name=bt_name)
        try:
            response = origin_method(*args, **kwargs)
        except Exception:
            with self.log_exceptions():
                if btr:
                    btr.add_exception(*sys.exc_info())
            raise
        finally:
            self.end_transaction_request(btr)

        return response


class ThreadEntryPointInterceptor(BaseInterceptor):
    def target_wrapper(self, target, btr):
        if not callable(target):
            return target

        @wraps(target)
        def thread_wrapper(*args, **kwargs):
            transport_trace = None
            with self.log_exceptions():
                set_current_ident_btr(btr)
                md = MethodData(args, kwargs, self.cls, target, btr)
                md.start_time = now()
                if btr.traceshot_svc and not getattr(btr, 'close_flag', False):
                    transport_trace = TransportTrace()
                    transport_trace.start_trace_call(md)
            try:
                response = target(*args, **kwargs)
            except Exception:
                with self.log_exceptions():
                    md.error = self.add_error(*sys.exc_info())
                raise
            finally:
                with self.log_exceptions():
                    if md.end_time is None:
                        md.end_time = now()
                    with btr.lock:
                        if btr.traceshot_svc and not getattr(btr, 'close_flag', False) and transport_trace:
                            transport_trace.end_trace_call(md)
                    unset_current_ident_btr()

            return response

        return thread_wrapper
