# Copyright (c) Bonree, Inc., and its affiliates
# 2015
# All Rights Reserved

"""Interceptor for Redis.

"""

from __future__ import unicode_literals

from Bonree import rule, config
from Bonree.lang import str
from Bonree.agent.interceptor.cache import CacheInterceptor
from Bonree.agent.proto.pb_type import BACKEND_TYPE_REDIS


def get_host_port(_redis):
    if 'host' not in _redis.connection_pool.connection_kwargs \
            or 'port' not in _redis.connection_pool.connection_kwargs:
        if not hasattr(_redis.connection_pool, 'nodes'):
            return '', '0', ''

        # redis cluster
        hosts = []
        urls = []
        for node in _redis.connection_pool.nodes.startup_nodes:
            _host, _port = node['host'], str(node['port'])
            _url = rule.url_conversion('redis', _host, _port)
            hosts.append(rule.peer_address_conversion('', _host, _port))
            urls.append(_url)
        host = ','.join(hosts)  # 'ip1:port1,ip2:port2...'
        port = '0'  # real port will in host
        url = ','.join(urls)  # 'redis://ip1:port1,redis://ip2:port2...'
    else:
        host = _redis.connection_pool.connection_kwargs['host']
        port = str(_redis.connection_pool.connection_kwargs['port'])
        url = rule.url_conversion('redis', host, port)
    return host, port, url



class RedisInterceptor(CacheInterceptor):
    def __init__(self, agent, cls):
        super(RedisInterceptor, self).__init__(agent, cls, rule.vendor_conversion('Redis'))

    def get_backend(self, host, port):
        """

        Parameters
        ----------
        host : redis server host
        port : redis server port

        """
        backend_name_format_string = rule.common_name()
        backend_properties = {
            'HOST': host,
            'PORT': str(port),
            'VENDOR': self.vendor,
            'URL': rule.url_conversion('redis', host, port)
        }
        backend = self.agent.backend_registry.get_backend(BACKEND_TYPE_REDIS, backend_properties,
                                                          backend_name_format_string)
        backend.connType = rule.connType_conversion('redis')
        return backend

    def make_tags(self, md, _redis, *args, **kwargs):
        tags  = {}
        host, port, url = get_host_port(_redis)
        tags['peer.address'] = url
        tags['db.type'] = rule.connType_conversion('redis')
        tags['db.statement'] = getattr(_redis, '_bron_redis_statement', md.operation)
        md.tags = tags

class RedisClientInterceptor(RedisInterceptor):
    def start_data(self, md, _redis, *args, **kwargs):
        host, port, url = get_host_port(_redis)
        md.backend = self.get_backend(host, port)
        if md.backend:
            operation = md.original_method.__name__
            if config.getEnableCaptureNosqlParams() and args:
                args_list = [str(i) for i in args]
                operation = operation + ' ' + ' '.join(args_list)
            md.operation = operation
            setattr(_redis, '_bron_redis_method', True)

    def end_data(self, md, _redis, *args, **kwargs):
        if hasattr(_redis, '_bron_redis_method'):
            delattr(_redis, '_bron_redis_method')
        if hasattr(_redis, '_bron_redis_statement'):
            delattr(_redis, '_bron_redis_statement')


class RedisExecuteInterceptor(RedisInterceptor):
    def start_data(self, md, _redis, *args, **kwargs):
        operation = ' '.join(str(x) for x in args).lower()
        if not config.getEnableCaptureNosqlParams():
            operation = operation.split(' ')[0]
        setattr(_redis, '_bron_redis_statement', operation)
        flag = getattr(_redis, '_bron_redis', None)
        if not flag:
            with self.log_exceptions():
                host, port, url = get_host_port(_redis)
                md.backend = self.get_backend(host, port)
                if md.backend:
                    md.operation = operation
        else:
            self.btrace_data = False



def intercept_redis_client(agent, mod):
    """
    module `redis` version < 3.0.0, `redis.Redis` extends `redis.StrictRedis`
    module `redis` version >= 3.0.0, `redis.StrictRedis` = `redis.Redis`
    """

    method1_list = ['exists', 'delete', 'type', 'keys', 'randomkey', 'rename', 'dbsize', 'expire', 'ttl', 'move',
                   'flushdb', 'flushall', 'set', 'get', 'getset', 'mget', 'setnx', 'setrange', 'mset',
                   'msetnx', 'incr', 'decr', 'append', 'substr', 'getrange', 'rpush', 'lpush', 'llen', 'lrange',
                   'ltrim', 'lindex', 'lset', 'lpop', 'rpop', 'blpop', 'brpop', 'rpoplpush', 'sadd', 'srem',
                   'spop', 'smove', 'scard', 'sismember', 'sinter', 'sinterstore', 'sunion', 'sunionstore', 'sdiff',
                   'sdiffstore', 'smembers', 'srandmember', 'zrem', 'zincrby', 'zrank', 'zrevrank', 'zrevrange',
                   'zrangebyscore', 'zcount', 'zcard', 'zremrangebyrank', 'zremrangebyscore', 'hset', 'hsetnx', 'hget',
                   'hmget', 'hmset', 'hincrby', 'hexists', 'hdel', 'hlen', 'hkeys', 'hvals', 'hgetall', 'save']
    method2_list = ['setex', 'lrem', 'zadd']

    if hasattr(mod, 'StrictRedis'):
        for method in method1_list:
            RedisClientInterceptor(agent, mod.StrictRedis).attach(method, btrace_data=True)
        RedisExecuteInterceptor(agent, mod.StrictRedis).attach('execute_command', btrace_data=True)
    else:
        for method in method1_list:
            RedisClientInterceptor(agent, mod.Redis).attach(method, btrace_data=True)
        RedisExecuteInterceptor(agent, mod.Redis).attach('execute_command', btrace_data=True)

    for method in method2_list:
        RedisClientInterceptor(agent, mod.Redis).attach(method, btrace_data=True)
