# Copyright (c) Bonree, Inc., and its affiliates
# 2020
# All Rights Reserved

"""Interceptors and utilities for dealing with ASGI-based apps/frameworks.
this module require python3.6+
"""

import sys
import asyncio
from types import MethodType, FunctionType
from functools import wraps

from Bonree import config
from Bonree.lang import items, str
from Bonree.lib import LazyWsgiRequest
from Bonree.agent.models import transactions
from Bonree.agent.interceptor.base import EntryPointInterceptor
from Bonree.agent.services.cross import make_cross_response_header
from Bonree.agent.interceptor.async_base import AsyncBaseInterceptor


class ASGIInterceptor(AsyncBaseInterceptor, EntryPointInterceptor):
    """
    doc: https://asgi.readthedocs.io/en/latest/index.html
    ASGI applications should be a single async callable:
        coroutine application(scope, receive, send)

    Legacy (v2.0) ASGI applications are defined as a callable:
        application(scope):
            return coroutine application_instance(receive, send)
    """
    def attach(self, method_name, patched_method_name=None, entry=True, btrace_data=True):
        origin_method = self.get_original_method(method_name)
        if origin_method is None:
            return
        if not asyncio.iscoroutinefunction(origin_method):
            return
        super(ASGIInterceptor, self).attach(method_name, patched_method_name='application_callable', entry=entry, btrace_data=btrace_data)

    def build_wsgi_environ(self, scope):
        """
        Builds a scope and request body into a WSGI environ object.
        """
        environ = {
            "REQUEST_METHOD": scope["method"],
            "SCRIPT_NAME": scope.get("root_path", ""),
            "PATH_INFO": scope["path"],
            "QUERY_STRING": scope["query_string"].decode("ascii"),
            "SERVER_PROTOCOL": "HTTP/%s" % scope["http_version"],
            "wsgi.version": (1, 0),
            "wsgi.url_scheme": scope.get("scheme", "http"),
            "wsgi.multithread": True,
            "wsgi.multiprocess": True,
            "wsgi.run_once": False,
        }
        # Get server name and port - required in WSGI, not in ASGI
        if "server" in scope:
            environ["SERVER_NAME"] = scope["server"][0]
            environ["SERVER_PORT"] = str(scope["server"][1])
        else:
            environ["SERVER_NAME"] = "localhost"
            environ["SERVER_PORT"] = "80"

        if "client" in scope:
            environ["REMOTE_ADDR"] = scope["client"][0]

        # Go through headers and make them into environ entries
        for name, value in scope.get("headers", []):
            name = name.decode("latin1")
            if name == "content-length":
                corrected_name = "CONTENT_LENGTH"
            elif name == "content-type":
                corrected_name = "CONTENT_TYPE"
            else:
                corrected_name = "HTTP_%s" % name.upper().replace("-", "_")
            # HTTPbis say only ASCII chars are allowed in headers, but we latin1 just in case
            value = value.decode("latin1")
            # if corrected_name in environ:
            #     value = environ[corrected_name] + "," + value
            environ[corrected_name] = value
        return environ

    async def application_callable(self, application, instance, scope, receive, send):
        if scope['type'] != 'http':
            return await application(instance, scope, receive, send)
        if config.CONTAINER_VERSION is None and 'asgi' in scope:
            container_version = 'asgi'
            if isinstance(scope['asgi'], dict) and 'version' in scope['asgi']:
                container_version = "asgi/%s" % scope['asgi']['version']
            config.set_container_version(container_version)
        environ = self.build_wsgi_environ(scope)
        btr = self.start_transaction_request(transactions.ENTRY_PYTHON_WEB, request=LazyWsgiRequest(environ))
        try:
            response = await application(instance, scope, receive, self._make_start_send_wrapper(send))
        except Exception:
            with self.log_exceptions():
                if btr:
                    btr.add_exception(*sys.exc_info())
            raise
        finally:
            self.end_transaction_request(btr)

        return response

    def _event_before_send(self, event):
        if isinstance(event, dict) and "type" in event and event["type"] == "http.response.start":
            with self.log_exceptions():
                btr = self.btr
                if not btr:
                    return
                if 'status' in event:
                    self.handle_http_status_code(btr, event['status'], 'HTTP %s' % event['status'])
                if 'headers' in event:
                    headers = event.get('headers', [])
                    self.header_process(headers, btr)
                    response_headers = []
                    for name, value in headers:
                        name = name.decode('latin1')
                        value = value.decode('latin1')
                        response_headers.append((name, value))
                    btr.response_headers = dict(response_headers)

    def _make_start_send_wrapper(self, send):
        """Deal with HTTP status codes, errors and EUM correlation.

        See:
            https://github.com/django/asgiref/blob/master/specs/asgi.rst
            https://github.com/django/asgiref/blob/master/specs/www.rst
        for more information.

        """
        if not isinstance(send, MethodType) and not isinstance(send, FunctionType):
            # hook condition is not match, return original value
            # eg: `send` is a `class` instance with `__call__`
            return send

        if asyncio.iscoroutinefunction(send):
            # @wraps(send)
            async def start_send_wrapper(event):
                self._event_before_send(event)
                return await send(event)
        else:
            # @wraps(send)
            def start_send_wrapper(event):
                self._event_before_send(event)
                return send(event)

        return start_send_wrapper

    def header_process(self, header, btr):
        response_cross_header = make_cross_response_header(self.agent, btr)
        for key, value in items(response_cross_header):
            header.append((key.encode('latin1'), value.encode('latin1')))
