# Copyright (c) Bonree, Inc., and its affiliates
# 2017
# All Rights Reserved

"""Interceptors and utilities for dealing with Odoo frameworks.

"""
import sys
import json
from functools import wraps

from Bonree.agent.interceptor.frameworks import ExitPointInterceptor
from Bonree.config import is_content_type, getgetJsConfigJs
from Bonree.lib import LazyWsgiRequest
from Bonree.agent.models import transactions
from Bonree.agent.interceptor.base import EntryPointInterceptor


class OdooWSGIInterceptor(EntryPointInterceptor):
    def _odoo(self, application, environ, start_response):
        btr = self.start_transaction_request(transactions.ENTRY_PYTHON_WEB, request=LazyWsgiRequest(environ))
        try:
            response = application(environ, self._make_start_response_wrapper(start_response))
        except Exception:
            with self.log_exceptions():
                if btr:
                    btr.add_exception(*sys.exc_info())
            raise
        finally:
            self.end_transaction_request(btr)

        return response

    def _make_start_response_wrapper(self, start_response):
        @wraps(start_response)
        def start_response_wrapper(status, headers, exc_info=None):
            """Deal with HTTP status codes, errors and EUM correlation.

            See https://www.python.org/dev/peps/pep-0333/#the-start-response-callable for more information.

            """
            with self.log_exceptions():
                btr = self.btr
                if btr:
                    # Store the HTTP status code and deal with errors.
                    status_code, msg = status.split(' ', 1)
                    self.handle_http_status_code(btr, status_code, msg)
                    btr.response_headers = dict(headers)

            return start_response(status, headers, exc_info)

        return start_response_wrapper


class OdooWebRequestInterceptor(ExitPointInterceptor):
    def end_data(self, md, req, httprequest):
        btr = self.btr
        if btr and httprequest:
            content_type = getattr(httprequest, 'content_type', None)
            if is_content_type(content_type):
                request_dict = {}
                for k, v  in httprequest.args.items():
                    request_dict[k] = v
                btr.request_content = json.dumps(request_dict)

class OdooResponseInitInterceptor(ExitPointInterceptor):
    def start_data(self, md, resp, *args, **kwargs):
        md.qcontext = kwargs.get('qcontext', {})

    def end_data(self, md, resp, *args, **kwargs):
        btr = self.btr
        if btr and resp.headers:
            content_type = resp.headers.get("Content-Type", None)
            if content_type and is_content_type(content_type):
                btr.response_content = json.dumps(md.qcontext)

class OdooResponseInterceptor(EntryPointInterceptor, ExitPointInterceptor):
    def _render(self, render, resp, *args, **kwargs):
        result = render(resp, *args, **kwargs)
        with self.log_exceptions():
            if self.btr:
                content_type = resp.content_type
                if 'text/html' in content_type and 'utf-8' in content_type and getgetJsConfigJs():
                    content = result.decode()
                    content = self.inject_js_code(content)
                    result = content.encode()
        return result

def intercept_odoo_wsgi_handler(agent, mod):
    OdooWSGIInterceptor(agent, mod).attach(['application', 'application_unproxied'], patched_method_name='_odoo')


def intercept_odoo6_wsgi_handler(agent, mod):
    OdooWSGIInterceptor(agent, mod).attach(['application'], patched_method_name='_odoo')

def intercept_odoo_request_handler(agent, mod):
    # OdooWebRequestInterceptor(agent, mod.WebRequest).attach('__init__')
    # OdooResponseInitInterceptor(agent, mod.Response).attach('__init__')
    OdooResponseInterceptor(agent, mod.Response).attach('render', patched_method_name='_render')
