# Copyright (c) Bonree, Inc., and its affiliates
# 2018
# All Rights Reserved

"""Interceptor for Sanic framework.

"""
import sys

from Bonree.agent.interceptor.base import OnlyStartInterceptor, OnlyEndInterceptor
from Bonree.agent.interceptor.base import EntryPointInterceptor
from Bonree.config import is_content_type
from Bonree.lang import items
from Bonree.agent.models import transactions
from Bonree.agent.interceptor.frameworks import ExitPointInterceptor
from Bonree.agent.services.cross import make_cross_response_header
from Bonree.config import setEnvironIp, setEnvironPort, getgetJsConfigJs

class SanicInterceptor(EntryPointInterceptor, ExitPointInterceptor):
    def _handle_request(self, handle_request, instance, request, *args, **kwargs):
        request.app = instance
        btr = self.start_transaction_request(transactions.ENTRY_PYTHON_WEB, request=request)
        try:
            response = handle_request(instance, request, *args, **kwargs)
            setattr(instance, '__bonr_bt__', btr)
            # if btr:
            #     content_type = request.headers.get("Content-Type", None)
            #     if is_content_type(content_type):
            #         btr.request_content = request.body
        except Exception:
            with self.log_exceptions():
                if btr:
                    btr.add_exception(*sys.exc_info())
            raise

        return response


    def _handle_response(self, handle_response, httpProtocol, response):
        btr = getattr(httpProtocol.request.app, '__bonr_bt__', None)
        with self.log_exceptions():
            if btr:
                if isinstance(response.status, int):
                    status_code = response.status
                    self.handle_http_status_code(btr, status_code, 'HTTP %s' % status_code)
                elif isinstance(response.status, tuple):
                    status_code, msg = response.status
                    self.handle_http_status_code(btr, status_code, msg)
                self.header_process(response.headers, btr)
                btr.response_headers = dict(response.headers)
                content_type = response.content_type
                if content_type and 'text/html' in content_type and 'utf-8' in content_type and getgetJsConfigJs():
                    # response.body is `bytes`, so need translate to `str`
                    content = response.body
                    if isinstance(content, bytes):
                        content = content.decode()
                        content = self.inject_js_code(content)
                        response.body = content.encode()
                # if is_content_type(content_type):
                #     btr.response_content = response.body
        resp = handle_response(httpProtocol, response)
        with self.log_exceptions():
            if btr:
                # delattr(httpProtocol.request.app, '__bonr_bt__')
                self.end_transaction_request(btr)
        return resp

    def _handle_response_send(self, send, response, *args, **kwargs):
        btr = self.btr
        with self.log_exceptions():
            if btr:
                if isinstance(response.status, int):
                    status_code = response.status
                    self.handle_http_status_code(btr, status_code, 'HTTP %s' % status_code)
                elif isinstance(response.status, tuple):
                    status_code, msg = response.status
                    self.handle_http_status_code(btr, status_code, msg)
                self.header_process(response.headers, btr)
                btr.response_headers = dict(response.headers)
                content_type = response.content_type
                if content_type and 'text/html' in content_type and 'utf-8' in content_type and getgetJsConfigJs():
                    # response.body is `bytes`, so need translate to `str`
                    content = response.body
                    if isinstance(content, bytes):
                        content = content.decode()
                        content = self.inject_js_code(content)
                        response.body = content.encode()
                # if is_content_type(content_type):
                #     btr.response_content = response.body
        resp = send(response, *args, **kwargs)
        with self.log_exceptions():
            if btr:
                # delattr(httpProtocol.request.app, '__bonr_bt__')
                self.end_transaction_request(btr)
        return resp

    def header_process(self, header, btr):
        response_cross_header = make_cross_response_header(self.agent, btr)

        if isinstance(header, dict):
            for key, value in items(response_cross_header):
                header[key] = value
        elif isinstance(header, list):
            for key, value in items(response_cross_header):
                header.append((key, value))
            

class SanicRunInterceptor(OnlyStartInterceptor):
    def start_data(self, md, *args, **kwargs):
        host, port = kwargs.get('host', '127.0.0.1'), kwargs.get('port', 8000)
        # print('host[%s] port[%s]' % (host, port,))
        setEnvironIp(host)
        setEnvironPort([port,])


class SanicErrorHandler(OnlyEndInterceptor):
    def end_data(self, md, *args, **kwargs):
        btr = self.btr
        if btr:
            btr.add_exception(*sys.exc_info())


def intercept_sanic(agent, mod):
    SanicInterceptor(agent, mod.Sanic).attach('handle_request', patched_method_name='_handle_request', btrace_data=True)
    SanicRunInterceptor(agent, mod.Sanic).attach('run')
    SanicErrorHandler(agent, mod.handlers.ErrorHandler).attach('response')


def intercept_sanic_response(agent, mod):
    # for sanic version < 21.0.0

    if hasattr(mod.HttpProtocol, 'stream_response'):
        SanicInterceptor(agent, mod.HttpProtocol).attach(['write_response', 'stream_response'], patched_method_name='_handle_response')
    elif hasattr(mod.HttpProtocol, 'write_response'):
        SanicInterceptor(agent, mod.HttpProtocol).attach('write_response', patched_method_name='_handle_response')


def intercept_sanic_response_send(agent, mod):
    # for sanic version > 21.0.0
    if hasattr(mod.BaseHTTPResponse, 'send'):
        SanicInterceptor(agent, mod.BaseHTTPResponse).attach('send', patched_method_name='_handle_response_send')
