import sys
import contextlib

from Bonree.agent.interceptor.frameworks import ExitPointInterceptor
from Bonree.agent.models.methods import MethodData
from Bonree.config import is_content_type
from Bonree.lang import PY3, items
from Bonree.agent.models.cross import SET_COOKIE
from Bonree.agent.models.transactions import ENTRY_PYTHON_WEB
from Bonree.agent.services.cross import make_cross_response_header
from Bonree.lib import LazyWsgiRequest, set_current_ident_btr, unset_current_ident_btr
from Bonree.agent.interceptor.base import EntryPointInterceptor, OnlyStartInterceptor
from Bonree.config import setEnvironIp, setEnvironPort

def header_process(agent, handler, btr):
    response_cross_header = make_cross_response_header(agent, btr)

    # handler.set_cookie(SET_COOKIE, response_cross_header[SET_COOKIE])
    for key, value in items(response_cross_header):
        handler.add_header(key, value)


class TornadoExecuteHandlerInterceptor(EntryPointInterceptor):

    def __execute(self, _execute, handler, *args, **kwargs):
        import tornado.wsgi
        try:
            # Tornado 6.0 remove `tornado.stack_context` module
            import tornado.stack_context as stack_context
        except ImportError:
            stack_context = None
        with self.log_exceptions():
            request_environ = tornado.wsgi.WSGIContainer.environ(handler.request)
            # when we build request_environ, we pop data from handler.request.headers.
            # it will cause problems when we trans this request from tornado to
            # other python web1 framwork. so we append what we just popped.
            if "CONTENT_TYPE" in request_environ:
                handler.request.headers["Content-Type"] = request_environ['CONTENT_TYPE']
            if "CONTENT_LENGTH" in request_environ:
                handler.request.headers["Content-Length"] = request_environ['CONTENT_LENGTH']

            self.get_body_arguments(handler, request_environ)

        btr = self.btr
        md = None
        if not btr:
            btr = self.start_transaction_request(ENTRY_PYTHON_WEB, request=LazyWsgiRequest(request_environ))
            md = MethodData(args, kwargs, self.cls, _execute, btr)
            self.start_trace_call(md, btrace_data=True)
        if btr:
            setattr(handler, '_bonr_transaction', btr)
            setattr(handler, '_bonr_method_data', md)

        if stack_context and btr:
            @contextlib.contextmanager
            def current_bt_manager():
                """Set and unset current_bt as tornado moves between execution contexts.

                By wrapping the handler's execution with this we can ensure that whenever the
                IOLoop is executing code for a particular BT, that BT is the 'current_bt'.
                For more information see http://www.tornadoweb.org/en/stable/stack_context.html.

                """
                set_current_ident_btr(btr)
                try:
                    yield
                except Exception:
                    # Currently can't figure out how to get here, so this code is untested.
                    btr.add_exception(*sys.exc_info())
                    raise
                finally:
                    unset_current_ident_btr()

            with stack_context.StackContext(current_bt_manager):
                result = _execute(handler, *args, **kwargs)
        else:
            result = _execute(handler, *args, **kwargs)

        return result

    def get_body_arguments(self, handler, request_environ):

        def parse_xml():

            import xml.sax
            import xml.sax.handler

            class XmlHandler(xml.sax.handler.ContentHandler):
                def __init__(self):
                    super(XmlHandler, self).__init__()
                    self.buffer = ''
                    self.mapping = {}

                def startElement(self, name, attrs):
                    self.buffer = ''

                def characters(self, content):
                    self.buffer += content

                def endElement(self, name):
                    self.mapping[name] = self.buffer

                def getDict(self):
                    return self.mapping

            if not handler.request.body.strip():
                body_arguments = handler.request.body
                xmlhandler = XmlHandler()
                xml.sax.parseString(body_arguments, xmlhandler)
                body_arguments = xmlhandler.getDict()
                return body_arguments

        def parse_json():
            if handler.request.body:
                body_arguments = handler.request.body
                import json
                body_arguments = json.loads(body_arguments)
                if isinstance(body_arguments, dict):
                    return body_arguments

        def parse_form():
            if handler.request.body_arguments:
                body_arguments = handler.request.body_arguments
                return body_arguments

        #: mapping of mimetypes to parsing functions
        parse_functions = {
            'multipart/form-data': parse_form,
            'application/x-www-form-urlencoded': parse_form,
            'application/json': parse_json,
            'text/xml': parse_xml
        }

        with self.log_exceptions():
            if "CONTENT_TYPE" in request_environ:
                content_type = request_environ['CONTENT_TYPE']
                if 'multipart/form-data' in content_type:
                    content_type = 'multipart/form-data'
                if content_type in parse_functions:
                    parse_func = parse_functions[content_type]
                    body_arguments = parse_func()
                    if body_arguments is not None:
                        request_environ['body_arguments'] = body_arguments

class TornadoRequestHandlerInterceptor(OnlyStartInterceptor):
    def start_data(self, md, *args, **kwargs):
        btr = self.btr
        if btr:
            btr.add_exception(*sys.exc_info())

class TornadoFinshHandlerInterceptor(EntryPointInterceptor):
    def _finish(self, finish, handler, *args, **kwargs):
        btr = getattr(handler, '_bonr_transaction', None)
        if btr:
            with self.log_exceptions():
                header_process(self.agent, handler, btr)
                self.handle_http_status_code(btr, handler._status_code, handler._reason)
                btr.response_headers = handler._headers

        result = finish(handler, *args, **kwargs)

        md = getattr(handler, '_bonr_method_data', None)
        if md:
            self.end_trace_call(md, btrace_data=True)
            self.end_transaction_request(getattr(handler, '_bonr_transaction'))

        return result

class TornadoFallbackHandlerInterceptor(EntryPointInterceptor):
    # When using FallbackHandler, the RequestHandler's finish method is
    # never called.  Wrap the custom fallback callable to end the bt here.
    def _initialize(self, initialize, handler, fallback):
        def _fallback(request):
            btr = self.btr
            if btr:
                with self.log_exceptions():
                    header_process(self.agent, handler, btr)

            fallback(request)

            if btr:
                with self.log_exceptions():
                    self.handle_http_status_code(btr, handler._status_code, handler._reason)
                    btr.response_headers = handler._headers
                md = getattr(handler, '_bonr_method_data', None)
                if md:
                    self.end_trace_call(md, btrace_data=True)
                self.end_transaction_request(btr)

        initialize(handler, _fallback)


class TornadoResponseInterceptor(ExitPointInterceptor):
    def _write(self, write, handler, chunk):
        with self.log_exceptions():
            btr = self.btr
            if btr:
                # request
                # content_type = handler.request.headers.get('Content-Type', None)
                # if is_content_type(content_type):
                #     btr.request_content = handler.request.body

                # response
                content_type = handler._headers.get('Content-Type', None)
                if content_type and 'text/html' in content_type:
                    chunk = self.inject_js_code(chunk)
                # if is_content_type(content_type):
                #     btr.response_content = chunk
        return write(handler, chunk)

class TornadoApplicationListenInterceptor(OnlyStartInterceptor):
    def start_data(self, md, origin_self, *args, **kwargs):
        port = args[0]
        address = kwargs.get('address', '')
        setEnvironIp(address)
        setEnvironPort([port,])
        # print('%s.%s' % (md.cls_name, md.method_name))
        # print('%s:%s' % (address, port))


def intercept_tornado_web(agent, mod):
    import tornado.web
    import tornado.wsgi
    import tornado.ioloop
    if not PY3:
        import tornado.stack_context
    TornadoExecuteHandlerInterceptor(agent, mod.RequestHandler).attach('_execute')
    TornadoRequestHandlerInterceptor(agent, mod.RequestHandler).attach('_handle_request_exception')
    TornadoFinshHandlerInterceptor(agent, mod.RequestHandler).attach('finish')

    TornadoResponseInterceptor(agent, mod.RequestHandler).attach('write', entry=True)
    TornadoFallbackHandlerInterceptor(agent, mod.FallbackHandler).attach('initialize')

    TornadoApplicationListenInterceptor(agent, mod.Application).attach('listen')
