# Copyright (c) Bonree, Inc., and its affiliates
# 2015
# All Rights Reserved

"""Interceptor for web1.py framework.

"""
import sys
from functools import wraps

from Bonree.agent.interceptor.base import OnlyStartInterceptor
from Bonree.agent.interceptor.base import EntryPointInterceptor
from Bonree.agent.interceptor.frameworks import ExitPointInterceptor
from Bonree.agent.interceptor.frameworks.wsgi import WSGIInterceptor
from Bonree.agent.models.methods import MethodData
from Bonree.config import is_content_type
from Bonree.lang import items, PY3
from Bonree.lib import LazyWsgiRequest
from Bonree.agent.models import transactions
from Bonree.agent.services.cross import make_cross_response_header
from Bonree.config import setEnvironIp, setEnvironPort, getgetJsConfigJs


def header_process(agent, header, btr):
    response_cross_header = make_cross_response_header(agent, btr)

    for key, value in items(response_cross_header):
        header.append((key, value))

class WebInterceptor(EntryPointInterceptor, ExitPointInterceptor):

    def start_data(self, md, application, *args, **kwargs):
        application.add_processor(self.auto_js_inject)

    def auto_js_inject(self, handler):
        body = handler()
        with self.log_exceptions():
            btr = self.btr
            if btr:
                from web import webapi
                headers = webapi.ctx.headers
                content_type = None
                for head_name, head_value in headers:
                    if head_name.lower() == 'content-type':
                        content_type = head_value
                if content_type and 'text/html' in content_type and 'utf-8' in content_type and getgetJsConfigJs():
                    if not PY3 and isinstance(body, str):
                        # `str` to `unicode`
                        body = body.decode('utf-8')
                    body = self.inject_js_code(body)
                    if not PY3:
                        body = body.encode('utf-8')
                # if is_content_type(content_type):
                #     btr.response_content = body

        return body

    def _wsgi(self, wsgifunc, application, *args, **kwargs):
        result = wsgifunc(application, *args, **kwargs)

        def wsgi(environ, start_response):
            fcgi_or_scgi = False
            btr = self.btr
            md = MethodData([environ], {}, self.cls, result, btr)
            if not btr:
                btr = self.start_transaction_request(transactions.ENTRY_PYTHON_WEB, request=LazyWsgiRequest(environ))
                md.btr = btr
                fcgi_or_scgi = True
            try:
                if fcgi_or_scgi:
                    md.args.append(self._make_start_response_wrapper(start_response))
                else:
                    md.args.append(start_response)

                self.start_trace_call(md, btrace_data=True)
                response = result(*md.args, **md.kwargs)
                self.end_trace_call(md, btrace_data=True)
            except Exception:
                with self.log_exceptions():
                    if btr:
                        btr.add_exception(*sys.exc_info())
                raise
            finally:
                if fcgi_or_scgi:
                    self.end_transaction_request(btr)

            return response

        return wsgi

    def _make_start_response_wrapper(self, start_response):
        @wraps(start_response)
        def start_response_wrapper(status, headers, exc_info=None):
            """Deal with HTTP status codes, errors and EUM correlation.

            See https://www.python.org/dev/peps/pep-0333/#the-start-response-callable for more information.

            """
            with self.log_exceptions():
                btr = self.btr
                if btr:
                    # Store the HTTP status code and deal with errors.
                    status_code, msg = status.split(' ', 1)
                    self.handle_http_status_code(btr, status_code, msg)
                    # Deal with response header and store the header.
                    header_process(self.agent, headers, btr)
                    btr.response_headers = dict(headers)

            return start_response(status, headers, exc_info)

        return start_response_wrapper


class WebRequestInterceptor(ExitPointInterceptor):
    def start_data(self, md, instance, status, headers, *args, **kwargs):
        btr = self.btr
        if not btr:
            return

        # Store the HTTP status code and deal with errors.
        status_code, msg = status.split(' ', 1)
        self.handle_http_status_code(btr, status_code, msg)
        # Deal with response header and store the header.
        header_process(self.agent, headers, btr)
        btr.response_headers = dict(headers)
        # get request
        # content_type = instance.env.get('CONTENT_TYPE', None)
        # if is_content_type(content_type):
        #     btr.request_content = instance.req.rfile.read()


class WebHttpServerRunSimpleInterceptor(OnlyStartInterceptor):
    def start_data(self, md, *args, **kwargs):
        server_address = args[1]
        host, port = server_address
        # print('host[%s] port[%s]' % (host, port,))
        setEnvironIp(host)
        setEnvironPort([port,])


class WebErrorHandler(OnlyStartInterceptor):
    def start_data(self, md, *args, **kwargs):
        btr = self.btr
        if btr:
            btr.add_exception(*sys.exc_info())


def interceptor_web(agent, mod):
    WebInterceptor(agent, mod.application).attach('wsgifunc', patched_method_name='_wsgi')
    WebErrorHandler(agent, mod.application).attach('internalerror')


def interceptor_web_static(agent, mod):
    WSGIInterceptor(agent, mod.LogMiddleware).attach('__call__', btrace_data=False)
    WebHttpServerRunSimpleInterceptor(agent, mod).attach('runsimple')


def interceptor_start_response(agent, mod):
    WebRequestInterceptor(agent, mod.WSGIGateway).attach('start_response', btrace_data=False)
