# Copyright (c) Bonree, Inc., and its affiliates
# 2015
# All Rights Reserved

"""Interceptors and utilities for dealing with WSGI-based apps/frameworks.

"""

import sys
import threading
from functools import wraps

import Bonree.agent
from Bonree import config
from Bonree.lang import items, str
from Bonree.lib import LazyWsgiRequest
from Bonree.agent.models import transactions
from Bonree.agent.interceptor.base import EntryPointInterceptor
from Bonree.agent.services.cross import make_cross_response_header


class WSGIInterceptor(EntryPointInterceptor):
    def attach(self, application, entry=True, btrace_data=True):
        super(WSGIInterceptor, self).attach(application, patched_method_name='application_callable', entry=entry, btrace_data=btrace_data)

    def application_callable(self, application, instance, environ, start_response):
        if config.CONTAINER_VERSION is None:
            if 'SERVER_SOFTWARE' in environ:
                config.set_container_version(environ['SERVER_SOFTWARE'])
            elif 'uwsgi' in sys.modules:
                config.set_container_version('uwsgi/%s' % str(getattr(sys.modules['uwsgi'], 'version', 'unknown')))
            elif 'mod_wsgi' in sys.modules:
                config.set_container_version('Apache/mod_wsgi (daemon)')

        btr = self.start_transaction_request(transactions.ENTRY_PYTHON_WEB, request=LazyWsgiRequest(environ))
        try:
            if btr:
                response = application(instance, environ, self._make_start_response_wrapper(start_response))
            else:
                response = application(instance, environ, start_response)
        except Exception:
            with self.log_exceptions():
                if btr:
                    btr.add_exception(*sys.exc_info())
            raise
        finally:
            self.end_transaction_request(btr)

        return response

    def _make_start_response_wrapper(self, start_response):
        # @wraps(start_response)
        def start_response_wrapper(status, headers, exc_info=None):
            """Deal with HTTP status codes, errors and EUM correlation.

            See https://www.python.org/dev/peps/pep-0333/#the-start-response-callable for more information.

            """
            with self.log_exceptions():
                btr = self.btr
                if btr:
                    # Store the HTTP status code and deal with errors.
                    status_code, msg = status.split(' ', 1)
                    self.handle_http_status_code(btr, status_code, msg)
                    # Deal with response header and store the header.
                    self.header_process(headers, btr)
                    btr.response_headers = dict(headers)

            return start_response(status, headers, exc_info)

        return start_response_wrapper

    def header_process(self, header, btr):
        response_cross_header = make_cross_response_header(self.agent, btr)

        for key, value in items(response_cross_header):
            header.append((key, value))


class WSGIMiddleware(object):
    def __init__(self, application=None):
        self._application = application
        self._configured = False
        self._interceptor = WSGIInterceptor(Bonree.agent.get_agent_instance(), None)
        self._load_application_lock = threading.Lock()

    def load_application(self):
        wsgi_callable = config.getWsgiCallableObject() or 'application'

        if not config.getWsgiScriptAlias() and not config.getWsgiModule():
            raise AttributeError(
                'Cannot get WSGI application: the Bonree agent cannot load your '
                'application. You must set either BONR_WSGI_SCRIPT_ALIAS or BONR_WSGI_MODULE '
                'in order to load your application.')

        if config.getWsgiModule():
            module_name = config.getWsgiModule()

            if ':' in module_name:
                module_name, wsgi_callable = module_name.split(':', 1)

            __import__(module_name)
            wsgi_module = sys.modules[module_name]
        else:
            import imp
            wsgi_module = imp.load_source('wsgi_module', config.getWsgiScriptAlias())

        if wsgi_callable.endswith('()'):  # "Quick" callback
            app = getattr(wsgi_module, wsgi_callable[:-2])
            app = app()
        else:
            app = getattr(wsgi_module, wsgi_callable)

        self._application = app

    def wsgi_application(self, environ, start_response):
        btr = self._interceptor.start_transaction_request(transactions.ENTRY_PYTHON_WEB, request=LazyWsgiRequest(environ))
        try:
            response = self._application(environ, start_response)
        except Exception:
            with self._interceptor.log_exceptions():
                if btr:
                    btr.add_exception(*sys.exc_info())
            raise
        finally:
            self._interceptor.end_transaction_request(btr)
        return response

    def __call__(self, environ, start_response):
        if not self._configured:
            Bonree.agent.configure(environ)
            self._configured = True

        if not self._application:
            self.load_application()

        if not self._application:
            # CORE-60212 - We are deadlocking inside imp.load_source when
            # WSGI_SCRIPT_ALIAS is used and uwsgi workers are restarted. We
            # actually don't need to go inside `load_application` multiple
            # times, which should prevent the imp.load_source deadlock while
            # improving initial request performance a bit. (We do double-
            # checked locking so we don't take the lock on every __call__,
            # only the initial burst of concurrent ones.)
            with self._load_application_lock:
                if not self._application:
                    self.load_application()

        return self.wsgi_application(environ, start_response)
