# Copyright (c) Bonr, Inc., and its affiliates
# 2018
# All Rights Reserved

"""Intercept ActiveMQ to ensure that ActiveMQ is reported correctly.

"""
from Bonree import rule
from Bonree.agent.interceptor.base import BaseInterceptor, OnlyEndInterceptor
from . import MessageQueueInterceptor
from Bonree.agent.proto.pb_type import BACKEND_TYPE_ACTIVEMQ

class ActiveMQInterceptor(MessageQueueInterceptor):
    def get_backend(self, host, port, destination='/default', params=None):
        backend_properties = {
            'HOST': host,
            'PORT': str(port),
            'URL': rule.url_conversion('activemq', host, port),
            'DATABASE': destination, # dbSchema
            'VENDOR': rule.vendor_conversion('activemq')
        }
        if params:
            if str(params).startswith('/'):
                backend_properties.update({
                    'URL': str(backend_properties.get('URL')) + str(params)
                })
            else:
                backend_properties.update({
                    'URL': str(backend_properties.get('URL')) + '/' + str(params)
                })

        backend_name_format_string = rule.common_name()
        backend = self.agent.backend_registry.get_backend(BACKEND_TYPE_ACTIVEMQ, backend_properties,
                                                          backend_name_format_string)
        backend.connType = rule.connType_conversion('ActiveMQ')
        return backend

    def make_tags(self, md, stompConnection, *args, **kwargs):
        tags  = {}
        host = port = ''
        try:
            host, port = stompConnection.transport.current_host_and_port
        except Exception:
            if stompConnection.transport._Transport__host_and_ports:
                host, port = stompConnection.transport._Transport__host_and_ports[0]

        tags['peer.address'] = rule.peer_address_conversion(ip=host, port=port)
        tags['message_bus.tag'] = md.original_method.__name__.upper()
        if md.original_method.__name__.upper() in ['SEND', 'SUBSCRIBE']:
            tags['message_bus.topic'] = args[0]
        tags['message_bus.group'] = kwargs.get('headers').get('id') if kwargs.get('headers') else ''
        md.tags = tags

class ActiveMQProtocolInterceptor(ActiveMQInterceptor):
    def start_data(self, md, stompConnection, *args, **kwargs):
        try:
            host, port = stompConnection.transport.current_host_and_port
        except Exception:
            if stompConnection.transport._Transport__host_and_ports:
                host, port = stompConnection.transport._Transport__host_and_ports[0]
        method_name = md.original_method.__name__.upper()
        if method_name == 'SEND':
            md.backend = self.get_backend(host, port, args[0])
        else:
            md.backend = self.get_backend(host, port)
            return
        # SEND
        if md.backend:
            destination, message = args[0], args[1]
            btr = self.btr
            if btr:
                btr.mq_data = message
            operation = 'producer'
            md.summary = operation
            md.mq_direction = 1
            if destination.startswith('/'):
                operation = operation + destination
            else:
                operation = operation + '/' + destination
            md.params = destination
            md.operation = operation

class ActiveMQSetListenerInterceptor(OnlyEndInterceptor):
    def end_data(self, md, stompConnection, name, listener):
        ActiveMQListenerOnMessageInterceptor(self.agent, listener.__class__, stompConnection).attach('on_message')

class ActiveMQListenerOnMessageInterceptor(ActiveMQInterceptor):
    def __init__(self, agent, cls, stompConnection):
        super(ActiveMQListenerOnMessageInterceptor, self).__init__(agent, cls)
        self.stompConnection = stompConnection

    def make_tags(self, md, listener, headers, message):
        tags  = {}
        with self.log_exceptions():
            host = port = ''
            stompConnection = self.stompConnection
            try:
                host, port = stompConnection.transport.current_host_and_port
            except Exception:
                if stompConnection.transport._Transport__host_and_ports:
                    host, port = stompConnection.transport._Transport__host_and_ports[0]

            tags['peer.address'] = rule.peer_address_conversion(ip=host, port=port)
            tags['message_bus.tag'] = md.original_method.__name__.upper()
            tags['message_bus.topic'] = headers.get('destination', '/default')
            tags['message_bus.group'] = headers.get('id', '')
            tags['component'] = 'ActiveMQ'
        md.tags = tags

    def start_data(self, md, listener, headers, message):
        host = port = ''
        stompConnection = self.stompConnection
        try:
            host, port = stompConnection.transport.current_host_and_port
        except Exception:
            if stompConnection.transport._Transport__host_and_ports:
                host, port = stompConnection.transport._Transport__host_and_ports[0]
        if isinstance(headers, dict):
            destination = headers.get('destination', '/default')
        else:
            destination = '/default'
        md.backend = self.get_backend(host, port, destination)
        if md.backend is None:
            return
        md.mq_data = message
        operation = 'consumer'
        md.summary = operation
        md.mq_direction = 2
        md.params = destination
        if destination.startswith('/'):
            operation = operation + destination
        else:
            operation = operation + '/' + destination
        md.operation = operation

class ActiveMQStompyInitInterceptor(OnlyEndInterceptor):
    def end_data(self, md, stomp, hostname, port):
        setattr(stomp.frame, 'host', hostname)
        setattr(stomp.frame, 'port', port)

class ActiveMQStompyFrameInterceptor(ActiveMQInterceptor):
    def start_data(self, md, frame, *args, **kwargs):
        md.params = frame.headers['destination'] if 'destination' in frame.headers else ''
        md.backend = self.get_backend(frame.host, frame.port, md.params)
        if md.backend:
            md.backend.connType = rule.connType_conversion('ActiveMQ')
        md.operation = frame.command


# stomp.py, mostly used lib
def intercept_activemq(agent, mod):
    protocol_method_list = ['abort', 'ack', 'begin', 'commit', 'connect', 'disconnect', 'send', 'subscribe', 'unsubscribe']
    ActiveMQProtocolInterceptor(agent, mod.Protocol10).attach(protocol_method_list)

    protocol_method_list.append('nack')
    ActiveMQProtocolInterceptor(agent, mod.Protocol11).attach(protocol_method_list)
    ActiveMQProtocolInterceptor(agent, mod.Protocol12).attach(protocol_method_list)


def intercept_listen_activemq(agent, mod):
    ActiveMQSetListenerInterceptor(agent, mod.BaseConnection).attach('set_listener')


# stompy, rarely used lib
def intercept_stompy(agent, mod):
    ActiveMQStompyInitInterceptor(agent, mod.Stomp).attach('__init__')


def intercept_stompy_frame(agent, mod):
    ActiveMQStompyFrameInterceptor(agent, mod.Frame).attach(['get_message', 'send_frame'])