# Copyright (c) Bonr, Inc., and its affiliates
# 2016
# All Rights Reserved

"""Intercept gRPC to ensure that RPC is reported correctly.

"""
import sys
from Bonree import rule
from Bonree.agent.interceptor.base import ExitCallInterceptor, EntryPointInterceptor, OnlyEndInterceptor
from Bonree.agent.models.transactions import ENTRY_PYTHON_BACKGROUND
from Bonree.agent.proto.pb_type import BACKEND_TYPE_GRPC


def _conversion_host_port(url, default_host='', default_port=0):
    """parse `url` into host and port
    :arg str url: URL to fetch, eg: '127.0.0.1', 'localhost:8000', '9000', '[::]:8000'
    """
    if not url:
        return default_host, default_port

    if url.isdigit():
        return default_host, int(url)

    str_list = url.split(':')

    if len(str_list) == 1:
        return url, default_port

    if len(str_list) == 2:
        return str_list[0], str_list[1]

    # ipv6 eg: [::]:400
    port = str_list[-1]
    host = ':'.join(str_list[:-1])
    return host, port


class GRPCInterceptor(ExitCallInterceptor):
    def start_data(self, md, channel, method, *args, **kwargs):
        host = getattr(channel, '__bonr_target__', None)
        host, port = _conversion_host_port(host)
        backend_name_format_string = rule.common_name('grpc')
        backend_properties = {
            'HOST': host,
            'PORT': port,
            'URL': rule.url_conversion('grpc', host, port)
        }
        md.backend = self.agent.backend_registry.get_backend(BACKEND_TYPE_GRPC, backend_properties,
                                                          backend_name_format_string)
        if md.backend:
            md.backend.connType = rule.connType_conversion('GRPC')
        md.operation = method


    def make_tags(self, md, *args, **kwargs):
        tags  = {}
        host = getattr(md.original_method, '__bonr_target__', None)
        if not host:
            return
        host, port = _conversion_host_port(host)
        tags['peer.address'] = rule.url_conversion('grpc', host, port)
        tags['peer.ip'] = host
        tags['peer.port'] = port
        tags['peer.service'] = 'grpc'
        md.tags = tags

class GRPCChannelInterceptor(OnlyEndInterceptor):
    def end_data(self, md, target, *args, **kwargs):
        bonr_target = target
        setattr(md.response, '__bonr_target__', bonr_target)


class GRPCServerInterceptor(EntryPointInterceptor):
    def _bind_transaction_args(self, rpc_event, state, behavior, *args, **kwargs):
        return rpc_event, behavior

    def attached_process_method(self, origin_method, *args, **kwargs):
        btr = None
        with self.log_exceptions():
            rpc_event, behavior = self._bind_transaction_args(*args, **kwargs)
            bt_name = 'unknown' #  use `bt_name = behavior.__name__`  when lowerst support python3.7
            if hasattr(behavior, '__name__'):
                bt_name = behavior.__name__
            else:
                if len(args) >= 4 and hasattr(args[3], '__name__'):
                    bt_name = getattr(args[3], '__name__')

            btr = self.start_transaction_request(ENTRY_PYTHON_BACKGROUND, name=bt_name)

        if btr:
            try:
                response = origin_method(*args, **kwargs)
            except Exception:
                with self.log_exceptions():
                    btr.add_exception(*sys.exc_info())
                raise
            finally:
                with self.log_exceptions():
                    btr.status_code = 0
                    self.end_transaction_request(btr)

            return response

        return origin_method(*args, **kwargs)


class XMLRPCServerInterceptor(EntryPointInterceptor):
    def attached_process_method(self, origin_method, instance, method, params):
        btr = None
        with self.log_exceptions():
            bt_name = method
            btr = self.start_transaction_request(ENTRY_PYTHON_BACKGROUND, name=bt_name)
        if btr:
            try:
                response = origin_method(instance, method, params)
            except Exception:
                with self.log_exceptions():
                    btr.add_exception(*sys.exc_info())
                raise
            finally:
                with self.log_exceptions():
                    btr.status_code = 0
                    self.end_transaction_request(btr)

            return response
        return origin_method(instance, method, params)


def intercept_grpc(agent, mod):
    GRPCInterceptor(agent, mod.Channel).attach(['unary_unary', 'unary_stream', 'stream_unary', 'stream_stream'])


def intercept_grpc_channel(agent, mod):
    GRPCChannelInterceptor(agent, mod).attach(['insecure_channel', 'secure_channel'])


def intercept_grpc_server(agent, mod):
    GRPCServerInterceptor(agent, mod).attach('_unary_response_in_pool', patched_method_name='attached_process_method')


def intercept_xmlrpc_server(agent, mod):
    XMLRPCServerInterceptor(agent, mod.SimpleXMLRPCServer).attach('_dispatch', patched_method_name='attached_process_method')
