# Copyright (c) Bonr, Inc., and its affiliates
# 2019
# All Rights Reserved

"""Intercept Kfaka to ensure that Kafka is reported correctly.

"""
# from Bonree.agent.interceptor.base import ExitCallInterceptor
from Bonree import rule
from Bonree.agent.interceptor.base import BaseInterceptor
from . import MessageQueueInterceptor
from Bonree.agent.proto.pb_type import BACKEND_TYPE_KAFKA


def get_address(obj):
    try:
        address = obj.config['bootstrap_servers']
        url_list = []
        if isinstance(address, str):
            url_list.append(address)
        else:
            for addr in address:
                url_list.append(addr)

        return ';'.join(url_list)
    except Exception:
        pass


class KafkaSendInterceptor(MessageQueueInterceptor):
    def make_tags(self, md, producer, topic, *args, **kwargs):
        tags  = {}
        address = get_address(producer)
        tags['peer.address'] = address
        tags['message.bus_topic'] = str(topic)
        md.tags = tags

    def start_data(self, md, producer, topic, *args, **kwargs):
        backend_name_format_string = rule.common_name('')
        address = get_address(producer)
        backend_properties = {
            'HOST': address,
            # 'PORT': '0',
            'URL': 'kafka://' + address,
            'DATABASE': '/' + topic,
            'VENDOR': rule.vendor_conversion('Kafka')
        }

        backend = self.agent.backend_registry.get_backend(BACKEND_TYPE_KAFKA, backend_properties,
                                                          backend_name_format_string)
        backend.connType = 'Producer'
        md.backend = backend
        value = kwargs.get('value', None)
        if value:
            md.mq_data = value

        md.summary = 'producer'
        md.operation = 'producer/' + topic
        md.mq_direction = 1


class KafkaPollInterceptor(MessageQueueInterceptor):
    def make_tags(self, md, consumer, timeout_ms=0, max_records=None, *args, **kwargs):
        tags  = {}
        address = get_address(consumer)
        tags['peer.address'] = address
        tags['message.bus.group'] = consumer.config.get('group_id', None)
        md.tags = tags

    def start_data(self, md, consumer, *args, **kwargs):
        backend_name_format_string = rule.common_name('')
        address = get_address(consumer)
        backend_properties = {
            'HOST': address,
            # 'PORT': '0',
            'URL': 'kafka://' + address,
            'DATABASE': '/defaultTopic',
            'VENDOR': rule.vendor_conversion('Kafka')
        }

        backend = self.agent.backend_registry.get_backend(BACKEND_TYPE_KAFKA, backend_properties,
                                                          backend_name_format_string)
        backend.connType = 'Consumer'
        md.backend = backend

    def end_data(self, md, *args, **kwargs):
        md.summary = 'consumer'
        md.mq_direction = 2
        msg_pack = md.response
        md.operation = 'consumer/defaultTopic'
        for _, consumeRecords in msg_pack.items():
            if not isinstance(consumeRecords, list) or len(consumeRecords) == 0:
                continue
            consumeRecord = consumeRecords[0]
            md.mq_data = consumeRecord.value
            md.operation = 'consumer/' + consumeRecord.topic
            if md.backend:
                md.backend.identifying_properties['DATABASE'] = '/' + consumeRecord.topic
            # TODO: deal with one more consumeRecord
            break
        transport_be = md.transport_be
        if transport_be:
            transport_be.update_be_call(md)


class KafkaConsumerInterceptor(MessageQueueInterceptor):
    def make_tags(self, md, consumer, *args, **kwargs):
        tags  = {}
        with self.log_exceptions():
            address = get_address(consumer)
            tags['peer.address'] = str(address)
            tags['message.bus.group'] = consumer.config.get('group_id', None)
        md.tags = tags

    def start_data(self, md, consumer, *args, **kwargs):
        backend_name_format_string = rule.common_name('')
        address = get_address(consumer)
        backend_properties = {
            'HOST': address,
            # 'PORT': '0',
            'URL': 'kafka://' + address,
            'DATABASE': '/defaultTopic',
            'VENDOR': rule.vendor_conversion('Kafka')
        }

        backend = self.agent.backend_registry.get_backend(BACKEND_TYPE_KAFKA, backend_properties,
                                                          backend_name_format_string)
        backend.connType = 'Consumer'
        md.backend = backend

    def end_data(self, md, consumer, *args, **kwargs):
        consumeRecord = md.response
        if md.backend:
            md.backend.identifying_properties['DATABASE'] = '/' + consumeRecord.topic
        md.summary = 'consumer'
        md.mq_direction = 2
        md.mq_data = consumeRecord.value
        md.operation = 'consumer/' + consumeRecord.topic
        transport_be = md.transport_be
        if transport_be:
            transport_be.update_be_call(md)


def intercept_kafka_python(agent, mod):
    KafkaSendInterceptor(agent, mod.KafkaProducer).attach('send')


def intercept_kafka_consumer(agent, mod):
    KafkaPollInterceptor(agent, mod.KafkaConsumer).attach('poll')
    KafkaConsumerInterceptor(agent, mod.KafkaConsumer).attach('__next__')