from functools import wraps
from types import FunctionType, MethodType

from Bonree import rule
from Bonree.agent.interceptor.base import CustomBtInterceptor, ExitCallInterceptor
from Bonree.agent.interceptor.frameworks.openstack import OpenStackBackGroundInterceptor
from Bonree.agent.interceptor.http.rabbitmq import RabbitInterceptor
from Bonree.agent.models.methods import MethodData
from Bonree.agent.models.transactions import ENTRY_PYTHON_BACKGROUND
from Bonree.agent.proto.pb_type import BACKEND_TYPE_RABBITMQ, BACKEND_TYPE_REDIS
from Bonree.agent.services.transport import TransportBe
from Bonree.logger import SYS_LOGGER


class OsloRabbitConsumerInterceptor(OpenStackBackGroundInterceptor):
    backend_name_format_string = rule.common_name()

    def __init__(self, agent, cls):
        super(OpenStackBackGroundInterceptor, self).__init__(agent, cls)
        self.vendor = rule.vendor_conversion('RabbitMQ')

    def get_backend(self, scheme, host, port, version, destination='defaultExchangeName/defaultRoutingKey'):
        """

        Parameters
        ----------
        host : rabbitmq server host
        port : rabbirmq server port

        """

        backend_properties = {
            'HOST': host,
            'PORT': str(port),
            'VENDOR': rule.vendor_conversion(scheme),
            'URL': rule.url_conversion(scheme, host, port),
            'DATABASE': destination,  # dbSchema
            'VERSION': version
        }
        tipe = BACKEND_TYPE_RABBITMQ
        if scheme == 'redis':
            tipe = BACKEND_TYPE_REDIS
        backend = self.agent.backend_registry.get_backend(tipe, backend_properties, self.backend_name_format_string)
        backend.connType = rule.connType_conversion(scheme)
        return backend

    def start_exit_call(self, md, *args, **kwargs):
        """Start an exit call.

        """
        # if not self.agent.check_and_open_sdk():
        #     return None
        # self.agent.update_config_response()
        # if not config.get_enabled():
        #     return None
        if not md.backend or md.exit_flag:
            return None
        btr = self.btr

        transport_be = TransportBe()
        md.transport_be = transport_be
        if md.is_rpc and btr:
            transport_be.start_be_rpc(md)
        else:
            transport_be.start_be_call(md)
        # with_btr = (btr is None)
        # BT_LOGGER.debug("start exitcall:[{0}.{1}] with btr[{2}] backend[{3}]".format(md.cls_name, md.method_name, with_btr, md.backend))

    def end_exit_call(self, md, *args, **kwargs):
        """End the exit call.

        """
        btr = self.btr
        transport_be = md.transport_be
        if transport_be is None:
            return

        if md.exc_info and btr:
            btr.add_exception(*md.exc_info)
        if md.is_rpc and btr:
            transport_be.end_be_rpc(md)
        else:
            transport_be.end_be_call(md)

    def _get_host_port_kombu_connection(self, channel=None):
        """
        connection is instance of kombu.connection.Connection
        channel is instance of kombu.transport.<pyamqp>.Channel
        """
        host, port, scheme = '', '', ''

        if channel is not None:
            host = channel.connection.host  # 'host[:port]' port default 5672
            scheme = 'amqp'
            port = '5672'
            if ':' in host:
                tmp = host.split(':')
                host = tmp[0]
                port = tmp[1]

        if scheme in ('amqp', 'librabbitmq'):
            scheme = 'rabbitmq'

        version = 'unknown'

        return scheme, host, port, version

    def make_tags(self, md, consumer, message, *args, **kwargs):
        with self.log_exceptions():
            tags = message.delivery_info.copy() if isinstance(message.delivery_info, dict) else dict()
            scheme, host, port, version = self._get_host_port_kombu_connection(message.channel)
            tags['peer.address'] = rule.peer_address_conversion(ip=host, port=port)
            md.tags = tags

    def start_data(self, md, consumer, message, *args, **kwargs):
        body = message.body
        scheme, host, port, _ = self._get_host_port_kombu_connection(message.channel)
        version = 'unknown'
        if hasattr(message.channel.connection, 'server_properties'):
            version = message.channel.connection.server_properties.get('version', 'unknown')
        delivery_info = message.delivery_info
        exchangeName = 'defaultExchangeName'
        routingKey = 'defaultRouteKey'
        if isinstance(delivery_info, dict):
            exchangeName = delivery_info.get('exchange', exchangeName)
            routingKey = delivery_info.get('routing_key', routingKey)
        destination = '%s/%s' % (exchangeName, routingKey)
        md.backend = self.get_backend(scheme, host, port, version, destination)
        md.mq_data = body
        md.operation = 'consumer/{0}'.format(destination)
        md.summary = 'consumer'
        md.mq_direction = 2


def intercept_oslo_message_rpc_client(agent, mod):
    """oslo_messaging.rpc.client"""
    methods = ['cast', 'call']
    for method in methods:
        OpenStackBackGroundInterceptor(agent, mod.RPCClient).set_bt_name(method).attach(method, 'custom_bt_entry', True, True)

    for method in methods:
        OpenStackBackGroundInterceptor(agent, mod._BaseCallContext).set_bt_name(method).attach(method, 'custom_bt_entry', True, True)


def intercept_oslo_message_rabbit(agent, mod):
    """oslo_messaging._drivers.impl_rabbit"""

    OsloRabbitConsumerInterceptor(agent, mod.Consumer).set_bt_name('_callback').attach('_callback', 'custom_bt_entry', True, True)

    cls = mod.Connection
    # for method in dir(mod.Connection):
    for method in ['direct_send']:
        if method.startswith('_') or method == 'ensure' or method == 'consume':
            continue
        obj = getattr(cls, method, None)
        if isinstance(obj, (FunctionType, MethodType)):
            OpenStackBackGroundInterceptor(agent, mod.Connection).set_bt_name(method).attach(method, 'custom_bt_entry', True, True)
