# Copyright (c) Bonree, Inc., and its affiliates
# 2015
# All Rights Reserved

"""Interceptor for RabbitMQ.

"""

from __future__ import unicode_literals

from Bonree import rule
from Bonree.lang import str
from . import MessageQueueInterceptor
from Bonree.agent.proto.pb_type import BACKEND_TYPE_RABBITMQ, BACKEND_TYPE_REDIS

def _get_host_port_kombu_connection(connection, channel=None):
    """
    connection is instance of kombu.connection.Connection
    channel is instance of kombu.transport.<pyamqp>.Channel
    """
    host, port, scheme = '', '', ''
    if connection is not None:
        host = connection.hostname
        port = connection.port
        scheme = connection.transport_cls
    if connection is None and channel is not None:
        host = channel.connection.host #'host[:port]' port default 5672
        scheme = 'amqp'
        port = '5672'
        if ':' in host:
            tmp = host.split(':')
            host = tmp[0]
            port = tmp[1]

    if scheme in ('amqp', 'librabbitmq'):
        scheme = 'rabbitmq'

    version = 'unknown'
    if connection and hasattr(connection.connection, 'server_properties'):
        version = connection.connection.server_properties.get('version')

    return scheme, host, port, version


class RabbitInterceptor(MessageQueueInterceptor):
    backend_name_format_string = rule.common_name()

    def __init__(self, agent, cls):
        super(RabbitInterceptor, self).__init__(agent, cls, rule.vendor_conversion('RabbitMQ'))

    def get_backend(self, scheme, host, port, version, destination='defaultExchangeName/defaultRoutingKey'):
        """

        Parameters
        ----------
        host : rabbitmq server host
        port : rabbirmq server port

        """

        backend_properties = {
            'HOST': host,
            'PORT': str(port),
            'VENDOR': rule.vendor_conversion(scheme),
            'URL': rule.url_conversion(scheme, host, port),
            'DATABASE': destination, # dbSchema
            'VERSION': version
        }
        tipe = BACKEND_TYPE_RABBITMQ
        if scheme == 'redis':
            tipe = BACKEND_TYPE_REDIS
        backend = self.agent.backend_registry.get_backend(tipe, backend_properties, self.backend_name_format_string)
        backend.connType = rule.connType_conversion(scheme)
        return backend

class RabbitReceiveInterceptor(RabbitInterceptor):
    def make_tags(self, md, consumer, *args, **kwargs):
        with self.log_exceptions():
            message = args[1]
            tags = message.delivery_info.copy() if isinstance(message.delivery_info, dict) else dict()
            scheme, host, port, version = _get_host_port_kombu_connection(consumer.connection, message.channel)
            tags['peer.address'] = rule.peer_address_conversion(ip=host, port=port)
            md.tags = tags

    def start_data(self, md, consumer, *args, **kwargs):
        body = args[0]
        message = args[1]
        scheme, host, port, _ = _get_host_port_kombu_connection(consumer.connection, message.channel)
        version = 'unknown'
        if hasattr(message.channel.connection, 'server_properties'):
            version = message.channel.connection.server_properties.get('version', 'unknown')
        delivery_info = message.delivery_info
        exchangeName = 'defaultExchangeName'
        routingKey = 'defaultRouteKey'
        if isinstance(delivery_info, dict):
            exchangeName = delivery_info.get('exchange', exchangeName)
            routingKey = delivery_info.get('routing_key', routingKey)
        destination = '%s/%s' % (exchangeName, routingKey)
        md.backend = self.get_backend(scheme, host, port, version, destination)
        md.mq_data = body
        md.operation = 'consumer/{0}'.format(destination)
        md.summary = 'consumer'
        md.mq_direction = 2

'''
class RabbitQueueInterceptor(RabbitInterceptor):
    def make_tags(self, md, queue, *args, **kwargs):
        with self.log_exceptions():
            tags  = {}
            tags['queue'] = queue.name 
            tags['queue_bind_routing_key'] = queue.routing_key
            tags['exchange'] = queue.exchange.name if '' != queue.exchange.name else 'default'
            connection = queue.channel.connection
            connection_class_str = str(connection.__class__)
            if "<class 'kombu.transport.librabbitmq.Connection'>" == connection_class_str:
                host, port = connection.hostname, connection.port
            elif "<class 'kombu.transport.pyamqp.Connection'>" == connection_class_str:
                host, port = connection.host.split(':')
            else:
                host, port = 'unknown', '0'
            tags['peer.address'] = rule.peer_address_conversion(ip=host, port=port)
            md.tags = tags

    def start_data(self, md, queue, *args, **kwargs):
        with self.log_exceptions():
            connection = queue.channel.connection
            connection_class_str = str(connection.__class__)
            if "<class 'kombu.transport.librabbitmq.Connection'>" == connection_class_str:
                host, port = connection.hostname, connection.port
            elif "<class 'kombu.transport.pyamqp.Connection'>" == connection_class_str:
                host, port = connection.host.split(':')
            else:
                return
            version = connection.server_properties.get('version')
            destination = '%s/%s' % (queue.exchange.name, queue.routing_key)
            md.backend = self.get_backend(host, port, version, destination)
            if md.backend:
                md.operation = 'receive/{0}/{1}'.format(queue.exchange.name, queue.routing_key)
                md.summary = 'receive'
                md.mq_direction = 2
'''

class RabbitPublishInterceptor(RabbitInterceptor):
    def make_tags(self, md, producer, *args, **kwargs):
        tags  = {}
        with self.log_exceptions():
            tags['exchange'] = producer.exchange.name
            scheme, host, port, _ = _get_host_port_kombu_connection(producer.connection)
            tags['peer.address'] = rule.peer_address_conversion(ip=host, port=port)
        md.tags = tags

    def start_data(self, md, producer, *args, **kwargs):
        scheme, host, port, version = _get_host_port_kombu_connection(producer.connection)
        md.mq_data = args[0]
        exchangeName = producer.exchange.name
        # SimpleQueue, routing_key is None
        routingKey = kwargs.get('routing_key', None)
        if routingKey is None:
            routingKey = getattr(producer, 'routing_key', None)
        if routingKey is None:
            routingKey = exchangeName
        destination = '%s/%s' % (exchangeName, routingKey)
        md.backend = self.get_backend(scheme, host, port, version, destination)
        md.operation = 'producer/{0}'.format(destination)
        md.summary = 'producer'
        md.mq_direction = 1


def intercept_rabbitmq(agent, mod):
    # RabbitQueueInterceptor(agent, mod.Queue).attach('consume')
    RabbitReceiveInterceptor(agent, mod.Consumer).attach('receive')
    RabbitPublishInterceptor(agent, mod.Producer).attach('publish')
