# Copyright (c) Bonree, Inc., and its affiliates
# 2018
# All Rights Reserved

"""
Definition of base entry and exit point interceptors.
"""
from __future__ import unicode_literals

from Bonree import rule
from Bonree.agent.interceptor.base import ExitCallInterceptor, EntryPointInterceptor, OnlyStartInterceptor
from Bonree.agent.proto.pb_type import BACKEND_TYPE_THRIFT
from Bonree.agent.models.cross import BR_CS_REQ, BR_CS_RSP
from Bonree.agent.services.cross import make_cross_server
from Bonree.lib import ThriftRequest

class ThriftInterceptor(ExitCallInterceptor):
    backend_name_format_string = rule.common_name

    def start_data(self, md, tprotocol, transport):
        md.backend = self.get_backend(transport.host, transport.port, 'Thrift', transport.path)
        md.operation = transport.__class__.__name__

    def make_tags(self, md, tprotocol, transport):
        tags  = {}
        host = transport.host
        port = transport.port
        tags['peer.address'] = rule.peer_address_conversion(ip=host, port=port)
        tags['peer.ip'] = host
        tags['peer.port'] = port
        md.tags = tags

    def get_backend(self, host, port, scheme, url=None):
        backend_properties = {
            'HOST': host,
            'PORT': port,
        }
        backend = self.agent.backend_registry.get_backend(BACKEND_TYPE_THRIFT, backend_properties,
                                                          self.backend_name_format_string(scheme))
        backend.connType = rule.connType_conversion(scheme)
        return backend


TType_STRING = 11 # see `thrift.Thrift.TType` STRING


class ThriftWriteBtHeader(OnlyStartInterceptor):
    def start_data(self, md, protocol, *args, **kwargs):
        btr = self.btr
        if not btr:
            return
        if btr.rpc_headers and BR_CS_REQ in btr.rpc_headers:
            # write req header
            # Write in the field name value to facilitate identification
            cross_head = btr.rpc_headers[BR_CS_REQ]
            if isinstance(cross_head, bytes):
                cross_head = cross_head.decode('latin1') # `bytes` to `str`
            protocol.writeFieldBegin(BR_CS_REQ, TType_STRING, -1)
            protocol.writeString('%s%s' % (BR_CS_REQ, cross_head))
            protocol.writeFieldEnd()
        elif btr.response_headers and BR_CS_RSP in btr.response_headers:
            # write resp header
            cross_head = btr.response_headers[BR_CS_RSP]
            if isinstance(cross_head, bytes):
                cross_head = cross_head.decode('latin1') # `bytes` to `str`
            protocol.writeFieldBegin(BR_CS_RSP, TType_STRING, -1)
            protocol.writeString('%s%s' % (BR_CS_RSP, cross_head))
            protocol.writeFieldEnd()


class ThriftProtocolSkip(EntryPointInterceptor):
    def _skip(self, skip, protocol, ttype):
        if ttype == TType_STRING:
            btr = self.btr
            if btr:
                maybe_header = protocol.readString()
                with self.log_exceptions():
                    if isinstance(maybe_header, str) and maybe_header.startswith(BR_CS_REQ):
                        headers = {BR_CS_REQ: maybe_header[len(BR_CS_REQ):] }
                        request = ThriftRequest(headers=headers)
                        btr.request = request
                        make_cross_server(btr, request)
                        btr.response_headers = {BR_CS_RSP: btr.cross_resp}
                    elif isinstance(maybe_header, str) and maybe_header.startswith(BR_CS_RSP):
                            resp_header = maybe_header[len(BR_CS_RSP):]
                            if not btr.rpc_headers:
                                btr.rpc_headers = {}
                            btr.rpc_headers[BR_CS_RSP] = resp_header
                return None

        return skip(protocol, ttype)


def interceptor_thrift(agent, mod):
    ThriftInterceptor(agent, mod.TProtocolBase).attach('__init__')


def interceptor_thrift_binary_field(agent, mod):
    ThriftWriteBtHeader(agent, mod.TBinaryProtocol).attach('writeFieldStop')


def interceptor_thrift_json_field(agent, mod):
    ThriftWriteBtHeader(agent, mod.TJSONProtocol).attach('writeFieldStop')


def interceptor_thrift_compact_field(agent, mod):
    ThriftWriteBtHeader(agent, mod.TCompactProtocol).attach('writeFieldStop')


def interceptor_thrift_skip(agent, mod):
    ThriftProtocolSkip(agent, mod.TProtocolBase).attach('skip', patched_method_name='_skip')
