# Copyright (c) Bonree, Inc., and its affiliates
# 2015
# All Rights Reserved

"""Interceptors for dealing with DB-API 2.0 compatible libraries.

"""
from Bonree import config, rule
from Bonree.lib import parse_args
from Bonree.lang import str, PY3
from Bonree.logger import BT_LOGGER
from .. import VENDOR_PROPERTY_MAX_LEN
from ..base import ExitCallInterceptor, BaseInterceptor, OnlyEndInterceptor
import copy


class DbAPICursorInterceptor(ExitCallInterceptor):
    SUPPORT_SQL_FORMAT = ("mysql", "postgresql", "oracle", "hive")
    def start_data(self, md, cursor, operation, *args, **kwargs):
        backend = None
        connection = self.get_connection(cursor)
        try:
            backend = getattr(connection, '_bonr_backend', None)
            if backend is None:
                # some version of Mysqldb returns a weekref of the connection
                # thus we have to try it again if it is none
                backend = getattr(connection.__repr__(), '_bonr_backend', None)
        except Exception as e:
            BT_LOGGER.error("Can't get backend, because {0}".format(e))

        if backend:
            oper_str = str(operation) or parse_args(args, kwargs)
            oper_str = oper_str.replace('\n', '').replace(r'"', '')

            md.backend = backend
            md.operation = oper_str

    def sql_format(self, md, cursor, operation, *args, **kwargs):
        if md.backend.connType not in self.SUPPORT_SQL_FORMAT:
            return

        md.summary = md.operation
        if md.backend.connType == "mysql":
            _operation = cursor._executed
            if _operation:
                if not PY3:
                    _operation = repr(_operation)[1:-1]
                md.operation = str(_operation)
        elif md.backend.connType == "postgresql":
            md.operation = str(cursor.query)
        elif md.backend.connType == "hive":
            if len(args) != 1:
                return
            parameters = args[0]
            if isinstance(parameters, (tuple, list)) or isinstance(parameters, dict):
                md.operation = operation % parameters
        elif md.backend.connType == "oracle":
            if len(args) != 1:
                return
            parameters = args[0]
            _operation = str(operation)
            if isinstance(parameters, (tuple, list)):
                for i, value in enumerate(parameters, start=1): # oracle first is :1
                    _operation = _operation.replace(str(':%d' % i), str(value))
            elif isinstance(parameters, dict):
                for name, value in parameters.items():
                    _operation = _operation.replace(str(':%s' % name), str(value))
            md.operation = _operation

        md.operation = md.operation[:config.getSqlMaxLength()]
        transport_be = md.transport_be
        if transport_be:
            transport_be.update_be_call(md)

    def end_data(self, md, cursor, operation, *args, **kwargs):
        btr = md.btr
        if not btr:
            return
        if not md.backend:
            return
        if args or kwargs:
            # SQL format
            self.sql_format(md, cursor, operation, *args, **kwargs)

        if not self._need_explain_sql(md, operation, btr):
            return
        explain_operation = str('EXPLAIN ' + str(operation))
        cursor = copy.copy(cursor)
        if hasattr(cursor, '_defer_warnings'):
            cursor._defer_warnings = True
        md.original_method(cursor, explain_operation, *args, **kwargs)
        bonr_explain = cursor.fetchall()
        if not bonr_explain:
            return
        md.sql_explain.append(str('\t').join([i[0] for i in cursor.description]))
        bonr_explain = list(bonr_explain)
        for index, item in enumerate(bonr_explain):
            if isinstance(item, dict):
                item_copy = copy.deepcopy(item)
                item = []
                key_list = [i[0] for i in cursor.description]
                for key in key_list:
                    item.append(item_copy.get(key))
                bonr_explain[index] = item
            elif isinstance(item, (list, tuple)):
                bonr_explain[index] = list(item)
            for i, v in enumerate(bonr_explain[index]):
                if v is None:
                    bonr_explain[index][i] = 'NULL'
        for value in bonr_explain:
            md.sql_explain.append('\n' + str('\t').join(str(s) for s in value))

    def _need_explain_sql(self, md, operation, btr):
        need_explain = md.need_explain(btr)
        # BT_LOGGER.debug('[%s.%s] need explain[%s] connType[%s]' % (md.cls_name, md.method_name, need_explain, md.backend.connType))
        if not need_explain:
            return False
        # mysql 5.5 & 5.0: SELECT
        # mysql 5.6 & 5.7: SELECT DELETE INSERT REPLACE UPDATE
        # postgresql: SELECT DELETE INSERT UPDATE
        # TODO: oracle: SELECT DELETE INSERT UPDATE
        if md.backend.connType == 'postgresql':
            return str(operation)[:20].lower().startswith(('select', 'update',
                                                      'insert', 'delete'))
        if md.backend.connType == 'mysql':
            if md.backend.identifying_properties.get('VERSION', '') < '5.6':
                return str(operation)[:20].lower().startswith('select')
            else:
                return str(operation)[:20].lower().startswith(('select', 'update',
                                                      'insert', 'delete', 'replace'))
        return False


    def make_tags(self, md, cursor, operation, *args, **kwargs):
        tags  = {}
        connection = self.get_connection(cursor)
        backend = getattr(connection, '_bonr_backend', None)
        if backend:
            tags['peer.address'] = rule.peer_address_conversion(backend.identifying_properties.get('URL', None))
            tags['db.instance'] = str(backend.identifying_properties.get('DATABASE', 'unknown'))
            _operation = operation
            if not PY3:
                _operation = repr(_operation)[1:-1]
            tags['db.statement'] = str(_operation) or parse_args(args, kwargs)
            tags['db.type'] = backend.connType
        md.tags = tags

    def get_connection(self, cursor):
        """Return the `Connection` object used to create this `Cursor`.

        """
        raise NotImplementedError


class DbAPIConnectInterceptor(OnlyEndInterceptor):
    backend_naming_format_string = '{DATABASE}-{VENDOR}-{HOST}'

    def __init__(self, agent, cls, vendor, get_backend_properties):
        self.vendor = vendor[:VENDOR_PROPERTY_MAX_LEN]
        self.get_backend_properties = get_backend_properties
        super(DbAPIConnectInterceptor, self).__init__(agent, cls)

    def end_data(self, md, connection, *args, **kwargs):
        md.backend = self.get_backend(connection, *args, **kwargs)
        if md.backend:
            setattr(md.args[0], '_bonr_backend', md.backend)

    def get_backend(self, connection, *args, **kwargs):
        if self.agent.backend_registry is None:
            return None
        host, port, dbname, dbversion = self.get_backend_properties(connection, *args, **kwargs)
        backend = self.agent.backend_registry.get_db_backend(self.vendor, host, port, dbname, dbversion)
        backend.connType = rule.connType_conversion(self.vendor)
        return backend

class DbAPIConnectCursorInterceptor(OnlyEndInterceptor):
    def __init__(self, agent, cls, cursor_interceptor):
        self.cursor_interceptor = cursor_interceptor
        super(DbAPIConnectCursorInterceptor, self).__init__(agent, cls)
        self.cur_cls_set = set()

    def end_data(self, md, *args, **kwargs):
        if type(md.response) not in self.cur_cls_set:
            self.cursor_interceptor(self.agent, type(md.response)).attach(['execute', 'executemany'],btrace_data=True)
            self.cur_cls_set.add(type(md.response))


def db_api_attach(agent, cls, cursor_interceptor, get_backend_properties, connect_func, vendor, btrace_data=True):
    DbAPIConnectInterceptor(agent, cls, vendor, get_backend_properties).attach(connect_func, btrace_data=False)
    DbAPIConnectCursorInterceptor(agent, cls, cursor_interceptor).attach('cursor', btrace_data=False)
