import time
from string import Formatter

from Bonree import config
from Bonree.agent.core import sqlparser
from Bonree.config import PY3, enableSQLExplain
from Bonree.lib import set_time_unit, current_thread_ident, get_func_signa, get_btr_from_attr

EXIT_SQL = 0
EXIT_NOSQL = 1
EXIT_REMOTE = 2

HOOK_FLAG = '_bonr_intercepted'

class BackendNamer(Formatter):
    """A utility class for generating backend names from their properties.

    """
    def format(self, format_string, properties, custom_backend_prefix=None):
        """Format the input string using the supplied properties.

        This function is similar to the string formater described in PEP-3101.
        However there some important differences.  The behaviour is best
        illustrated with some examples::

            >>> format_string = 'http://{HOST}:{PORT}{URL}?{QUERY STRING}'
            >>> format(format_string, {'HOST': 'Bonree.com'})
            'http://Bonree.com'
            >>> format(format_string, {'HOST': 'Bonree.com', 'PORT': '80'})
            'http://Bonree.com:80'
            >>> format(format_string, {'QUERY STRING': 'name=tom'})
            'name=tom'
            >>> format(format_string, {'PORT': '80', 'QUERY STRING': 'foo=bar'})
            '80?foo=bar'
            >>> format(format_string, {'HOST': '', 'PORT': '', 'URL': '/path/'})
            '/path/''
            >>> format(format_string, {'HOST': 'tom.com'},
            ... custom_backend_prefix='BACKEND')
            'BACKEND - http://tom.com'

        Some things to note:

        1)  Missing replacement fields are skipped, along with their preceeding
            literal text.
        2)  Apart from the first field, literal text preceeding a replacement
            field is skipped unless there is something in front of it in the
            output.
        3)  Empty properties are ignored.
        4)  If a custom prefix is provided, it is inserted at the front of the
            output, with a ' - ' separator.
        5)  'SERVER POOL' is a special case.  We take the last item in the
            newline separated string.

        """
        result = []
        for i, (literal_text, field_name, format_spec, conversion) in \
                enumerate(self.parse(format_string)):

            if field_name is not None:
                value = properties.get(field_name, '')

                if value:
                    if result or i == 0:
                        # We only want to add the prefix if this is the very
                        # first item, or there is already text in the string.
                        result.append(literal_text)

                    # SERVER POOL is newline separated, take the last item.
                    if field_name == 'SERVER POOL':
                        value = value.split('\n')[-1]

                    result.append(value)

        result = ''.join(str(i) for i in result)
        result = ' - '.join(x for x in (custom_backend_prefix, result) if x)

        return result


def exit_call_summary(b_type, operation):
    handler = {
        EXIT_SQL: sqlparser.parse_sql_cmd,
        EXIT_NOSQL: sqlparser.parse_nosql_cmd,
        EXIT_REMOTE: sqlparser.parse_remote_cmd,
    }

    # to match our backend type and pb backend type, there is a formula: b_type/100,
    if PY3:
        return handler[b_type//100](b_type, operation)
    else:
        return handler[b_type/100](b_type, operation)


class Backend(object):
    """A backend is the destination of an exit call.

    Parameters
    ----------
    tipe : enum
        The type of the backend. more detail see pb_backend.proto
    identifying_properties : dict
        A dictionary of properties that uniquely identify this backend.
    name : str
        The display name of the backend.

    """
    __slots__ = ('tipe', 'identifying_properties', 'name', 'connType', 'be')

    def __init__(self, exit_point_type, identifying_properties, name):
        super(Backend, self).__init__()
        self.tipe = exit_point_type
        self.identifying_properties = identifying_properties
        self.name = name
        self.connType = None
        self.be = None

    def trans_to_dic(self):
        dic = {
            'type': self.tipe,
            'connType': str(self.connType),
            'name': self.name,
            'host': self.identifying_properties.get('HOST', 'unknown'),
            'port': int(self.identifying_properties.get('PORT', '0')),
            'url': str(self.identifying_properties.get('URL', self.name))
        }

        # Set database fields if interceptor has.
        for name in ('dbVendor', 'dbVersion', 'dbSchema'):
            if name in self.identifying_properties:
                dic[name] = str(self.identifying_properties[name])

        return dic

    def __str__(self):
        return self.name


class MethodData(object):
    __slots__ = ('original_method', 'args', 'kwargs', 'response', 'custom', 'cls_name', 'method_name', 'backend',
                 'operation', 'params', 'exc_info', '_summary', 'is_rpc', 'cross_head', 'rpc_be_call', 'response_method',
                 'sql_explain', 'start_time', 'end_time', 'transport_trace', 'tags', 'transport_be', 'be_options',
                 'be_call', 'exit_flag', 'error', '_mq_data', 'mq_direction', 'btr')
    def __init__(self, args, kwargs, cls, original_method, btr=None):
        # attach method
        self.original_method = original_method

        self.args = list(args)
        self.kwargs = kwargs
        self.btr = btr
        self.response = None
        self.custom = False

        # name
        self.cls_name = cls.__name__
        # original_method maybe a `function` or a `obj` with `__call__`
        self.method_name = getattr(original_method, '__name__', '__call__')

        # exit_call
        self.backend = None
        self.operation = None
        self.params = ''
        self._summary = None

        # rpc
        self.is_rpc = False
        self.cross_head = None
        self.rpc_be_call = None
        self.response_method = None

        # sql
        self.sql_explain = []

        # time
        self.start_time = self.time()
        self.end_time = None

        # trace
        self.transport_trace = None
        self.tags = {}

        # be
        self.transport_be = None
        self.be_options = None
        self.be_call = None
        self.exit_flag = False

        # error
        self.error = None
        self.exc_info = []

        # mq
        self._mq_data = None
        self.mq_direction = 0


    @property
    def mq_data(self):
        if self._mq_data:
            return self._mq_data
        else:
            return self.btr.mq_data

    @mq_data.setter
    def mq_data(self, value):
        self._mq_data = value

    @property
    def rpc_response_header(self):
        if not self.response_method:
            return {}

        return self.response_method(self.response)


    @property
    def summary(self):
        if self._summary:
            return self._summary
        else:
            return exit_call_summary(self.backend.tipe, self.operation)

    @summary.setter
    def summary(self, value):
        self._summary = value

    @property
    def method_name_parameter(self):
        try:
            return self.method_name + get_func_signa(self.original_method)
        except Exception:
            return self.method_name

    @property
    def line_number(self):
        try:
            return self.original_method.__code__.co_firstlineno
        except Exception:
            return 0

    @property
    def have_error(self):
        return True if self.error or self.exc_info else False

    @property
    def is_async(self):
        return current_thread_ident() != self.btr.concurrent_ident if self.btr else False

    def handle_rpc_http_status_code(self, response):
        status_code = getattr(response, 'status', None)
        if not isinstance(status_code, int):
            return

        if status_code < 400:
            return

        msg = 'HTTP %s' % status_code
        self.error = self.create_rpc_http_status_error(msg)

    def create_rpc_http_status_error(self, msg):
        error = {
            'errorName': msg,
            'errorSumary': msg,
            'errorDetails': msg,
        }
        return error

    @staticmethod
    def time():
        return set_time_unit(time.time())

    def need_explain(self, btr):
        if not enableSQLExplain():
            return False

        if btr.traceshot_svc:
            duration_time = self.time() - self.start_time
            return duration_time > config.getSqlExeTimeExplainThreshold() * 1000
        return False
