# Copyright (c) Bonree, Inc., and its affiliates
# 2015
# All Rights Reserved

"""Models and mappers for business transactions request.

Atttributes
-----------
ENTRY_WSGI : int, const
    A constant representing the WSGI entry point type.
ENTRY_FLASK : int, const
    A constant representing the Flask entry point type.
ENTRY_DJANGO : int, const
    A constant representing the Django entry point type.

Examples
--------
Transactions should be created using a transaction factory. To get a
transaction factory, you first need a timer factory such as the `timer` method
of an `Agent` instance:

    >>> from Bonree.agent.core.agent import Agent
    >>> agent = Agent()

Once you have the timer factory, create a transaction factory:

    >>> factory = make_transaction_req_factory(agent.timer)

You can now use `factory` to create transactions:

    >>> btr = factory()

The returned transaction has a timer that has not been started. You can start
the timer by calling `start` on it:

    >>> btr.timer.start()

To conveniently set attributes of the created transaction, pass keyword
arguments to the factory that correspond to the transaction's attributes:

    >>> btr = factory(name='my-custom-bt', registered_id=42)
    >>> btr.name
    'my-custom-bt'
    >>> btr.registered_id
    42

You can, of course, always set an attribute of a BT directly:

    >>> btr.entry_type = ENTRY_PYTHON_WSGI

"""
import threading
from uuid import uuid4

from Bonree import config
from Bonree.logger import BT_LOGGER
from Bonree.lang import values
from Bonree.lib import current_thread_ident
from Bonree.agent.proto import pb_type
from Bonree.agent.models.errors import ExceptionInfo, ErrorInfo

ENTRY_PYTHON_WEB = pb_type.BT_PYTHON_WEB
ENTRY_PYTHON_WSGI = ENTRY_PYTHON_WEB
ENTRY_PYTHON_BACKGROUND = pb_type.BT_BACKGROUND_TASK

HTTP_ERROR_DISPLAY_NAME = 'HTTP ERROR ({code})'

Entry_Name = {
    ENTRY_PYTHON_WEB: "ENTRY_PYTHON_WEB",
    ENTRY_PYTHON_WSGI: "ENTRY_PYTHON_WSGI",
    ENTRY_PYTHON_BACKGROUND: "ENTRY_PYTHON_BACKGROUND"
}


def get_entry_name(entry_type):
    return Entry_Name.get(entry_type, None)


def make_transaction_req_factory(agent, timer_factory):
    """Return a factory that creates transactions.

    Parameters
    ----------
    agent: Agent
    timer_factory : callable
        A callable with no required arguments that returns a timer object
        compatible with `Bonree.agent.core.timer.Timer`.

    Returns
    -------
    factory : callable
        A callable with no positional arguments that returns a new `TransactionReq`
        instance when called. Any keyword arguments to this callable are set as
        attributes of the returned transaction.

    """
    def factory(**kwargs):
        btr = TransactionReq(agent, timer_factory)

        for k, v in kwargs.items():
            setattr(btr, k, v)

        return btr

    return factory


class TransactionReq(object):
    """A business transaction request.

    Parameters
    ----------
    agent:
    timer_factory : callable
        A callable taking no parameters that returns a new instance of
        :py:class:`Bonree.agent.core.timer.Timer` when called

    Attributes
    ----------
    lock : threading.Lock
        A lock that used to synchronize access to the BT for the purposes of
        marking it completed and reporting it
    request_id : long
        A random 64-bit identifier for identifying requests concerning this BT.
    timer : Bonree.agent.core.timer.Timer
        A timer for keeping track of this BT's duration.
    completed : bool
        True if this TransactionReq has completed (i.e., `end_transaction_reqeust` has
        been called for it), even if the TransactionReq hasn't been sent.
    crossbt_hdr: CrossBTHeader
        Non-None if crossbt is discovered
    thread_exceptions : dict{thread_ident: errors.ExceptionInfo}
        A dict of uncaught exceptions that have been recorded during this BT.

    """
    __slots__ = ('agent', 'agent_id', 'name', 'entry_type', 'close_flag', 'lock', '_timer_factory', 'timer',
                 'concurrent_ident', 'status_code', 'request', 'response_headers', 'request_content', 'response_content',
                 'rpc_headers', 'cross_req_created', 'cross_resp', 'http_errors_count', 'exceptions_count',
                 'thread_exceptions', '_exception_set', 'http_errors', 'err_dict', 'detect_logs', 'traceshot_svc',
                 'cross_svc', 'mq_data', 'tags', 'transport_sdk', '__weakref__', '__inject_js__')
    def __init__(self, agent, timer_factory):
        self.agent = agent
        self.agent_id = 0
        self.name = None
        self.entry_type = None
        self.close_flag = False

        self.lock = threading.RLock()
        self._timer_factory = timer_factory
        self.timer = timer_factory()

        # Identification of the BT.
        self.concurrent_ident = current_thread_ident()
        self.status_code = None
        self.request = None
        self.response_headers = None
        self.request_content = None
        self.response_content = None
        self.transport_sdk = None

        #
        self.rpc_headers = None

        # cross bt
        self.cross_req_created = False  # a bt only need call `so.get_or_create_cross_req` once, so we set a flag.
        self.cross_resp = None # b'o:123;;'

        # Error and exception handling.
        self.http_errors_count = 0
        self.exceptions_count = 0
        self.thread_exceptions = {}
        self._exception_set = {}
        self.http_errors = []
        self.err_dict  = {}

        # Logs
        self.detect_logs = []

        # Trace
        self.traceshot_svc = None

        # Cross
        self.cross_svc = None

        # MQ
        self.mq_data = None

        # tags
        self.tags = {}

    @property
    def has_errors(self):
        return self.exceptions_count > 0 or self.http_errors_count > 0

    @property
    def has_exceptions(self):
        return self.exceptions_count > 0

    def add_http_error(self, status_code, msg):
        """Add an error to the BTR if the status code should trigger an error.
        If the status code is in the error config and enabled, or the status
        code is >= 400, create an ErrorInfo object and add it to the BTR.

        Parameters
        ----------
        status_code : errors code

        msg : Error describe string

        See Also
        --------
        add_exception
            For reporting uncaught exceptions that occurred during the BT

        """
        error_name = HTTP_ERROR_DISPLAY_NAME.format(code=status_code)

        if not config.getErrorIgnoreErrMsg(error_name):
            BT_LOGGER.debug("Get a ignore error: {0}".format(error_name))
            return

        http_error = ErrorInfo(error_name, msg)

        self.err_dict[http_error.name] = self.err_dict[http_error.name] + 1 if http_error.name in self.err_dict else 1
        self.http_errors_count += 1
        self.http_errors.append(http_error)


    def get_all_errors(self):
        result = []

        for excep in values(self.thread_exceptions):
            result.extend(excep)

        result.extend(self.http_errors)
        return result

    def _is_repeated_exception(self, exc_value, exceptin_set):
        """Return True if we have already added this exception to the BT.

        The reason for the recursive nature of this function is due to code I
        have seen code which traps exceptions and wraps them in other classes
        e.g.::

            try:
                raise InnerException('Stop that!')
            except InnerException as e:
                raise MoreGenericException('An exception happened!', e)

        """
        if exc_value in exceptin_set:
            return True
        for arg in exc_value.args:
            if isinstance(arg, Exception):
                return self._is_repeated_exception(arg, exceptin_set)
        return False

    def add_exception(self, exc_type, exc_value, exc_tb):
        """Add an exception to a BT.

        The arguments are those returned by sys.exc_info().

        Parameters
        ----------
        exc_type : type
            The type (class) of the raised exception
        exc_value
            The value of the raised exception
        exc_tb : traceback
            The traceback associated with the raised exception

        See Also
        --------
        add_http_error, add_logged_error
            For reporting errors logged during the BT

        """
        if exc_value is None:
            return

        # In case when a exception is raise. we can't get the http status code.
        if self.status_code is None:
            self.status_code = 500

        exception_info = ExceptionInfo(exc_type, exc_value, exc_tb)

        if not config.getErrorIgnoreExceptions(exception_info.name):
            BT_LOGGER.debug("Get a ignore exception: {0}".format(exception_info.name))
            return
        self.err_dict[exception_info.name] = self.err_dict[exception_info.name] + 1 if exception_info.name in self.err_dict else 1

        current_ident = current_thread_ident()

        exc_set = self._exception_set.setdefault(current_ident, set())

        if not self._is_repeated_exception(exc_value, exc_set):
            with self.lock:
                self.exceptions_count += 1
                exc_set.add(exc_value)
                thread_exceptions = self.thread_exceptions.setdefault(current_ident, [])
                thread_exceptions.append(exception_info)


    def __str__(self):
        return self.name