# -*- coding: utf-8 -*-
# Copyright (c) Bonree, Inc., and its affiliates
# 2015
# All Rights Reserved

"""Service for retrieving agent configuration on a regular basis.

"""
import os
import sys
import time
import threading

import Bonree
from Bonree.agent.services.wrapper import *
from Bonree.lib import dic_to_string
from Bonree.logger import SYS_LOGGER, SDK_LOGGER
from Bonree import config, get_version
from Bonree.agent.proto import pb_type


AGENT_INSTALL_DIR = os.path.dirname(os.path.realpath(os.path.join(Bonree.__file__, '..')))
# AGENT_INSTALL_DIR like "/opt/bonree/apm/agent/python/4.1.0-Alpha14"
APM_INSTALL_DIR = os.path.realpath(os.path.join(AGENT_INSTALL_DIR, '..', '..', '..'))
install_dir_stat = os.stat(AGENT_INSTALL_DIR)
AGENT_INSTALL_TIME = int(install_dir_stat.st_mtime * 1000)


class ConfigService():
    """A service that fetches and updates configuration.

    """
    def __init__(self, agent):
        self.agent = agent
        self.updateConfigResponseTs = time.time()
        self.sdk_is_open = False
        self.sdk_open_lock = threading.Lock()
        self.interval = 20


    def get_sdk_options(self):
        vm_info = ''
        vim_option = ''

        try:
            from distutils import sysconfig
            vm_info = '%d.%d.%d' % sys.version_info[:3]
            vim_option = dic_to_string(sysconfig.get_config_vars(), sep='\r\n')
        except Exception:
            pass

        options = SdkOptions()
        options.agent_type = pb_type.AGENT_PYTHON
        options.pid = os.getpid()
        # options.common_config_filepath.with_value(os.path.join(APM_INSTALL_DIR, 'conf', 'common.yml'))
        container = ContainerInfo()
        try:
            container.agent_version.with_value(get_version())
        except:
            container.agent_version.with_value(AGENT_INSTALL_DIR.split('/')[-1])
        container.agent_install_timestamp = AGENT_INSTALL_TIME
        container.agent_install_dir.with_value(AGENT_INSTALL_DIR)
        listens = config.getListens()
        SYS_LOGGER.info('container listens[%s]' % (listens))
        container.listens.with_value(listens)
        container.container_version.with_value(config.get_container_version() or '')
        container.container_install_dir.with_value('')
        container.vm_info.with_value(vm_info)
        container.vm_startup_options.with_value(self.agent.startup_arguments)
        container.vm_system_options.with_value(vim_option)
        options.container_info = pointer(container)
        options.log_cfg = SdkLogCfg()
        options.flags = 0
        return options

    def check_and_open_sdk(self):
        SDK_LOGGER.debug("check_and_open_sdk start")
        if self.sdk_is_open:
            return True
        SDK_LOGGER.info("check_and_open_sdk, open sdk now")
        open_result = False
        with self.sdk_open_lock:
            if self.sdk_is_open:
                # check again
                return True
            options = self.get_sdk_options()
            open_result = config.open_sdk(options)
            self.sdk_is_open = open_result
        SDK_LOGGER.info("check_and_open_sdk, open sdk result[%s]" % (open_result,))
        return open_result

    def update_config_response(self):
        nowTs = time.time()
        if nowTs - self.updateConfigResponseTs < self.interval:
            # update config response per `self.interval` second
            return
        SDK_LOGGER.info("update_config_response need update lastTs[%s] nowTs[%s]" % (self.updateConfigResponseTs, nowTs))

        enable_agent = False
        if config.updateSDKConfig(self.agent):
            isCircuitBreak = config.getCircuitBreakerBool()
            isAgentEnable = config.getAgentEnable()
            enable_agent = (not isCircuitBreak) and isAgentEnable
            SDK_LOGGER.info("update_config_response isCircuitBreak[%s] isAgentEnable[%s] enable_agent[%s]" % (isCircuitBreak, isAgentEnable, enable_agent))
        else:
            SDK_LOGGER.warning("update_config_response update SDKConfig Fail")
        config.set_enabled(enable_agent)
        if enable_agent:
            # when sdk enable. update config per 60 seconds
            self.interval = 60
        self.updateConfigResponseTs = nowTs
        return 0