# Copyright (c) Bonree, Inc., and its affiliates
# 2015
# All Rights Reserved

import random
import threading
import time

from Bonree import config
from Bonree.agent.models.cross import NEED_SAMPLE
from Bonree.agent.proto.pb_config import TRACE_MODE_SAMPLE, TRACE_MODE_SMART

class BtStartModeTrace(object):
    __slots__ = ('start_timing', 'traced_count')

    def __init__(self):
        self.start_timing = 0
        self.traced_count = 0

    def can_trace(self, now, n, period):
        # every `period` seconds can trace `n` call chains.
        if now - self.start_timing > period:
            self.start_timing = now
            self.traced_count = 1
            return True

        if self.traced_count < n:
            self.traced_count += 1
            return True

        return False


class StartModeFactory(object):
    def __init__(self):
        self.last_clean_timing = 0
        self.registered_names = {}
        self.lock = threading.Lock()

    def get_bt_smart_mode(self, bt_name):
        try:
            return self.registered_names[bt_name]
        except KeyError:
            bt_smart_mode = BtStartModeTrace()
            self.registered_names[bt_name] = bt_smart_mode
            return bt_smart_mode

    def bt_can_trace(self, bt_name, now, n, period):
        with self.lock:
            bt_smart_mode = self.get_bt_smart_mode(bt_name)
            return bt_smart_mode.can_trace(now, n, period)

    def clean_expire(self, now=None, period=5):
        if not self.registered_names:
            return
        if now is None:
            now = time.time()
        if now - start_mode_helper.last_clean_timing < 90:
            return

        names = {}
        for k in self.registered_names:
            bt_smart_mode = self.registered_names[k]
            if now - bt_smart_mode.start_timing > period:
                continue
            names[k] = bt_smart_mode
        self.last_clean_timing = now
        self.registered_names = names


start_mode_helper = StartModeFactory()


def make_traceshot_server(trace_header, transport_sdk, bt_name):
    fetch_trace = False
    now = None
    if (trace_header and trace_header.get(NEED_SAMPLE) == '1'):
        fetch_trace = True
    elif config.getTraceEnable():
        trace_mode = config.getTraceMode()
        # 1 TRACE_MODE_SAMPLE
        # 2 TRACE_MODE_THRESHOLD (deprecated proto7.1.0)
        # 3 TRACE_MODE_SMART
        if trace_mode == TRACE_MODE_SAMPLE :
            if random.randint(0, 999) < config.getTraceDefaultGathererPercent():
                fetch_trace = True
        elif trace_mode == TRACE_MODE_SMART:
            level = config.getSmartModeLevel()
            now = time.time()
            period = 5
            fetch_trace = start_mode_helper.bt_can_trace(bt_name, now, level, period)

    start_mode_helper.clean_expire(now)

    if fetch_trace:
        transport_sdk.set_tracer()
    return fetch_trace


component_dict = {
    # Entry points
    'bottle': 'Bottle',
    'flask': 'Flask',
    'django': 'Django',
    'cherrypy': 'CherryPy',
    'tornado': 'Tornado',
    'pyramid': 'Pyramid',
    'odoo': 'Odoo',
    'openerp': 'Odoo',
    'web': 'Web.py',
    'sanic': 'Sanic',
    'apistar': 'Apistar',
    'quart': 'Quart',
    'starlette': 'Starlette',

    # HTTP exit calls
    'httplib': 'httplib',
    'urllib3': 'urllib',
    'requests': 'Requests',
    'kombu': 'Kombu',
    'aiohttp': 'Aiohttp',

    # RPC
    'grpc': 'Grpc',
    'stompy': 'ActiveMQ',
    'stomp': 'ActiveMQ',
    'kafka': 'Kafka',

    # SQL exit calls
    'psycopg2': 'PostgreSQL',
    'pymysql': 'MySQL',
    'mysql': 'MySQL',
    'MySQLdb': 'MySQL',
    'cx_Oracle': 'Oracle',

    # Caches
    'memcache': 'Memcache',
    'redis': 'Redis',
    'pyssdb': 'SSDB',

    # MongoDB
    'pymongo': 'Mongodb',

    #Other
    'Bonree': {'psycopg2': 'PostgreSQL', 'cx_oracle': 'Oracle'}
}


def module_name_to_component_name(module_name):
    if module_name is None:
        return None

    module_name_str = module_name.split('.')[0]

    if module_name_str == 'Bonree':
        module_name_str = module_name.split('.')[-1]
        if module_name_str in component_dict['Bonree']:
            return component_dict['Bonree'][module_name_str]

    if module_name_str in component_dict:
        return component_dict[module_name_str]
    return None


def module_name_to_component(module_name):
    """deprecated"""
    if module_name is None:
        return {}
        
    module_name_str = module_name.split('.')[0]

    if module_name_str == 'Bonree':
        module_name_str = module_name.split('.')[-1]
        if module_name_str in component_dict['Bonree']:
            return {'component': component_dict['Bonree'][module_name_str]}

    if module_name_str in component_dict:
        return {'component': component_dict[module_name_str]}
    return {}