# Copyright (c) Bonree, Inc., and its affiliates
# 2015
# All Rights Reserved
import threading

from Bonree import config
from Bonree.agent.models.errors import ErrorInfo
from Bonree.agent.services.cross import br_cs_req_from_header
from Bonree.lang import str
from Bonree.lib import current_thread_ident, args2str, now, arr_http_header, arr_http_parameters
from Bonree.logger import BT_LOGGER
from Bonree.agent.models.transactions import ENTRY_PYTHON_WEB, ENTRY_PYTHON_WSGI
from Bonree.agent.services.wrapper import *

class TransportSDK(object):
    __slots__ = ('so', 'transaction', 'bt', 'entry_type', 'tracer', 'tracer_node', )
    def __init__(self):
        self.so = config.get_sdk_so()
        self.transaction = None
        self.bt =Instance()
        self.entry_type = None
        self.tracer = None
        self.tracer_node = None

    def is_http_transaction(self):
        return self.entry_type == ENTRY_PYTHON_WEB

    def start_bt_transaction(self, name, entry_type, request=None):
        self.transaction = Instance()
        self.so.open_transaction(pointer(self.transaction))
        bt_options = BtOptions()
        bt_options.type = entry_type
        bt_options.name = bt_options.internal_name = str_cache(name)
        self.entry_type = entry_type
        if self.is_http_transaction() and request is not None:
            bt_options.uri = str_cache(request.path)

        self.so.get_or_create_bt(pointer(self.transaction), pointer(bt_options), POINTER(HttpReqData)(), pointer(self.bt))

    def end_sdk_transaction(self, btr):
        btr.traceshot_svc = False
        request = btr.request
        if request is not None and self.is_http_transaction():
            http_req = HttpReqData()
            http_req.url.with_value(request.url)
            http_req.http_method.with_value(request.method)
            if config.getTraceCollectHttpReqHeadersEnable() and request.headers:
                http_req.req_headers.with_value(arr_http_header(request.headers))
            if config.getTraceEnableCollectHttpParams():
                http_req.parameters.with_value(arr_http_parameters(request.args))
            # if btr.request_content:
            #     length = config.getTraceCollectRequestContentLen()
            #     if length:
            #         http_req.req_content =  Str().with_value(btr.request_content[:length])

            self.so.bt_set_http_req(pointer(self.bt), pointer(http_req))

        finish_bt_option = FinishBtOptions()
        if btr.has_exceptions:
            finish_bt_option.flags = 2
        if self.is_http_transaction():
            http_resp = HttpRespData()
            http_resp.status_code = str_cache(btr.status_code)
            http_resp.res_headers.with_value(arr_http_header(btr.response_headers))
            # if btr.response_content:
            #     length = config.getTraceCollectResponseContentLen()
            #     if length:
            #         http_resp.resp_content.with_value(btr.response_content[:length])
            http_resp = pointer(http_resp)
        else:
            http_resp = POINTER(HttpRespData)()

        if self.tracer:
            for k, v in btr.tags.items():
                if not v or str(v) == '0':
                    continue
                self.so.set_trace_tag(pointer(self.tracer), pointer(str_cache(k)), pointer(str_cache(v)))
            self.so.set_trace_tag(pointer(self.tracer), pointer(str_cache('http.status_code')), pointer(str_cache(btr.status_code)))
            if self.tracer_node:
                self.set_trace_node_finish()
            self.so.finish_tracer(pointer(self.tracer))
            self.tracer = None

        self.make_errors(btr)


        self.so.finish_bt(pointer(self.bt), pointer(finish_bt_option), http_resp)
        self.so.close_transaction(pointer(self.transaction))
        btr.close_flag = True

    def set_trace_node_finish(self):
        node = self.tracer_node
        finish_span_options = FinishSpanOptions()
        finish_span_options.finish_time_stamp = now()
        while node:
            if not node.finished:
                self.so.finish_span(pointer(node.span), pointer(finish_span_options))
                node.finished = True
            node = node.next_node
        self.tracer_node = None

    def set_tracer(self):
        self.tracer = Instance()
        self.so.get_or_create_tracer(pointer(self.transaction), pointer(self.tracer))

    def make_errors(self, btr):
        error_list = btr.get_all_errors()
        if error_list:
            for error in error_list:
                bt_error = BtError()
                bt_error.time_stamp = error.time
                bt_error.error_name.with_value(error.name)
                bt_error.error_sumary.with_value(error.sumary)
                bt_error.error_details.with_value(error.detail)
                if isinstance(error, ErrorInfo):
                    bt_error.error_type = 1
                    bt_error.error_type_str.with_value("ERROR_TYPE_HTTP")
                else:
                    bt_error.error_type = 2
                    bt_error.error_type_str.with_value("ERROR_TYPE_EXCEPTION")

                self.so.add_bt_error(pointer(self.bt), pointer(bt_error))


class TransportBe(object):
    __slots__ = ('so', 'transport_sdk', 'transaction', )
    def __init__(self):
        self.so = config.get_sdk_so()


    def start_be_call(self, md):
        btr = md.btr
        self.transport_sdk = btr.transport_sdk if btr else None
        self.transaction = self.transport_sdk.transaction if self.transport_sdk else None

        if not self.transaction:
            self.transaction = Instance()
            self.so.open_transaction(pointer(self.transaction))

        if btr and btr.close_flag:
            return

        be_options = BeCallOptions()
        if md.backend.be:
            be_options.be = md.backend.be
        else:
            self.make_be_call(md.backend, be_options.be)
        # class_method
        # be_options.class_method = ClassMethod()
        be_options.class_method.class_name = str_cache(md.cls_name)
        be_options.class_method.method_name = str_cache(md.method_name)

        # call_detail
        # be_options.call_detail = BeCallDetail()
        be_options.call_detail.detail.with_value(md.operation)
        be_options.call_detail.is_async = int(md.is_async)

        be_options.start_time_stamp = md.start_time

        be_call = Instance()
        self.so.start_be_call(pointer(self.transaction), pointer(be_options), pointer(be_call))
        md.be_options = be_options
        md.be_call = be_call
        # return self.be_call

    def update_be_call(self, md):
        btr = md.btr
        if (btr and btr.close_flag) or not md.be_call:
            return
        be_options = md.be_options
        if be_options is None:
            return
        be_options.call_detail.detail.with_value(md.operation)
        if md.be_call:
            self.so.update_be_call(pointer(md.be_call), pointer(be_options))
        if md.rpc_be_call:
            self.so.update_be_call(pointer(md.rpc_be_call), pointer(be_options))


    def end_be_call(self, md):
        btr = md.btr
        if (btr and btr.close_flag) or not md.be_call:
            return

        be_finish_options = FinishBeCallOptions()
        be_finish_options.flags = int(md.have_error)

        if md.error:
            be_finish_options.error_name.with_value(md.error.get('errorName'))
            be_finish_options.error_detail.with_value(md.error.get('errorSumary'))

        if md.sql_explain:
            be_finish_options.sql_explain_plan.with_value(' '.join(md.sql_explain))

        be_finish_options.finish_time_stamp = md.end_time

        if btr:
            if md.mq_data:
                mq_info = MQInfo()
                mq_info.mq_size = len(md.mq_data)
                mq_info.direction = md.mq_direction
                self.so.set_mq_info(pointer(md.be_call), pointer(mq_info))
                btr.mq_data = None
            self.so.finish_be_call(pointer(md.be_call), pointer(be_finish_options))
        else:
            self.so.finish_be_call(pointer(md.be_call), pointer(be_finish_options))

    def start_be_rpc(self, md):
        btr = md.btr
        self.transport_sdk = btr.transport_sdk if btr else None
        self.transaction = self.transport_sdk.transaction if self.transport_sdk else None

        if btr and btr.close_flag:
            return

        cross_req_head = CrossReqHead()
        if btr.request and btr.request.headers:
            req_headers = Str().with_value(br_cs_req_from_header(btr.request.headers))
            self.so.get_cross_req_head_from_head(pointer(req_headers), pointer(cross_req_head))

        cross_req_options = CrossReqOptions()
        cross_req_options.is_sampling = int(config.getEnableTrackRemoteCall())
        cross_req_options.cross_req = cross_req_head

        if not btr.cross_req_created:
            cross_req = Instance()
            self.so.get_or_create_cross_req(pointer(self.transaction), pointer(cross_req_options), pointer(cross_req))
            btr.cross_req_created = True

        rpc_be_options = BeCallOptions()
        if md.backend.be:
            rpc_be_options.be = md.backend.be
        else:
            self.make_be_call(md.backend, rpc_be_options.be)

        # rpc_be_options.class_method = ClassMethod()
        rpc_be_options.class_method.class_name.with_value(md.cls_name)
        rpc_be_options.class_method.method_name.with_value(md.method_name)

        # rpc_be_options.call_detail = BeCallDetail()
        rpc_be_options.call_detail.detail.with_value(md.operation)
        rpc_be_options.call_detail.is_async = int(md.is_async)

        md.rpc_be_call = Instance()
        md.be_options = rpc_be_options
        self.so.start_rpc(pointer(self.transaction), pointer(rpc_be_options), pointer(cross_req_head),
                          pointer(md.rpc_be_call))
        md.cross_head = cross_req_head.data
        BT_LOGGER.debug('Make trace request header: {0}'.format(cross_req_head.data))
        # return self.rpc_be_call


    def end_be_rpc(self, md):
        if (md.btr and md.btr.close_flag) or not md.rpc_be_call:
            return

        rpc_be_finish_options = FinishBeCallOptions()
        rpc_be_finish_options.flags = int(md.have_error)

        resp_headers = Str().with_value(arr_http_header(md.rpc_response_header))
        cross_req_head = CrossReqHead()
        self.so.get_cross_resp_head_from_head(pointer(resp_headers), pointer(cross_req_head))
        self.so.finish_rpc(pointer(self.transaction), pointer(md.rpc_be_call), pointer(rpc_be_finish_options),
                           pointer(cross_req_head))


    def make_be_call(self, backend, be):
        be.type = int(backend.tipe)
        be.conn_type.with_value(backend.connType)
        be.name.with_value(backend.name)
        be.host.with_value(backend.identifying_properties.get('HOST', 'unknown'))
        be.port = int(backend.identifying_properties.get('PORT', '0'))
        be.url.with_value(backend.identifying_properties.get('URL', backend.name))

        if 'dbVendor' in backend.identifying_properties:
            be.db_vendor.with_value(backend.identifying_properties['dbVendor'])
        if 'dbVersion' in backend.identifying_properties:
            be.db_version.with_value(backend.identifying_properties['dbVersion'])
        if 'dbSchema' in backend.identifying_properties:
            be.db_schema.with_value(backend.identifying_properties['dbSchema'])


class TransportTrace(object):
    __slots__ = ('so', 'transaction', 'tracer', 'finished', 'next_node', 'transport_sdk',
                 'span', )
    def __init__(self):
        self.so = config.get_sdk_so()
        self.transaction = None
        self.tracer = None
        self.finished = False
        self.next_node = None

    def trace_id_for_cross_server(self, btr):
        # get trace id from sdk
        trace_id_struct = TraceId()
        if self.so.get_trace_id(pointer(self.tracer), pointer(trace_id_struct)) == 0:
            btr.cross_svc.trace_id = trace_id_struct.data.decode('latin1')  # `bytes` to `str`

    def start_trace_call(self, md):
        btr = md.btr
        self.transport_sdk = btr.transport_sdk if btr else None
        self.transaction = self.transport_sdk.transaction if self.transport_sdk else None

        if not self.transaction:
            return

        self.tracer = self.transport_sdk.tracer
        if not self.tracer:
            return

        self.next_node = btr.transport_sdk.tracer_node
        btr.transport_sdk.tracer_node = self

        span_options = StartSpanOptions()
        # span_options.thread_info = SpanThreadInfo()
        # span_options.method_info = SpanMethodInfo()
        span_options.start_time_stamp = md.start_time

        span_options.thread_info.thread_name = str_cache(threading.current_thread().name)
        span_options.thread_info.thread_id = current_thread_ident()
        # method_info
        span_options.method_info.class_name = str_cache(md.cls_name)
        span_options.method_info.method_name.with_value(md.method_name_parameter)


        # span_options.method_info.ret_value = str_cache('')
        # span_options.method_info.params.with_value(','.join(args2str(md.method_name, *md.args, **md.kwargs)))
        span_options.method_info.line_number = int(md.line_number)

        self.span = Instance()
        self.so.start_span(pointer(self.tracer), pointer(span_options), pointer(self.span))

        if btr.cross_svc.trace_id is None:
            self.trace_id_for_cross_server(btr)

        if md.be_call:
            self.so.link_to_be_call(pointer(self.span), pointer(md.be_call))
        elif md.rpc_be_call:
            self.so.link_to_be_call(pointer(self.span), pointer(md.rpc_be_call))

    def end_trace_call(self, md):
        if not self.transaction:
            return
        if not self.tracer:
            return
        btr = md.btr
        if not btr:
            return

        if btr.close_flag:
            return

        if self.finished:
            return

        if md.error:
            trace_error = TraceError()
            trace_error.error_summary.with_value(md.error.get('errorSumary'))
            trace_error.error_details.with_value(md.error.get('errorDetails'))
            self.so.set_trace_error(pointer(self.span), pointer(trace_error))

        if btr.detect_logs:
            for log_dict in btr.detect_logs:
                trace_log = TraceLog()
                trace_log.level.with_value(log_dict.get('level'))
                trace_log.log_msg.with_value(log_dict.get('logMsg'))
                self.so.add_trace_log(pointer(self.tracer), pointer(trace_log))

        for k, v in md.tags.items():
            if not v or str(v) == '0':
                continue
            self.so.set_tag(pointer(self.span), pointer(str_cache(k)), pointer(str_cache(v)))
        if md.end_time is None:
            md.end_time = now()
        finish_span_options = FinishSpanOptions()
        finish_span_options.finish_time_stamp = md.end_time
        self.so.finish_span(pointer(self.span), pointer(finish_span_options))
        self.finished = True
