import os
import re
import sys
import logging
import platform
import socket

import Bonree
from Bonree.agent.proto.pb_config import HTTP_CODE_TYPE_ERROR
from Bonree.agent.proto.pb_type import BT_PYTHON_WEB
from Bonree.agent.services.wrapper import *
from Bonree.lib import LazyWsgiRequest, ms_to_us, TimeCache
from Bonree.lang import values, bytes_t, range, str
from Bonree.confparser import ConfigParser

on_off = lambda v: str(v).lower() in ('on', 'true', 'yes', 'y', 't', '1')

AGENT_INSTALL_DIR = os.path.dirname(os.path.realpath(os.path.join(Bonree.__file__, '..')))
# AGENT_INSTALL_DIR like "/opt/bonree/apm/agent/python/4.1.0-Alpha14"
APM_INSTALL_DIR = os.path.realpath(os.path.join(AGENT_INSTALL_DIR, '..', '..', '..'))
DEFAULT_HOME_DIR = os.path.join(APM_INSTALL_DIR, 'logs', 'python')
PY3 = int(sys.version.split('.')[0]) > 2

_CONFIG_OPTIONS = {
    # log
    'enable': ('LOGGING_ENABLE', on_off),
    'dir': ('LOGS_DIR', None),
    'level': ('LOGGING_LEVEL', None),
    'process': ('LOG_PROCESS', on_off),
    # wsgi
    'script': ('WSGI_SCRIPT_ALIAS', None),
    'callable': ('WSGI_CALLABLE_OBJECT', None),
    'module': ('WSGI_MODULE', None),
    # js
    'js_site': ('JS_SITE', on_off),
    # bonree_log
    'bonree_log': ('BONREE_LOG', on_off)
}


# Logging
LOGS_DIR = None
LOGGING_ENABLE = True
LOGGING_LEVEL = 'INFO'
LOG_PROCESS = False

# WSGI
WSGI_MODULE = None
WSGI_SCRIPT_ALIAS = None
WSGI_CALLABLE_OBJECT = None

# JS
JS_SITE = False

# bonree_log
BONREE_LOG = False



# ----------------------------------------
# Interval time
TRANSPORT_INTERVAL_S = 1
RELOAD_CONFIGURATION_INTERVAL_S = 120
SNAPSHOT_PROFILER_INTERVAL_S = 0.001

# logger
LOGGING_MAX_BYTES = 20000000
LOGGING_MAX_NUM_FILES = 5


# -----------------------------------------
STATE_CFG_CHANGED = 0x1

# .SO
SO = None

SO_PATH = os.environ.get('BONREE_SMARTAGENT_SDK_PATH', None)

# config
SO_CONFIG = Instance()

APP_CONFIG = Instance()
ERROR_CONFIG = Instance()
DEBUG_CONFIG = Instance()
BACKEND_CONFIG = Instance()
JS_CONFIG = Instance()
CALL_GRAPH_CONFIG = Instance()
APP_GATHER_CONFIG = Instance()
SNAPSHOT_GATHER_CONFIG = Instance()
BT_CONFIG = Instance()
BT_TRANS_CONFIG = BTThresholdConfig()
TRACE_CONFIG = Instance()
SOCKET_CONFIG = Instance()
CIRCUIT_BREAKER_CONFIG = Instance()

agent_enable = False

def get_enabled():
    return agent_enable

def set_enabled(value):
    global agent_enable
    if agent_enable != value:
        from Bonree.logger import SDK_LOGGER
        SDK_LOGGER.info("Switch agent enable: [{0}]->[{1}]".format(agent_enable, value))
        agent_enable = value

def response_to_bool(response):
    return response == 0

def open_sdk(options):
    if not SO_PATH:
        from Bonree.logger import SDK_LOGGER
        SDK_LOGGER.error("Not available SDK path")
        return False

    global SO
    if SO is None:
        SO = CDLL(SO_PATH)
    sdk_response = SO.open_sdk(pointer(options))
    is_success = response_to_bool(sdk_response)
    if not is_success:
        from Bonree.logger import SDK_LOGGER
        SDK_LOGGER.error("open_sdk fail sdk_response[%s]" % (sdk_response))
    return is_success



def get_sdk_so():
    return SO

CALL_BACK_FUNC = None
def sdk_set_callback_fuc(method):
    if SO:
        global CALL_BACK_FUNC
        FUNC = CFUNCTYPE(c_int, POINTER(AgentState))
        CALL_BACK_FUNC = FUNC(method)
        SO.reg_agent_state_callback(CALL_BACK_FUNC)
        return True
    else:
        return False

def update_cfg_obj():
    return response_to_bool(SO.get_config(pointer(SO_CONFIG)))

# AgentConfig
def getAgentEnable():
    return bool(SO.agent_enable(pointer(SO_CONFIG)))

@TimeCache()
def getAgentId():
    agent_id = SO.get_agent_id(pointer(SO_CONFIG))
    if agent_id > 0:
        return agent_id
    return None

# ErrorConfig
def updateErrorConfig():
    return SO.get_error_config(pointer(SO_CONFIG), pointer(ERROR_CONFIG))

def getErrorIgnoreExceptions(err_name):
    response = SO.ignore_exception(pointer(ERROR_CONFIG), pointer(Str().with_value(err_name)))
    return response_to_bool(response)

def getErrorIgnoreErrMsg(err_name):
    response = SO.ignore_err_msg(pointer(ERROR_CONFIG), pointer(Str().with_value(err_name)))
    return response_to_bool(response)


ERROR_MSG = Str().with_value('')
ERROR_TYPE = c_uint(0)
def getErrorCusHttpErrors(status_code):
    status_code = c_uint(status_code)
    response = SO.get_custom_http_code_desc(pointer(ERROR_CONFIG), status_code, pointer(ERROR_MSG), pointer(ERROR_TYPE))
    return response_to_bool(response)

def getErrorMsg(err_code):
    if getErrorCusHttpErrors(err_code):
        msg = str(ERROR_MSG.value)
        err_type = ERROR_TYPE.value
        if not err_type:
            err_type = HTTP_CODE_TYPE_ERROR
        freeErrorCusHttpErrors()
        return msg, err_type
    return None, None

def freeErrorCusHttpErrors():
    return response_to_bool(SO.free_custom_http_code_desc(pointer(ERROR_MSG)))

@TimeCache()
def getPythonEnableDetectLogging():
    if not SO:
        return False
    return bool(SO.detect_logging_enable(pointer(ERROR_CONFIG)))

@TimeCache()
def getPythonEnableDetectLogBook():
    return bool(SO.detect_log_book_enable(pointer(ERROR_CONFIG)))

@TimeCache()
def getPythonEnableDetectEliot():
    return bool(SO.detect_eliot_enable(pointer(ERROR_CONFIG)))

# DebugConfig
def updateDebugConfig():
    return SO.get_debug_config(pointer(SO_CONFIG), pointer(DEBUG_CONFIG))

def getDebugConfigEnable():
    return bool(SO.debug_enable(pointer(DEBUG_CONFIG)))

def getDebugConfigEnableFullLog():
    return bool(SO.full_log_enable(pointer(DEBUG_CONFIG)))


# AppConfig
def updateAppConfig():
    return SO.get_app_config(pointer(SO_CONFIG), pointer(APP_CONFIG))

def getAppInstrumentLevel():
    # 1 PRODUCTION  2 DEVELOPMENT
    return SO.get_instrument_level(pointer(APP_CONFIG))

@TimeCache()
def getEnableDataAggregate():
    return bool(SO.data_aggregate_enable(pointer(APP_CONFIG)))

def updateNamingConfigs(agent):
    bt_naming = BtNaming()
    SO.get_bt_naming_cfg(pointer(APP_CONFIG), BT_PYTHON_WEB, pointer(bt_naming))
    if bool(bt_naming.enable):
        rules = [{'type': bt_naming.type, 'properties': str(bt_naming.properties.value)}]
        agent.naming_registry.update(rules)
    SO.free_bt_naming_cfg(pointer(bt_naming))


# BackendConfig
def updateBackendConfig():
    return SO.get_backend_config(pointer(SO_CONFIG), pointer(BACKEND_CONFIG))

def getSqlRecordType():
    return SO.get_sql_record_type(pointer(BACKEND_CONFIG))

def getSqlExeTimeThreshold():
    return SO.get_sql_exe_time_threshold(pointer(BACKEND_CONFIG))

@TimeCache()
def getSqlExeTimeExplainThreshold():
    return SO.get_sql_exe_time_explain_threshold(pointer(BACKEND_CONFIG))

@TimeCache()
def getSqlMaxLength():
    if SO:
        return SO.get_sql_max_length(pointer(BACKEND_CONFIG))

@TimeCache()
def getSqlMaxRecords():
    return SO.get_sql_max_records(pointer(BACKEND_CONFIG))

def getNosqlExeTimeThreshold():
    return SO.get_nosql_exe_time_threshold(pointer(BACKEND_CONFIG))

@TimeCache()
def getEnableCaptureNosqlParams():
    if SO:
        return SO.enable_capture_nosql_params(pointer(BACKEND_CONFIG))

def getNosqlMaxParamLength():
    if SO:
        return SO.get_nosql_max_param_length(pointer(BACKEND_CONFIG))

def getRemoteCallExeTimeThreshold():
    return SO.get_remote_call_exe_time_threshold(pointer(BACKEND_CONFIG))

def getRemoteCallGetURISegmentType():
    return SO.get_remote_call_get_uri_seg_type(pointer(BACKEND_CONFIG))

def getRemoteCallURISegments():
    return SO.get_remote_call_uri_segments(pointer(BACKEND_CONFIG))


# JsConfig
def updateJsConfig():
    return SO.get_inject_js_config(pointer(SO_CONFIG), pointer(JS_CONFIG))

JS_CODE = Str()
@TimeCache()
def getgetJsConfigJs():
    if not SO:
        return False
    return response_to_bool(SO.get_inject_js(pointer(JS_CONFIG), pointer(JS_CODE)))

def freeJsConfigJs():
    SO.free_inject_js(pointer(JS_CODE))

@TimeCache()
def getJsCode():
    js_code = str(JS_CODE.value)
    freeJsConfigJs()
    return js_code


# CallGraphConfig
def updateCallGraphConfig():
    resp = SO.get_call_graph_config(pointer(SO_CONFIG), pointer(CALL_GRAPH_CONFIG))
    min_method_time = getMinMethodTime()
    if min_method_time:
        new_period = min_method_time / 1000000.0 / 2.0
        if new_period > 0.001:
            global SNAPSHOT_PROFILER_INTERVAL_S
            SNAPSHOT_PROFILER_INTERVAL_S = new_period
    return resp

def getMinMethodTime():
    return ms_to_us(SO.get_method_time_threshold(pointer(CALL_GRAPH_CONFIG)))

def getMinSQLQueryTime():
    return ms_to_us(SO.get_sql_query_time_threshold(pointer(CALL_GRAPH_CONFIG)))


def getExcludeNames(name):
    name = Str().with_value(name)
    return bool(SO.in_exclude_names(pointer(CALL_GRAPH_CONFIG), pointer(name)))

def getIncludeNames(name):
    name = Str().with_value(name)
    return bool(SO.in_include_names(pointer(CALL_GRAPH_CONFIG), pointer(name)))

# AppGatherConfig
def updateAppGatherConfig():
    return SO.get_app_gatherer_config(pointer(SO_CONFIG), pointer(APP_GATHER_CONFIG))

def getEnableTrackBT():
    return bool(SO.track_bt_enable(pointer(APP_GATHER_CONFIG)))

def getEnableCollectErrorAndException():
    return bool(SO.collect_error_and_exception_enable(pointer(APP_GATHER_CONFIG)))

def getEnableTrackSQL():
    return bool(SO.track_sql_enable(pointer(APP_GATHER_CONFIG)))

def getEnableTrackNoSQL():
    return bool(SO.track_nosql_enable(pointer(APP_GATHER_CONFIG)))

@TimeCache()
def getEnableTrackRemoteCall():
    return SO.track_remote_call_enable(pointer(APP_GATHER_CONFIG))

def getEnableTrackAsyncThread():
    return bool(SO.track_async_thread_enable(pointer(APP_GATHER_CONFIG)))

def getGathererPercent():
    return SO.get_gatherer_percent(pointer(APP_GATHER_CONFIG))


# SnapshotGatherConfig
HOT_METHOD_ARG_TRUE  = []
HOT_METHOD_LIST  = []
def updateSnapshotGatherConfig():
    resp = SO.get_snapshot_gatherer_config(pointer(SO_CONFIG), pointer(SNAPSHOT_GATHER_CONFIG))

    global HOT_METHOD_ARG_TRUE, HOT_METHOD_LIST
    HOT_METHOD_ARG_TRUE  = []
    HOT_METHOD_LIST  = []
    for hot in getHotspotMethods():
        class_name = getHotspotMethodClassName(hot)
        methods_name = getHotspotMethodNames(hot)

        HOT_METHOD_LIST.append({'className': class_name, 'methodNames': methods_name})

    # hook hotshot
    from Bonree.agent.interceptor.custommethod.hotspot import hotshot_hook
    hotshot_hook(HOT_METHOD_LIST, SO, CALL_GRAPH_CONFIG)

    freeHotspotMethods(HOT_METHOD, HOT_METHOD_SIZE)
    return resp

def getHotMethodList():
    return HOT_METHOD_LIST

def getExceedThresholdPercent():
    return SO.get_exceed_threshold_percent(pointer(SNAPSHOT_GATHER_CONFIG))

def getExceedErrorPercent():
    return SO.get_exceed_error_percent(pointer(SNAPSHOT_GATHER_CONFIG))

def getSnapshotCountPerMin():
    return SO.get_snapshot_count_per_min(pointer(SNAPSHOT_GATHER_CONFIG))

def getSnapshotMin():
    return SO.get_snapshot_min(pointer(SNAPSHOT_GATHER_CONFIG))

def getSnapshotInterval():
    return SO.get_snapshot_interval(pointer(SNAPSHOT_GATHER_CONFIG))

def getSnapshotTryTimesPerMin():
    return SO.get_snapshot_try_times_per_min(pointer(SNAPSHOT_GATHER_CONFIG))

def getSnapshotPerMinPerTransaction():
    return SO.get_snapshot_per_min_per_transaction(pointer(SNAPSHOT_GATHER_CONFIG))

def getSnapshotPercent():
    return SO.get_snapshot_percent(pointer(SNAPSHOT_GATHER_CONFIG))

HOT_METHOD = POINTER(HotMethodcfg)()
HOT_METHOD_SIZE = c_int(0)
def getHotspotMethods():
    if response_to_bool(SO.get_hot_spot_methods(pointer(SNAPSHOT_GATHER_CONFIG), pointer(HOT_METHOD), pointer(HOT_METHOD_SIZE))):
        hot  = []
        for i in range(HOT_METHOD_SIZE.value):
            hot.append(HOT_METHOD[i])
        return hot

def getHotspotMethodClassName(method):
    return str(method.class_name.value)

def getHotspotMethodNames(method):
    names = method.method_names
    names_size = method.method_name_size
    enables = method.enable_collect_params
    names_list  = []
    for i in range(names_size):
        names_list.append(str(names[i].value))
        if bool(enables[i]):
            HOT_METHOD_ARG_TRUE.append(str(names[i].value))
    return names_list

def freeHotspotMethods(hot_methods, size):
    return response_to_bool(SO.free_hot_spot_methods(hot_methods, size))

def getEnableAllSnapshot():
    return bool(SO.all_snapshot_enable(pointer(SNAPSHOT_GATHER_CONFIG)))

# BtConfig
def updateBtConfig():
    return SO.get_bt_configs(pointer(SO_CONFIG), pointer(BT_CONFIG))

def updateBtFilterConfig():
    return SO.get_bt_filter_config(pointer(SO_CONFIG), pointer(BT_CONFIG))

def getFilterURIPatterns(url):
    return bool(SO.in_bt_blacklist(pointer(BT_CONFIG), pointer(Str().with_value(url))))

def getSnapshotURIPatterns(url):
    return bool(SO.in_bt_whitelist(pointer(BT_CONFIG), pointer(Str().with_value(url))))

def getIncludeURIPatterns(url):
    return bool(SO.In_snapshot_blacklist(pointer(BT_CONFIG), pointer(Str().with_value(url))))


# BtTransConfig
def updateBtTransConfig():
    return SO.get_default_bt_cfg(pointer(SO_CONFIG), pointer(BT_TRANS_CONFIG))

def getBtTransDefaultBtThreshold():
    return ms_to_us(BT_TRANS_CONFIG.bt_threshold)

def getBtTransDefaultBtThresholdSlow():
    return ms_to_us(BT_TRANS_CONFIG.bt_threshold_slow)

def getBtTransDefaultBtThresholdVerySlow():
    return ms_to_us(BT_TRANS_CONFIG.bt_threshold_very_slow)

BT_TRANS_THRESHOLD = BTThresholdConfig()
def getBtTransBtCfgs(name):
    return response_to_bool(SO.get_bt_cfg(pointer(BT_TRANS_CONFIG), pointer(Str().with_value(name)), pointer(BT_TRANS_THRESHOLD)))

def getBtTransBtThreshold():
    return ms_to_us(BT_TRANS_THRESHOLD.bt_threshold)

def getBtTransBtThresholdSlow():
    return ms_to_us(BT_TRANS_THRESHOLD.bt_threshold_slow)

def getBtTransBtThresholdVerySlow():
    return ms_to_us(BT_TRANS_THRESHOLD.bt_threshold_very_slow)


# TraceConfig
def updateTraceConfig():
    return SO.get_trace_config(pointer(SO_CONFIG), pointer(TRACE_CONFIG))

@TimeCache()
def getTraceEnable():
    return bool(SO.trace_enable(pointer(TRACE_CONFIG)))

@TimeCache()
def getTraceMode():
    # 1 TRACE_MODE_SAMPLE   2 TRACE_MODE_THRESHOLD  3 TRACE_MODE_SMART
    return SO.get_trace_mode(pointer(TRACE_CONFIG))

@TimeCache()
def getTraceDefaultGathererPercent():
    return SO.get_default_gatherer_percent(pointer(TRACE_CONFIG))

@TimeCache()
def getSmartModeLevel():
    return SO.get_smart_mode_level(pointer(TRACE_CONFIG))

def getTraceCollectHttpReqHeaders(header):
    return bool(SO.check_header_need_collect(pointer(TRACE_CONFIG), pointer(Str().with_value(header))))

def getTraceNeedCollectHeaders():
    header_list = POINTER(Str)()
    header_list_size = c_int(0)

    if response_to_bool(SO.get_need_collect_headers(pointer(TRACE_CONFIG), pointer(header_list), pointer(header_list_size))):
        hlist = []
        for i in range(header_list_size.value):
            hlist.append(header_list[i].value)
        return hlist

@TimeCache()
def getTraceCollectHttpReqHeadersEnable():
    return bool(SO.collect_http_req_headers_enable(pointer(TRACE_CONFIG)))

@TimeCache()
def getTraceEnableAddTraceidToApplog():
    if not SO:
        return False
    return bool(SO.enable_add_traceid_to_applog(pointer(TRACE_CONFIG)))

@TimeCache()
def getTraceEnableCollectHttpParams():
    return bool(SO.collect_http_params_enable(pointer(TRACE_CONFIG)))

@TimeCache()
def getTraceCollectRequestContentLen():
    return SO.get_collect_request_content_len(pointer(TRACE_CONFIG))

@TimeCache()
def getTraceCollectResponseContentLen():
    return SO.get_collect_response_content_len(pointer(TRACE_CONFIG))

# SocketConfig
def updateSocketConfig():
    return response_to_bool(SO.get_socket_config(pointer(SO_CONFIG), pointer(SOCKET_CONFIG)))

def getSocketMetricBytes(host, port):
    return SO.get_socket_metric_bytes(pointer(SOCKET_CONFIG), pointer(Str().with_value(host)), c_short(port))


# circuit breaker
def updateCircuitBreakerConfig():
    return SO.get_circuit_breaker_config(pointer(SO_CONFIG), pointer(CIRCUIT_BREAKER_CONFIG))

def getCircuitBreakerBool():
    return bool(SO.is_circuit_breaked(pointer(CIRCUIT_BREAKER_CONFIG)))


# Logging
def getLogsDir():
    if LOG_PROCESS:
        return os.path.join(LOGS_DIR, ('python_agent'+ str(os.getpid()) +'.log'))
    else:
        return os.path.join(LOGS_DIR, ('python_agent.log'))

def getApacheLogsDir():
    if LOG_PROCESS:
        return os.path.join(LOGS_DIR, 'apache_log', ('python_agent'+ str(os.getpid()) +'.log'))
    else:
        return os.path.join(LOGS_DIR, 'apache_log', ('python_agent.log'))

def getApacheLogsPath():
    return os.path.join(LOGS_DIR, 'apache_log')

def getLoggingEnable():
    return LOGGING_ENABLE

def getLoggingLevel():
    return LOGGING_LEVEL

# WSGI
def getWsgiModule():
    return WSGI_MODULE


def getWsgiScriptAlias():
    return WSGI_SCRIPT_ALIAS


def getWsgiCallableObject():
    return WSGI_CALLABLE_OBJECT

# js
def getJsSite():
    return JS_SITE

# environ config
def getEnvironIp():
    return ENVIRON_IP

def getEnvironPort():
    return ENVIRON_PORT
# environ config
def setEnvironIp(ip):
    global ENVIRON_IP
    ENVIRON_IP = ip
    # print('environ_ip %s' % (ENVIRON_IP,))

def setEnvironPort(port):
    global ENVIRON_PORT
    ENVIRON_PORT = port

USE_MICRO_SECOND = True
def getUseMicroSecond():
    return USE_MICRO_SECOND

# config
def getTrabsportInterval():
    return TRANSPORT_INTERVAL_S


def getReloadConfigurationInterval():
    return RELOAD_CONFIGURATION_INTERVAL_S


def getSnapshotProfilerInterval():
    return SNAPSHOT_PROFILER_INTERVAL_S


def getLoggingMaxBytes():
    return LOGGING_MAX_BYTES


def getLoggingMaxFiles():
    return LOGGING_MAX_NUM_FILES


@TimeCache()
def enableSQLExplain():
    return bool(SO.enable_sql_explain(pointer(BACKEND_CONFIG)))


# other config
MAX_CONCURRENT_SNAPSHOT = 20
MAX_EXIT_CALL_PER_THREAD = 100
MAX_EXIT_CALL_PER_BTR = 500
MAX_CALL_ELEMENT_PER_SNAPSHOT = 5000

# environ config
ENVIRON_CONFIG_FILE = None
ENVIRON_IP = None
ENVIRON_PORT = None

# container version
CONTAINER_VERSION = None


def get_container_version():
    return CONTAINER_VERSION


def set_container_version(version):
    global CONTAINER_VERSION
    CONTAINER_VERSION = version

# other config
def getMaxConcurrentSnapshot():
    return MAX_CONCURRENT_SNAPSHOT


def getMaxExitCallPerThread():
    return MAX_EXIT_CALL_PER_THREAD


def getMaxExitCallPerBtr():
    return MAX_EXIT_CALL_PER_BTR


def getMaxCallElementperSnapshot():
    return MAX_CALL_ELEMENT_PER_SNAPSHOT

def getListens():
    if isinstance(ENVIRON_PORT, list):
        ip_port = ''
        for port in ENVIRON_PORT:
            ip_port +=  str(ENVIRON_IP) + ":" + str(port) + '|'
        return ip_port[:-1]
    elif ENVIRON_PORT is None:
        return str(ENVIRON_IP)
    return str(ENVIRON_IP) + ":" + str(ENVIRON_PORT)


def is_content_type(content_type):
    if not content_type:
        return False

    if isinstance(content_type, bytes_t):
        content_type = content_type.decode()

    type_list = ['application/json', 'application/xml']
    for type in type_list:
        if type in content_type:
            return True
    return False


def collect_request_content(content):
    if content is None:
        return ''
    length = getTraceCollectRequestContentLen()
    if length <= 0:
        return None
    # content = bytes(content)
    return content[:length]


def collect_response_content(content):
    if content is None:
        return ''
    length = getTraceCollectResponseContentLen()
    if length <= 0:
        return None
    # content = bytes(content)
    return content[:length]


def updateSDKConfig(agent):
    if not update_cfg_obj():
        return False
    else:
        # ErrorConfig
        updateErrorConfig()
        # DebugConfig
        updateDebugConfig()
        # AppConfig
        updateAppConfig()
        # BackendConfig
        updateBackendConfig()
        # JsConfig
        updateJsConfig()
        # CallGraphConfig
        updateCallGraphConfig()
        # AppGatherConfig
        updateAppGatherConfig()
        # SnapshotGatherConfig
        # updateSnapshotGatherConfig()
        # BtConfig
        updateBtConfig()
        # BtFilterConfig
        updateBtFilterConfig()
        # BtTransConfig
        # updateBtTransConfig()
        # TraceConfig
        updateTraceConfig()
        # SocketConfig
        updateSocketConfig()
        # CircuitBreakerConfig
        updateCircuitBreakerConfig()
        # updateNamingConfigs
        updateNamingConfigs(agent)
        return True


class Config(ConfigParser):
    def __init__(self, defaults=None):
        ConfigParser.__init__(self, defaults=None)
        self._SECT_TMPL = r"""(?P<header>[^]]+):$"""
        self.SECTCRE = re.compile(self._SECT_TMPL, re.VERBOSE)

    def optionxform(self, optionstr):
        return optionstr


def parse_yaml_option(config, context, config_option):
    if config is None:
        return

    logger = logging.getLogger('Bonree.agent')

    for section in config.sections():
        for key, value in config.items(section):
            if key not in config_option:
                logger.debug('skipping unrecognized option %r ' % section)
                continue
            else:
                env_name, operation = config_option[key]
                if operation:
                    value = operation(value)
                context[env_name] = value


def parse_yaml_file(filename):
    """Parse an Bonree configuration file.

    """
    logger = logging.getLogger('Bonree.agent')
    logger.info("find config file: %s" % filename)
    try:
        config = Config()
        config.read(filename)

        conf_dict = {}
        parse_yaml_option(config, conf_dict, _CONFIG_OPTIONS)

        return conf_dict
    except Exception:
        logger.exception('Parsing config file failed.')


def parse_environ(environ=None, prefix='BONR_'):
    """Read Bonree configuration from an environment dictionary.

    Parameters
    ----------
    environ : mapping, optional
        A dict of environment variable names to values (strings). If not
        specified, `os.environ` is used.

    Other Parameters ted to have to be
        recognized as Bonree configuration. Defaults to `BONR_`.

    """
    request = environ
    environ = environ if environ is not None else os.environ

    config = {}
    # '/opt/bonree/apm/agent/python/{version}/bonree.yml'
    config_file = os.path.join(AGENT_INSTALL_DIR, 'bonree.yml')

    if os.path.exists(config_file):
        config = parse_yaml_file(config_file)

    option_descrs = {}

    for name, handler in values(_CONFIG_OPTIONS):
        option_descrs[prefix + name] = (name, handler)

    for option in environ.keys():
        if option not in option_descrs:
            continue

        name, handler = option_descrs[option]

        try:
            value = environ[option]
            if handler:
                value = handler(value)
            config[name] = value
        except Exception:
            pass
    return config


def merge(config):
    """Merge configuration into the module globals and update the computed defaults.

    """
    mod = globals()
    mod.update(config)
    update_computed_defaults(config)


def update_computed_defaults(config):
    global LOGS_DIR

    if LOGS_DIR is None:
        LOGS_DIR = DEFAULT_HOME_DIR

    if 'LOGS_DIR' in config and getLoggingEnable():
        logs_dir = config.get("LOGS_DIR")
        logs_dir = logs_dir.replace('{smartagent_apm_dir}', APM_INSTALL_DIR)
        if os.path.exists(logs_dir) and os.access(logs_dir, os.W_OK):
            LOGS_DIR = logs_dir
        else:
            try:
                os.makedirs(logs_dir, mode=0o777)
                LOGS_DIR = logs_dir
            except Exception:
                config['LOGS_DIR'] = DEFAULT_HOME_DIR
                LOGS_DIR = DEFAULT_HOME_DIR
