# Copyright (c) Bonree, Inc., and its affiliates
# 2015
# All Rights Reserved

"""Utilities for Bonree Python code.

"""
from __future__ import unicode_literals

import sys
import time
import weakref
import hashlib
import threading
import traceback
import wsgiref.util
import inspect
from functools import wraps
from itertools import chain
from logging import Formatter
from contextlib import contextmanager

from Bonree.lang import iteritems, SimpleCookie, parse_qsl, bytes_t, items

# default_log_formatter = Formatter('%(asctime)s [%(levelname)-5s] %(name)s <%(process)d|%(thread)d>: %(message)s', "%Y-%m-%d %H:%M:%S %Z")
default_log_formatter = Formatter('%(asctime)s [%(levelname)-5s] %(name)s <%(process)d|%(thread)d>: %(message)s')

class TimeCache(object):
    def __init__(self, expire=60):
        self.expire = expire
        self.start = 0
        self._cache_val = None

    def __call__(self, func):
        @wraps(func)
        def inner():
            now = time.time()
            if now - self.start > self.expire:
                self._cache_val = func()
                self.start = now
            return self._cache_val

        return inner

def current_thread_ident():
    return threading.current_thread().ident


def get_btr_from_attr():
    return get_current_btr(threading.current_thread())


def get_current_btr(thread_object):
    btr_weakref = getattr(thread_object, '_bonr_btr', None)
    if btr_weakref:
        return btr_weakref()
    else:
        return None


def set_current_ident_btr(btr):
    """the agent add the btr to the threading object
    for further be/bt correlation
    """
    concurrent_object = threading.current_thread()
    setattr(concurrent_object, '_bonr_btr', weakref.ref(btr))


def unset_current_ident_btr():
    concurrent_object = threading.current_thread()
    setattr(concurrent_object, '_bonr_btr', None)

def set_current_ident_be_call(be_call):
    """the agent add the btr to the threading object
    for further be/bt correlation
    """
    if be_call:
        concurrent_object = threading.current_thread()
        setattr(concurrent_object, '_bonr_be_call', weakref.ref(be_call))

def unset_current_ident_be_call():
    concurrent_object = threading.current_thread()
    if hasattr(concurrent_object, '_bonr_be_call'):
        setattr(concurrent_object, '_bonr_be_call', None)

def get_current_be_call():
    be_call_weakref = getattr(threading.current_thread(), '_bonr_be_call', None)
    if be_call_weakref:
        return be_call_weakref()
    else:
        return None



def now():
    return set_time_unit(time.time())

def now_ms():
    return int(time.time()*1000)

def normalize_header(hdr):
    """Normalize a header name.

    Normalization involves stripping leading and trailing spaces, turning
    underscores into dashes, and lowercasing the name.

    NOTE: WSGI headers come in like "HTTP_USER_AGENT". The caller is
    responsible for removing the "HTTP_" prefix before normalizing the header.

    """
    return hdr.strip().lower().replace('_', '-')


def arr_http_parameters(http_dict):
    if not http_dict:
        return
    parameters = ''
    for k, v in items(http_dict):
        parameters = parameters + "{0}={1}".format(k, v) + r'&'
    return parameters[:-1]

def arr_http_header(http_dict):
    if not http_dict:
        return
    parameters = ''
    for k, v in items(http_dict):
        parameters = parameters + "{0}: {1}".format(k, v) + r'\r\n'
    return parameters[:-4]


class LazyWsgiRequest(object):
    """Lazily read request line and headers from a WSGI environ.

    This matches enough of the Werkzeug Request API for the agent's needs: it
    only provides access to the information in the request line and the
    headers that is needed for the agent. (Since the agent doesn't inspect
    the request body, we don't touch any of that.)

    Parameters
    ----------
    environ : dict
        A WSGI environment.

    Attributes
    ----------
    headers : dict
        A dictionary of the HTTP headers. The headers are lowercase with
        dashes separating words.
    method : str
        The request method (e.g., GET).
    url : str
        The URL of the request (reconstructed according to PEP 333).
    cookies : dict
        The cookies passed in the request header (if any).
    path : str
        The path part of the request. Note that unlike raw WSGI, this will be
        just '/' if it would otherwise be empty.
    args : dict
        The query parameters. This is not a multi-dict: if a parameter is
        repeated multiple times, one of them wins.
    referer : str
        The HTTP Referer string.
    user_agent : str
        The HTTP User-Agent string.
    is_ajax : bool
        True if this request is AJAX.
    is_mobile : bool
        True if this request is from mobile.

    """
    DEFAULT_PORTS = {
        'http': 80,
        'https': 443,
    }
    __slots__ = ('environ', '_headers', '_host', '_port', '_http_host', '_url', '_path', '_config_path', '_args',
                 '_cookies')

    def __init__(self, environ):
        self.environ = environ
        self._headers = self._host = self._port = self._http_host = self._url = self._path = self._config_path = self._args = self._cookies = None

    @property
    def headers(self):
        if self._headers is not None:
            return self._headers

        headers = {}
        for key, value in iteritems(self.environ):
            if key.startswith('HTTP_'):
                header_name = normalize_header(key[5:])
                headers[header_name] = value

        self._headers = headers
        return headers

    @property
    def method(self):
        return self.environ.get('REQUEST_METHOD')

    @property
    def is_secure(self):
        return self.environ.get('wsgi.url_scheme') == 'https'

    @property
    def host(self):
        if self._host is None:
            host = self.environ.get('HTTP_X_FORWARDED_HOST')

            if host:
                self._host = host.split(',')[0].strip()
            else:
                self._host = self.environ.get('HTTP_HOST')

            if self._host is None:
                host = self.environ.get('SERVER_NAME')
                port = str(self.environ.get('SERVER_PORT'))

                if (self.environ.get('wsgi.url_scheme'), port) not in (('http', '80'), ('https', '443')):
                    self._host = '%s:%s' % (host, port)
                else:
                    self._host = host

        return self._host

    @property
    def port(self):
        host = self.host
        if host:
            host_or_port = host.split(':')
            if len(host_or_port) >= 2:
                self._port = [int(host_or_port[1])]
        if not self._port:
            self._port = [int(self.environ.get('SERVER_PORT'))]
        return self._port

    @property
    def url(self):
        if self._url is None:
            self._url = wsgiref.util.request_uri(self.environ, True)
        return self._url

    @property
    def cookies(self):
        if self._cookies is not None:
            return self._cookies

        cookies = {}
        cookie_header = self.environ.get('HTTP_COOKIE', None)

        if cookie_header:
            cookie_jar = SimpleCookie()

            try:
                cookie_jar.load(cookie_header)
            except Exception:
                self._cookies = {}
                return self._cookies

            for name in cookie_jar:
                cookies[name] = cookie_jar[name].value

        self._cookies = cookies
        return cookies

    @property
    def path(self):
        if self._path is not None:
            return self._path

        self._path = self.environ.get('PATH_INFO', '/')
        return self._path

    @property
    def config_path(self):
        if self._config_path is not None:
            return self._config_path

        self._config_path = self.environ.get('BONR_CONFIG_FILE')
        return self._config_path

    @property
    def args(self):
        if self._args is not None:
            return self._args

        query_string = self.environ.get('QUERY_STRING', '')
        self._args = dict((k, v) for k, v in parse_qsl(query_string))
        if 'body_arguments' in self.environ:
                for k, v in self.environ.get('body_arguments').items():
                    self._args[k] = v
        return self._args

    @property
    def referer(self):
        return self.environ.get('HTTP_REFERER')

    @property
    def user_agent(self):
        return self.environ.get('HTTP_USER_AGENT')

    @property
    def is_ajax(self):
        rum_header = self.environ.get('HTTP_ADRUM')
        if rum_header and 'isAjax:true' in rum_header:
            return True
        x_requested_header = self.environ.get('HTTP_X_REQUESTED_WITH')
        if x_requested_header == 'XMLHttpRequest':
            return True
        return False

    @property
    def is_mobile(self):
        rum_header = self.environ.get('HTTP_ADRUM_1')
        return rum_header and 'isMobile:true' in rum_header


class ThriftRequest(object):
    def __init__(self, bt_name=None, headers=None):
        self.bt_name = bt_name
        self.headers = headers

    @property
    def url(self):
        return self.bt_name

    @property
    def path(self):
        return self.bt_name


def get_request_host_and_port(request):
    """Return a tuple of the host and port from an incoming HTTP request.

    Parameters
    ----------
    request : request-like object
        The request must have `environ` and `host` attributes with the same
        meaning as those attributes in Werkzeug's BaseRequest class.

    Returns
    -------
    (host, port)
        The host and the port. The port is set even if it's the default for
        requests of this scheme.

    """
    host = request.host
    port = None

    if ':' in host:
        host, port = host.split(':', 1)

    return host, (port or request.environ.get('SERVER_PORT'))


def parse_args(args, kwargs):
    """Convert function arguments into a neat string.

    e.g.
       def foo(*args, **kwargs):
           print parse_args(args, kwargs)

       info = {'name': 'Dan', 'beard': True}
       foo(info, expires='2015-03-18')

    would print "{'name': 'Dan', 'beard': True}, expires='2015-03-18'"

    """
    args = (repr(x) for x in args)
    kwargs = ('%s=%r' % x for x in iteritems(kwargs))
    return ', '.join(chain(args, kwargs))


def dic_to_string(dic, format_string='{0}={1}', sep=',', filter='='):
    """
    Trans dictionary to a string
    format_string: the format about dictionary how to trans to string
    sep: string list join by sep
    fliter: if dic key or value include fliter, this key-value while not include in string
    """
    strlist = []
    for k, v in dic.items():
        if str(k).find(filter) != -1 or str(v).find(filter) != -1:
            continue
        strlist.append(format_string.format(k, v))

    return sep.join(strlist)


def args2str(method_name, *args, **kwargs):
    from Bonree import config
    host_list = config.getHotMethodList()
    method_name_list = set()
    for method_list in host_list:
        for method in method_list['methodNames']:
            method_name_list.add(method)

    str_args = []
    if method_name in method_name_list:

        for arg in args:
            if isinstance(arg, (str, int, float, bool, bytes_t)):
                str_args.append(encode_to_utf8(arg))
            elif arg != None:
                str_args.append(encode_to_utf8(arg.__class__.__name__))

        for value in kwargs.values():
            if isinstance(value, (str, int, float, bool, bytes_t)):
                str_args.append(encode_to_utf8(value))
            elif value != None:
                str_args.append(encode_to_utf8(value.__class__.__name__))
        return str_args

    return str_args


def encode_to_utf8(input):
    from Bonree.lang import str
    return str(input)


func_signa_cache = {}

def get_func_signa(method):
    PY3 = sys.version_info[0] > 2
    if PY3:
        method_name = method.__qualname__
        try:
            return func_signa_cache[method_name]
        except KeyError:
            sign = str(inspect.signature(method))
            func_signa_cache[method_name] = sign
            return sign
    else:
        sign_tuple = inspect.getargspec(method)
        return inspect.formatargspec(*sign_tuple)


def set_time_unit(time):
    return int(1000000*time)

def ms_to_us(time):
    return int(1000*time)

def us_to_ms(time):
    return int(time / 1000)