# Copyright (c) Bonree, Inc., and its affiliates
# 2015
# All Rights Reserved
import re
import os
import sys
import logging
import logging.handlers

from Bonree import config
from Bonree.lib import default_log_formatter

__Logger = None


class MyRotatingFileHandler(logging.handlers.RotatingFileHandler):
    def _open(self):
        prevUmask=os.umask(0o011)
        rtv=logging.handlers.RotatingFileHandler._open(self)
        os.umask(prevUmask)
        return rtv

def configure_logging():
    """Configure the Bonree agent logger.

    By default, we configure a log file which rolls over at midnight and keeps a week's worth of logs.  If the logging
    debug flag is set, we set the logging level to DEBUG and enable a handler to stderr.

    """
    if not config.getLoggingEnable():
        return
    try:
        level = get_log_level()

        logger = logging.getLogger('Bonree.agent')
        logger.setLevel(level)

        rotating_file_handler = MyRotatingFileHandler(config.getLogsDir(),
                                                    maxBytes=config.getLoggingMaxBytes(),
                                                    backupCount=config.getLoggingMaxFiles() - 1)

        rotating_file_handler.setLevel(level)
        rotating_file_handler.setFormatter(default_log_formatter)
        logger.addHandler(rotating_file_handler)

        # if config.getLoggingLevel().upper() == 'DEBUG':
        #     stream_handler = logging.StreamHandler(sys.stdout)
        #     stream_handler.setLevel(level)
        #     stream_handler.setFormatter(default_log_formatter)
        #     logger.addHandler(streaM_handler)

        logger.propagate = False

    except Exception:
        logging.getLogger('Bonree.agent').exception('Logging configuration failed.')


def get_log_level():
    default_logging_level = logging.WARNING
    allowed_logging_levels = {'WARNING': logging.WARNING, 'INFO': logging.INFO, 'DEBUG': logging.DEBUG}

    level = config.getLoggingLevel().upper()
    return allowed_logging_levels.get(level, default_logging_level)


# def get_log_filename():
#     non_alphanumeric = re.compile(r'\W+')
#     sanitize = lambda x: non_alphanumeric.sub('_', x)
#     filename = '-'.join(map(sanitize, [config.getAgentName(), config.getAgentName()])) + '.log'
#     return os.path.join(config.LOGS_DIR, filename)


def get_logger():
    global __Logger
    if __Logger is None and config.getLoggingEnable():
        __Logger = logging.getLogger('Bonree.agent')
    return __Logger


def dumplog(self, level, str_msg, dic_msg):
    pass


def dic_msg_format(msg):
    from json import dumps
    return dumps(msg, indent=1)


class Logger(object):
    def __init__(self, label):
        super(Logger, self).__init__()
        self.label = label

    def _log(self, level, str_msg, dic_msg):
        agent_log = get_logger()
        if agent_log:
            msg = "{0}{1}{2}".format(self.label, str_msg, dic_msg_format(dic_msg) if dic_msg else '')
            getattr(agent_log, level)(msg)

    def debug(self, str_msg, dic_msg=None):
        self._log('debug', str_msg, dic_msg)

    def info(self, str_msg, dic_msg=None):
        self._log('info', str_msg, dic_msg)

    def warning(self, str_msg, dic_msg=None):
        self._log('warning', str_msg, dic_msg)

    def error(self, str_msg, dic_msg=None):
        self._log('error', str_msg, dic_msg)

    def critical(self, str_msg, dic_msg=None):
        self._log('critical', str_msg, dic_msg)

    def exception(self, str_msg, dic_msg=None):
        self._log('exception', str_msg, dic_msg)


class TransactionLogger(Logger):
    pass


class ProxyLogger(Logger):
    pass


class BonreeLogger(Logger):
    def _log(self, level, str_msg, dic_msg):
        if config.BONREE_LOG:
            msg = "{0}{1}{2}".format(self.label, str_msg, dic_msg_format(dic_msg) if dic_msg else '')
            getattr(get_logger(), level)(msg)


SYS_LOGGER = Logger('<sys>')
SDK_LOGGER = ProxyLogger('<sdk>')
BT_LOGGER = TransactionLogger('<transaction>')
BONREE_LOGGER = BonreeLogger('<bonree>')
# SANPSHOT_LOGGER = SnapshotLogger('<snapshot>')


class _IgnoreException(object):
    def __enter__(self):
        return None

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            SYS_LOGGER.exception(exc_val)
        # if `__ exit__` return `True` , the exception is handled gracefully
        return True


only_log_exception = _IgnoreException()
