# Copyright (c) Bonree, Inc., and its affiliates
# 2015
# All Rights Reserved

import errno
import os
import shlex

from Bonree.scripts.bonrpy import CommandInvocationError, CommandExecutionError

USAGE = "run <options> -- <command> [args...]"

ABOUT = """Run a program with the agent enabled.

Use this command to instrument a Python application. If <command> is not the
path to an executable, the command is looked for in your PATH. The command may
be a Python interpreter or a supported server that embeds Python (like uwsgi
and gunicorn).

For example, if you normally run your application using gunicorn:

    gunicorn -w 4 -b unix:acme.sock acme.app:app

You can run the same application instrumented by Bonree with:

    {bonrpy} run -c acme.cfg -- gunicorn -w 4 -b unix:acme.sock acme.app:app

The -c (--config-file) option specifies an agent configuration file.

The agent requires the app, node, cluster, tier, and node names. If any of these
values is not set, your application will be run without instrumentation (the agent
will be disabled).
"""

OPTIONS = {
    'config-file': {
        'short': 'c',
        'help': 'the config file to use',
        'value': True,
        'value_help': '<file>',
    }
}


def command(options, args):
    import Bonree.autoinject
    pythonpath = [os.path.dirname(Bonree.autoinject.__file__), '.']

    if not args:
        raise CommandInvocationError('missing command: run <options> -- <command> [args...]')

    environ = os.environ

    if 'PYTHONPATH' in environ:
        pythonpath.append(environ['PYTHONPATH'])

    environ['PYTHONPATH'] = os.pathsep.join(pythonpath)

    if 'config-file' in options:
        environ['BONR_CONFIG_FILE'] = options['config-file']
    if 'app' in options:
        environ['BONR_APP_NAME'] = options['app']
    if 'tier' in options:
        environ['BONR_TIER_NAME'] = options['tier']
    if 'node' in options:
        environ['BONR_NODE_NAME'] = options['node']
    if 'cluster' in options:
        environ['BONR_CLUSTER_NAME'] = options['cluster']

    if 'ssl' in options:
        environ['BONR_SSL_ENABLED'] = 'on'
    elif 'no-ssl' in options:
        environ['BONR_SSL_ENABLED'] = 'off'

    environ['BONR_STARTUP_ARGUMENTS'] = '\r\n'.join(args)

    try:
        os.execvpe(args[0], args, environ)
    except OSError as exc:
        if exc.errno == errno.ENOENT:
            raise CommandExecutionError('%s: no such file or directory' % args[0])
        elif exc.errno == errno.EPERM:
            raise CommandExecutionError('%s: permission denied' % args[0])
        raise
